#' Determinant of a Symbolic Matrix by Sweeping
#'
#' @param A A square matrix
#' @param loc vector of pivot locations
#' @param log = 1 to calculate log determinant
#' @param simplify = 0 not to use Simplify nor remove_paren \cr
#' = 1 to use Simplify
#' @param debug = 1 to print details
#'
#' @details
#' Definition of Sweep Operator\cr
#' C <- matSweep( A, p ) means: \cr
#'  c[i,j] <- c[i,j] - c[i,p]*c[p,j]/c[p,p] \cr
#'  c[p,] <- c[p,]/c[p,p] \cr
#'  c[,p] <- - c[,p]/c[p,p] \cr
#'
#' # Inversion of A.  \cr
#' matSweep( A, 1:ncol(A) ) == solve(A) \cr
#' This is equivalent to Inv(A).
#' \cr\cr
#' The determinant of A is given as the product of all the pivots, c[p,p],
#' when inverting A using sweep operator.
#'
#' @return
#' The determinant of A
#'
#' @examples
#'
#' # determinant of 2 x 2 A
#' A <- demomat(2,2,root="a")
#' Det(A)
#' Det(A,log=1)
#'
#' # n x n ( slow !)
#' set.seed(1701)
#' n=3
#' B <- demomat(n,n,root="b")
#' dd <- Det(B)
#' Bn <- matrix(runif(n*n),n)
#' Print(Eval(dd,B=Bn),det(Bn))
#'
#' # Fully Symbolic Partitioned Matrix: No SMdim attribute but
#' # fullsymb attribute
#' Ap=demomat(2,2,root="A", fullsymb=1)
#' Print(Ap)
#' Det(Ap)
#' Det(Ap, 2:1)
#'
#' @export
#'

Det <- function( A, loc=1:nrow(A), log=0, simplify=0, debug=0 ){
 # matrix sweep operator
 # Shin-ichi Mayekawa
 # 20150823
 # 20150823 Maiko
 # simplify: 20150823,24
 # simplify=1: 20150906
 # error check: 20150908
 # 20150915
 # fullsymb attribute: 20200528
 #

 # errors
 if( !is.matrix(A) ){
  cat("\n error1:(matSweep) Input must be a matrix! \n")
  return()
 }

 nr=nrow(A); nc=ncol(A)
 if( nr != nc ){
  cat("\n error1:(matSweep) Input must be square! \n")
  return()
 }

 lenp=length(loc)

 # SMdim and fullsymb attribute
 SMdim=attributes(A)$SMdim
 fullsymb=attributes(A)$fullsymb

 if( !is.null(fullsymb) ){

   #
   # Here, A is a fully symbolic matrix.
   #
   pivots=character(nr)
   C=A
   for( pv in loc[1:lenp] ){

     cpivot=C[pv,pv]

     pivots[lenp-pv+1]=cpivot

     r1=NULL; if( pv > 1 ) r1=1:(pv-1)
     r2=NULL; if( pv < nc ) r2=(pv+1):nc
     r12=c(r1,r2)

     if( debug ) Print(pv,r12, cpivot)

     cpvrow=C[pv,]; cpvcol=C[,pv]

     for( j in r12 ){
       for( k in r12 ){
         # c(jk)=c(jk)-c(jp)*c(pk)/c(pp)
         C[j,k]=
           paste( C[j,k], "-(", C[j,pv], ")%*%inv(", cpivot, ")%*%(", C[pv,k]
                  , ")", sep="" )
       }
     }

     C[pv,]=paste( "inv(", cpivot, ")%*%(", cpvrow, ")", sep="" )
     C[,pv]=paste( "-(", cpvcol, ")", "%*%inv(", cpivot, ")" , sep="" )
     C[pv,pv]=paste( "inv(", cpivot, ")", sep="" )

     if( simplify >= 2 ){
       C=Expand(C)
       C=Simplify(C) # 20151006
     }

   } # end of pivot


   # final treatment
   if( simplify >=1  ){
     pivots=Simplify(Expand(pivots))
   }

   pivots=map("det", pivots)
   if( log == 1 ) res=Sum(map("log", pivots))
   else res=Prod(pivots)

   return( res )

 } # end of full symbolic A matrix



 #
 # Here, A is a semi-symbolic matrix.: no fullsymb attribute
 #

 pivots=character(nr)
 C=A
 for( pv in loc[1:lenp] ){

  cpivot=C[pv,pv]

  pivots[lenp-pv+1]=cpivot

  r1=NULL; if( pv > 1 ) r1=1:(pv-1)
  r2=NULL; if( pv < nc ) r2=(pv+1):nc
  r12=c(r1,r2)

  if( debug ) Print(pv,r12, cpivot)

  cpvrow=C[pv,]; cpvcol=C[,pv]

  for( j in r12 ){
   for( k in r12 ){
    # c(jk)=c(jk)-c(jp)*c(pk)/c(pp)
    C[j,k]=paste( C[j,k], "-(", C[j,pv], ")*(", C[pv,k], ")/(", cpivot, ")"
                  , sep="" )
   }
  }

  C[pv,]=paste( "(", cpvrow, ")", "/(", cpivot, ")" , sep="" )
  C[,pv]=paste( "-(", cpvcol, ")", "/(", cpivot, ")" , sep="" )
  C[pv,pv]=paste( "1/(", cpivot, ")", sep="" )

  # C=Simplify(C)

 } # end of pivot loop


 # final treatment
 if( simplify >= 2  ){
  pivots=Expand(pivots)
 }

 # Print(pivots, map("log",pivots))
 if( log == 1 ) res=Sum(map("log",pivots))
 else res=Prod(pivots)

 if( simplify >=2  ){
   res=Expand(res)
 }

 return( res )

} # end of Det


