#' Determinant of a Symbolic Matrix by Laplace Expansion
#'
#' @param A A square matrix (numeric or symbolic)
#' @param row row number for Laplace Expansion
#' @param col col number for Laplace Expansion
#' @param maxlevel depth of recursive calls
#' @param simplify = 0 not to use Simplify nor remove_paren \cr
#' = 1 to use Simplify
#' @param debug = 1 to print details
#' @param level Do Not Modify This.
#'
#' @details
#' THe Laplace Expansion is defined as:
#' \preformatted{
#'  DetL(A) = \sum_{j=1}^n  A[i,j] x cofactor(A,i,j)
#'  cofactor(A,i,j) = (-1)^{i+j} DetL( A[-i,-j] )
#' }
#' where DetL( A[-i,-j] ) is called the minor M(i,j) of A.
#'
#' @return
#' The determinant of A
#'
#' @examples
#'
#' # determinant of 2 x 2 matrix
#' Aa2 <- demomat(2,2,root="a")
#' DetL(Aa2)
#'
#' # Triangular matrix
#' L=demomat(4,4,root="L", shape="lowert")
#' DetL(L, maxlevel=3)
#' U=demomat(4,4,root="U", shape="uppert")
#' DetL(U, maxlevel=3)
#'
#' # 3 x 3 matrix
#' Aa3 <- matrix(letters[1:9],3,3,byrow=1)
#' # With 2nd order minors
#' DetL(Aa3)
#' DetL(Aa3, row=2)
#'
#' # full Laplace Expansion
#' ( d1 <- DetL(Aa3, maxlevel=2) )
#' ( d2 <- DetL(Aa3, row=2, maxlevel=2) )
#' ( d3 <- DetL(Aa3, row=NULL, col=1, maxlevel=2) )
#'
#' # Numerical verification
#' set.seed(1701)
#' An <- matrix( runif(9), 3, 3 )
#' Printm(det(An))
#' Printm( Eval(d1, Aa3=An), Eval(d2, Aa3=An), Eval(d3, Aa3=An) )
#'
#' # Comparison with Det function using MatSwp.
#' ( d4 <- Det(Aa3) )
#' Printm( Eval(d4, Aa3=An) )
#'
#'
#' # 4 x 4 matrix
#' Aa4 <- matrix(letters[1:16],4,4,byrow=1)
#' DetL(Aa4)
#' # Below, A2 stands for a submatrix of the first level.
#' # Note that not all the A2 are the same!!
#' DetL(Aa4, maxlevel=2)
#' DetL(Aa4, maxlevel=3)
#'
#' # General n x n case. (When n is large, it is slow.)
#' n <- 4
#' ( B <- demomat(n,n,root="b") )
#' ( dd <- DetL(B, maxlevel=n) )
#'
#' # Numerical verification
#' set.seed(1701)
#' Bn <- matrix( runif(n*n) ,n )
#' Printm( Eval(dd,B=Bn), det(Bn) )
#'
#' @export
#'


DetL <- function( A, row=1, col=NULL, maxlevel=1, level=0
                  , simplify=0, debug=0 ){
 # determinant by Laplace Expansion
 # Shin-ichi Mayekawa
 # 20250228,0301
 # Detl --> DetL: 20250303
 #

 #
 # Seems the we cannot use modify_math_operators() here,
 # possibly because of the recursive call.
 #
 
 ##########################################################################
 ################ Infix Operators from lazy.symbplic2 package #############
 ##########################################################################
 `+` <- function( A, B=NULL ){
  # + operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   if( is.null(B) ) return( Plus(0,A))
   else return( Plus(A,B) )
  }
  else{
   if( is.null(B) )  return(  .Primitive("+")(0,A) )
   else return(  .Primitive("+")(A,B) )
  }
 } # end of +

 `-` <- function( A, B=NULL ){
  # - operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   if( is.null(B) ) return( Minus(0,A))
   else return( Minus(A,B) )
  }
  else{
   if( is.null(B) )  return(  .Primitive("-")(0,A) )
   else return(  .Primitive("-")(A,B) )
  }
 } # end of -

 `*` <- function( A, B ){
  # * operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   return( Times(A,B) )
  }
  else{
   return(   .Primitive("*")(A, B) )
  }
 } # end of *
 ##########################################################################
 ##########################################################################
 ##########################################################################






 if( debug ) printm(level,A)

 # errors
 if( !is.matrix(A) ){
  cat("\n error1:(DetL) Input must be a matrix! \n")
  return()
 }
 if( maxlevel <= 0 ){
  cat("\n error1:(DetL) maxlevel must be a positive integer. \n")
  return()
 }

 nr=nrow(A); nc=ncol(A)
 if( nr != nc ){
  cat("\n error1:(DetL) Input must be square! \n")
  return()
 }

 # SMdim and fullsymb attribute
 SMdim=attributes(A)$SMdim
 fullsymb=attributes(A)$fullsymb

 if( !is.null(fullsymb) ){
  cat("\n error1:(DetL) Cannot handle fully symbolic matrix yet. \n")
  stop()
 } # end of full symbolic A matrix


 # stop by level
 if( level >= maxlevel ){
  AA=deparse(substitute(A))
  ii=get("i", envir=parent.frame())
  jj=get("j", envir=parent.frame())
  if( debug ) printm(A,AA,ii,jj)
  AA=strsplit(AA,"[", fixed=TRUE)[[1]][1]
  if( level > 1 ) AA=paste(AA,level)
  if( debug ) Printm(AA)
  dd=paste("DetL(", AA, "[-", ii, ",-", jj, "])", sep="")
  return(dd)
 }

 # end of recursive call
 if( nr <= 2 ){
  if( nr == 2 ) dd=A[1,1]*A[2,2]-A[1,2]*A[2,1]
  else if( nr == 1 ) dd=A[1,1]
  return( dd )
 }


 #
 # Here, nrow(A) > 2
 #

 level=level+1

 if( !is.null(row) ){

  i=row

  dd=0
  for( j in 1:nr ){

   dd=dd+(-1)^(i+j) * A[i,j] * DetL( A[-i,-j, drop=FALSE]
                                     , level=level, maxlevel=maxlevel )

  }

 }
 else if( !is.null(col) ){

  j=col

  dd=0
  for( i in 1:nc ){

   dd=dd+(-1)^(i+j) * A[i,j] * DetL( A[-i,-j, drop=FALSE]
                                     , level=level, maxlevel=maxlevel )

  }

 }

 if( debug ) printm( level, dd )
 # final treatment

 if( simplify ){
  if( simplify >=2  ){
   dd=Expand( dd )
  }
  dd=Simplify( dd )
 }

 return( dd )

} # end of DetL


