#' Evaluate Symbolic Matrix after Replacement
#'
#' @param ..exp A symbolic matrix whose elements are to be evaluated
#' after the replacement.
#' @param ... Replacement of the form pat1=pat2
#' @param values A character vector consisting of several replacements
#' separated by semicolons.
#' @param fullsymb = 1 to treat \code{exp} as fully symbolic expression. \cr
#' See the description of \code{matReplace} function.
#' @param check = 0 not to check if the expression is evaluatable.
#' (See below.)
#' @param ntogoback n parameter of sys.parent function
#' @param debug debug parameter for matReplace
#'
#' @details
#' This function first replaces the elements of pat1 matrix included in exp
#' with the corresponding elements of pat2 matrix. \cr
#' Then, if the resulting expression contains numbers only, it evaluates
#' the result. \cr\cr
#'
#' See the description of matReplace function for details. \cr
#' \cr
#' This function does NOT attempt to evaluate the expression
#' if the result of the replacement has any alphabets. \cr
#' Therefore, if the expression has a function call, with check=1,
#' it will not be evaluated.\cr
#' Use check=0 if you  are sure that the resulting expression is valid.
#'
#' @return A numeric matrix of the same size as exp
#'
#' @examples
#'
#' # matrix replacement
#' A <- demomat(2,2, root="a")
#' B <- demomat(2,2, root="b")
#' An <- demomat(2,2)
#' Bn <- demomat(2,2)*1000
#' ApB <- Plus(A,B)
#' ApBn <- Eval(ApB, A=An, B=Bn)
#' ApBn2 <- Eval(ApB, values=c("A=An; B=Bn") )
#' Print(ApB, ApBn, An+Bn)
#' Print(ApBn, ApBn2)
#'
#' AtB <- matTimes(A,B)
#' AtBn <- Eval(AtB, A=An, B=Bn)
#' Print(AtB, AtBn, An%*%Bn)
#'
#' # scalar replacement
#' val <- "a=1; b=2; c=3; d=4; e=5"
#' exp <- "(a+b*c-d)/e"
#' Eval(exp,values=val)
#'
#' # matrix replacement: fully symbolic expression
#' res=Eval("matTimes(t(A),B)"
#'         , A=demomat(2,2), B=demomat(2,2), fullsymb=1)
#'
#' # Below causes an error.
#' # Eval(A)
#'
#' @export
#'

Eval <- function( ..exp, ..., values=NULL, fullsymb=NULL
                  , check=1, ntogoback=2, debug=0 ){
 # evaluate symbolic expression which consists of numbers
 # Shin-ichi Mayekawa
 # 20150817,18
 # replacement: 20150907
 # values: 20150918
 # check: 20150929
 # ..exp: 20200609
 # fullsymb: 20200630
 # modify_math_operators: 20250315
 #

 # modified operators
 modify_math_operators( where="here", print=0 )

 # avoid partial match
 exp=..exp

 # store ...  (not used)
 name <- as.list(substitute(list(...)))[-1L]
 val <- list(...)

 # replace
 if( length(val) > 0  |  !is.null(values) ){
  exp=matReplace( exp, ..., values=values, fullsymb=fullsymb
                  , ntogoback=ntogoback, debug=debug )
 }

 # all numeric?
 if( check == 1  &  length(grep("[[:alpha:]]", exp)) > 0 ){
  cat("\n The input expression cannot be evaluated.\n")
  return( exp )
 }

 # evaluate
 exp[]=unlist( lapply(parse(text=exp),eval) )

 if( is.matrix(exp) )  exp=matrix( as.numeric(exp), nrow(exp) )
 else exp=as.numeric(exp)

 return( exp )

} # end of Eval


