#' Expand Expressions by Distributing a Multiplicative Factor
#'
#' @param expr A vector or list of symbolic expressions
#' @param method = 1 to expand a*(...) \cr
#' = 2 to expand (...)*a
#' @param all = 1 to apply Expand repeatedly
#' @param simplify = 0 to skip the simplification of the result
#' @param verify = 0 not to check the result by \code{is_valid}
#' @param debug = 1 to print intermediate result
#'
#' @details
#' This function expands a single occurence of \cr
#' \code{a*(b+c)} and \code{(a+b)*c} \cr
#' to \cr
#'  \code{a*b+a*c}  and \code{a*c+b*c}, resp., \cr
#' where \code{a} is a single multiplicative factor and
#' \code{b} and \code{c} are additive terms. \cr\cr
#' Matrix multiplications such as \%*\% and \%@\% can be used. \cr
#' \cr
#' Repeated use may expand it further.
#'
#' @return
#' Expanded expression
#'
#' @examples
#' Expand(c("a*(b+c)","-(b+c)*a"))
#' Expand(c("a/(b*c)","-(b*c)/a"))
#'
#' Expand("(a+b)*(F(x)-G(y))") |> print() |> Expand()
#'
#' Expand("-a*(b-c)*(d+e)") |> print() |> Expand()
#'
#' Expand( "((+a*(b-c*d)*e+f)*g)*h - (((+a*b-c*d)*e+f)*g)*h" ) |>
#' print() |> Expand()
#'
#' Expand("-t(A)%*%(B-C)%*%D")
#'
#' Expand("-inv(A)%*%(A-B)")
#' Expand("-inv(A)%*%(A-B)", simplify=0)
#'
#' Expand("-(A-B)%*%(C-D)+E")
#' Expand("-(A-B)%*%(C-D)+E", all=1)
#'
#' @export
#'

Expand <- function( expr, all=0, method=1:2, simplify=1, verify=1, debug=0 ){
 # expand multiplicative factor
 # Shin-ichi Mayekawa
 # 20150906
 # post: 20150907
 # recursive: 20150907
 # -(...): 20150909
 # last_term and first_term: 20150918
 # bugfix: 20150919
 # onetwo -> method: 20150920
 # bugfix for method=2: 20150921,22
 # special operators: 20150922
 # (Use Expand7.R if this version does not work.)
 # func(...)*a in method2: 20150928
 # first_term corrected for summation: 20151003,04
 # bugfix for (a+b)*(c+d): 20151004
 # first Simplify removed: 20200519
 # bugfix for %*% and %@%; ` and ~ added to op list: 20200519
 # first Simplify with bug fix] 20200520
 # function Simplify fixed: 20200520
 # grep pattern improved: 20200530
 # roxygen2: 20200610
 # final gsub: 20200627
 # verify: 20200627
 # all: 20200627
 # initial value of pre0=post0="": 20200628
 # bugfix for a/(b/c): 20200816
 # new first_term for method=1: 20200818,19
 # bugfix for a/(b/c): 20200819,20,21
 # debug after trim: 20230428
 # numeric input: 20230503
 # bug fix for "F(a/(b/c))": 20230513
 # exp -> expr: 20230513
 # .Last.value --> pipe: 20250314
 #



 ##################### first and last term #######################
  first_term <- function( expr, valid=1, markso=0, debug=0 ){
    # new version which uses find_a_term,
    # Shin-ichi Mayekawa
    # 20200818
    # exp -> expr: 20230513
    #

    if( markso ){
      expr=mark_soper( expr )
    }
    res=find_a_term( expr )
    term=paste(res$pre,res$term,sep="")
    fromto=cbind(res$loc,res$loc+res$len-1)
    if( markso ){
      expr=unmark_soper( expr )
    }

    return( list( term=term, fromto=fromto ) )

  } # end of first_term


 last_term <- function( expr, valid=1, debug=0 ){
  # find the last term of expr which may be incomplete
  # Shin-ichi Mayekawa
  # 20150917
  # valid: 20150918
  # exp -> expr: 20230513
  #
  # return
  #  A list of term and fromto
  #  where term contains the last term (with + or - sign if any)
  #  and fromto contains the begining location of the last term with sign.
  #

  if( debug ) Print("####### Top of last_term ######",expr)
  # simple case which ends with + or -
  lenexp=nchar(expr)
  if( substr(expr, lenexp,lenexp) %in% c("+","-","!","#") ){
   return( list(term="", fromto=cbind(lenexp,lenexp)) )
  }
  else if( substr(expr, lenexp-2,lenexp) %in% c("%p%", "%m%") ){
   return( list(term="", fromto=cbind(lenexp-2,lenexp)) )
  }

  # reverse order
  rexp=revchar(expr)
  rexp=chartr( "(", "}", rexp)
  rexp=chartr( ")", "(", rexp)
  rexp=chartr( "}", ")", rexp)

  res=find_terms( rexp, markso=0 )
  if( debug ) Print(res)
  nt=res$nterms

  term=res$term[1]
  if( nt == 1 ){
   # a or a+  ->   a or +a  in rev order
   oper=res$oper[1]
  }
  else{
   # +a  ->  a+  in rev order
   oper=res$oper[2]
  }

  if( debug ) Print(rexp, nt, oper, term)

  # back to the original order
  term=revchar(term)
  term=chartr( "(", "}", term)
  term=chartr( ")", "(", term)
  term=chartr( "}", ")", term)

  res=paste( oper, term, sep="" )

  ft=cbind(lenexp-nchar(res)+1,lenexp)

  if( debug ) Print(res, ft)


  if( valid == 1 ){

   loc=gregexpr( "(", res, fixed=1 )[[1]]
   newf=ft[1,1]
   if( any( loc > 0 ) ){
    if( debug ) Print(loc)
    lenloc=length(loc)
    for( i in lenloc:1 ){
     bp=locbalpar( res, loc[i] )
     if( debug ) Print(lenloc,i,loc[i],bp)
     if( bp == 0 ){
      newf=loc[i]+1
      break
     }
    }
    ft[1,1]=ft[1,1]+newf-1
    if( ft[1,1] > nchar(rexp) ){
     ft=cbind(0,0)
     res=""
    }
    else{
     # Print(res,newf)
     res=substr(res,newf,nchar(res))
    }
   }
  } # end of valid

  if( debug ) Print( res, ft )
  if( debug ) Print("####### Bottom of first_term ######")

  return( list( term=res, fromto=ft ) )

 } # end of last_term
 ##################### first and last term #######################





 # fixed point iteration
 if( all == 1 ){
  res0=expr
  for( llll in 1:10 ){
   if( debug ) cat("\n\n####### Top of fixed point iteration ", llll,"...\n")
   res=Recall( res0, all=0, method=method, simplify=simplify
               , verify=verify, debug=debug )
   if( all(res0 == res) ) break
   res0=res
  }
  if( length(res) == 1 ) res=c(res)
  return( res )
 }




 ############# main starts here. #####################################

 # is numeric?
 if( all(is.numeric(expr)) ) return(expr)

 # is expr a scalar?
 if( is.vector(expr) ) lenexp=length(expr)
 else lenexp=nrow(expr)*ncol(expr)

 # if not scalar, use recursive call
 if( lenexp > 1  |  is.matrix(expr) ){
  expr[]=unlist(lapply( expr, Expand, debug=debug ))
  return( expr )
 }


 ############# main starts here. #####################################

 # compress expr
 expr=gsub( " ", "", expr )

 if( debug ){
   Print("\nTop of Expand\n\n after trimming\n",expr)
   Print(analyze_expr(expr)$tab)
 }

 # mark special operators
 expr=mark_soper( expr )

 # simplify first
 expr=Simplify( expr, markso=0 )         # 20200520
 # expr=remove_paren( expr, method=1 )


 if( debug ) Print("\nAfter Simplify",expr)

 if( 1 %in% method ){
  if( debug )  Print("\n\n***** Top of method=1:  a*(...)  or a/(...) ")

  # chage -(...) to -1*(...)
  expr=gsub("-(","-1*(", expr, fixed=1)


  # find additive terms in expr
  res=find_terms( expr, markso=0 )
  oper=res$oper; term=res$term
  nterms=res$nterms

  if( debug ) Print(nterms, oper, term)

  # for each additive term in expr
  for( i in 1:nterms ){

   ti=term[i]; oi=oper[i]

   if( substr(ti,1,1) == "(" ) ti=paste("1*",ti,sep="")
   term[i]=ti

   if( debug ){
     Print("\nTop of",i,ti,oi)
     Print(analyze_expr(ti)$tab)
   }

   # Process possible pre op (body) post


   # Must decompose each term into   oper pre OP body post
   # where
   # body is in a pair of matching parens,
   # OP is a multiplicative operator, that is, * or ` or /,
   # pre and post are a multiplicative factors to (body) or empty.
   # pre will be later decomposed into pre0 OP1 pre and
   # post will be later decomposed into OP2 and post0, where
   # OP1 and OP2 are + - ( ) ^ or $,
   # pre must not have ( and post must not have ).
   #
   # Example of a single term expression:
   # expr="-f( a*(b+c)*d, param )*e"
   #  oper  = -
   #  pre0  = f(
   #  pre   = a
   #  OP    = *
   #  body  = b+c
   #  post  = *d
   #  post0 = , param)
   #

   # find a*(...)  or actually, find *(  .
   # loc=regexpr( "[$&~*/`~]\\(", ti )[[1]]          #  20200519
   loc=regexpr( "[*/`~]\\(", ti )[[1]]          #  20200520
   if( loc > 0 ) loc2=locbalpar( ti, loc+1 ) # % NOT taken care of
   else loc2=0 # not used.

   if( debug ) Print("\ntop of i-th term @1", i, oi,ti,loc, loc2 )

   if( loc > 0 ){
    # The i-th term of expr contains a*(...)  or  a/(...)
    # loc is the location of [*/`~](   and   loc2 is th loc of balancing ).

    if( debug ) Print( "\na*(...)  or a/(...) found!!!!!!!!!!!!",ti,"\n\n")

    # decompose the i-th term ti into pre0, pre,op,body, post and post0
    pre=substr(ti,1,loc-1)

    # 202200819 -> 20200820 deleted
    ###    if( pre != "" && !is_valid(pre)$valid ) next

    body=substr(ti,loc+2,loc2-1)  # no enclosing parens
    op=substr(ti,loc,loc)
    op1=op; if( op == "/" ) op1="//"  # for Print only.
    post=""
    if( loc2 < nchar(ti) ) post=substr(ti,loc2+1,nchar(ti))

    if( debug ) Print("\nfirst decomp",pre,op1,body,post)


    # decompose pre into pre0 and pre: pre0 OP1 pre OP body post OP2 post0
    pre0=""
    res=find_terms( pre, markso=0 )
    nt=res$nterms
    if( nt > 1 ){
     pre0=paste( paste( res$oper[1:(nt-1)], res$term[1:(nt-1)], sep="" )
                 , collapse="" )
     pre=paste(res$oper[nt],res$term[nt], sep="" )
    }
    loc=gregexpr("(",pre,fixed=1)[[1]]
    if( any(loc > 0) ){                  # 20151004 modified
      # Print("***",expr,pre0)
     # move paren to pre0
     for( j in length(loc):1 ){
      if( locbalpar(pre,loc[j]) > 0 ) next
      loc=loc[length(j)]
      pre0=paste(pre0,substr(pre,1,loc),sep="")
      pre=substr(pre,loc+1,nchar(pre))
     }
    }

    if( debug ) Print("2nd pre",pre0,pre,"\n")


    # decompose post
    post0=""
    if( post != "" ){
     res=first_term(post, debug=debug)
     f0=res$fromto[1,2]
     if( f0 < nchar(post) ) post0=substr(post,f0+1,nchar(post))
     post=res$term
    }
    if( debug ) Print("2nd post",post,res$oper,res$term,"\n")


    if( debug )  Print("\nfinal 1", oi, pre0,pre,op,op1,body, post, post0)

    # decompose body into terms
    res=find_terms( body, markso=0 )
    operi=res$oper; termi=res$term
    operi[operi == ""]="+"
    ntermsi=res$nterms

    if( debug ) Print("\nbody consists of:",ntermsi,operi,termi)

    nfactorsi=0                   # 20200816
    if( ntermsi == 1 ){
     res=find_factors( body, markso=0 )
     nfactorsi=res$nfactors
    }

    if( debug ){
     cat("\n*********** FINAL ************\n")
     Print(expr)
     Print(oi, pre0, "/", pre, op1, body, post, "/", post0)
     Print("\nbody consists of:",ntermsi,nfactorsi,operi,termi)
     Print(analyze_expr(body)$tab)
    }


    new=pre0

    if( !( op == "/"  &  (ntermsi > 1 | nfactorsi > 1) ) ){    # 20200816

      if( debug ){
       Print('Here, op == "/" and nternsi > 1 and nfactors1 > 1')
       Print(pre0)
      }

     # for each term in body, distribute pre and op,
     # Skip the division by multiple terms in paren.
     for( j in 1:ntermsi ){

      oij=operi[j]; tij=termi[j]
      cond1=( op == "/"  &  nterms(tij) > 1 ) #????????????????????????????
      if( cond1 ) tij=paste( "(", tij, ")" ,sep="" )
      if( debug ){
        Print("\nTop of j", j,new)
        Print(oij,pre,op, tij, post)
      }

      if( pre0 == "" ){
        # expression does not contain function whose head is in pre0.
        if( oi %in% c("","+","!") ){
          # Here, oi (oper[i]) is positive
          new=paste( new, oij, pre, op, tij, post, sep="")
        }
        else if( pre0 == "" ){
         # Here, oi (oper[i]) is negative,
         # so chage the sign of oij which is associated with each term of body.
         if( oij == "-" ) oij="+"
         else if( oij == "+" ) oij="-"
         else if( oij == "#" ) oij="!"
         else if( oij == "!" ) oij="#"
         new=paste( new, oij, pre, op, tij, post, sep="")
         oper[i]="+"
        }
      }
      else if( pre0 != "" ){
        # expression contains function whose head is in pre0.
        # Leave oi (oper[i]) as it is.
        new=paste( new, oij, pre, op, tij, post, sep="")
      }

     } # end of j loop


    } # end of division many terms or factors
    else if(  op == "/"  &  ntermsi == 1 & nfactorsi == 2 ){    # 20200820

     # Here, possible a/(b/d)
     # a/(b/c) -> a*c/b
     if( debug ){
      Print('\nHere, op == "/" and nternsi = 1 and nfactors1 = 2')
      Print(pre0, new)
     }

     sign="([+-]*)"
     varnum="([[:digit:]]+|[[:alpha:]][[:alnum:]_]*)[^[]"  # 20200820
     #         #1   #2      /  #3
     pat=paste("^",sign,varnum,"/",varnum,sep="")
     if( regexpr( pat, body) > 0 ){
       # a/(b/c) found:  no paren
       newbody=sub(pat, "\\1\\3/\\2", body)
       op="*"; op1=op
       oij=substr(newbody,1,1)
       if( oij != "-" ) oij=""
       else{
        newbody=substr(newbody,2,nchar(newbody))
       }

       if( debug ){
         cat("\n** a/(b/c) found. **\n")
         Print(body, newbody, oij)
       }


       if( oi %in% c("","+","!") ){
         new=paste( new, oij, pre, op, newbody, post, sep="")
       }
       else{ ## 20200818 do not combine oi and oij
        if( pre0 == "" ){
         # Here, oi (oper[i]) is negative and pre0 is null.
         if( oij == "-" ) oij=""
         else if( oij != "-" ) oij="-"
         else if( oij == "#" ) oij="!"
         else if( oij == "!" ) oij="#"
         new=paste( new, oij, pre, op, newbody, post, sep="")
         oper[i]="+"
        }
        else{
         # Here, pre0 is not null
          new=paste( new, oij, pre, op, newbody, post, sep="")
        }
       }
       if( debug ) Print(oi,oper[i],oij,new)

     }
     else{
      # a/(b/c) not found: recover original expression
      if( debug ) Print(" a/(b/c) not found:", pre0,pre,op,body)
      new=paste(pre0,pre,op,"(",body,")",post,sep="")
     }

    } # end of division with one term denominator
    else{ # 20200819

    if( debug ) Print("\nHere, op is not '/'", pre0,pre,op,body)
     new=paste(pre0,pre,op,"(",body,")",post,sep="")

    } # end of all divisions

    new=paste(new,post0,sep="")
    new=gsub("+-","-",new, fixed=1)
    new=gsub("--","+",new, fixed=1)
    new=gsub("++","+",new, fixed=1)
    new=gsub("-+","-",new, fixed=1)
    new=gsub("!#","#",new, fixed=1)
    new=gsub("##","!",new, fixed=1)
    new=gsub("!!","!",new, fixed=1)
    new=gsub("#!","#",new, fixed=1)

    term[i]=new
    # oper[i]="+" ##################### <-------------- 20151004, 0819


    if( debug ) Print(new,"\n")

   } # a*(...) found

  } # end of i loop for terms

  expr=paste( paste( oper, term, sep="" ), collapse="" )

 } # end of method = 1





 if( 2 %in% method ){

  if( debug )  Print("\n\n***** Top of  (...)*a  or (...)/a \n\n")
  if( debug )  Print(expr)

  # compress expr
  expr=gsub( " ", "", expr )

  # chage -(...) to -1*(...)
  expr=gsub("-(","-1*(", expr, fixed=1)


  # find additive terms in expr
  res=find_terms( expr, markso=0 )
  oper=res$oper; term=res$term
  nterms=res$nterms

  if( debug ) Print(nterms, oper, term)

  # for each additive term in expr
  for( i in 1:nterms ){

   ti=term[i]; oi=oper[i]
   if( substr(ti,1,1) == "(" ) ti=paste("1*",ti,sep="")
   term[i]=ti

   # Process possible pre (body) op post  or  (...)/a

   loc=regexpr( "\\)[$&~*/`~]", ti )[[1]]              #  20200519
   loc=regexpr( "\\)[*/`~]", ti )[[1]]              #  20200530
   if( loc > 0 ) loc2=locbalpar( ti, loc )
   else loc2=0 # not used.

   if( loc2 > 1 ){
    # if finction, set loc=0  20150928
    if( regexpr("[[:alnum:]._]",substr(ti,loc2-1,loc2-1))[[1]] > 0 ) loc=0
   }

   if( debug ) Print("\n\ntop of i-th term @2", i, oi,ti,loc, loc2 )

   if( loc > 0 ){
    # The i-th term of expr contains (...)*a  or  (...)/a

    if( debug ) Print( "(...)*a  or (...)/a found!!!!!!!!!!!!!!!")
    # decompose the i-th term ti into pre,op,body and post
    post=substr(ti,loc+2,nchar(ti)) # % NOT taken care of
    body=substr(ti,loc2+1,loc-1)
    op=substr(ti,loc+1,loc+1) # % NOT taken care of
    op1=op; if( op == "/" ) op1="//"
    pre=""
    if( loc2 > 1 ) pre=substr(ti,1,loc2-1)

    if( debug ) Print("first decomp",pre,body,op1,post)

    # decompose pre into pre0 and pre
    pre0=""
    if( pre != "" ){
     # pre is the last oper + term of pre
     # pre0 is the rest of pre
     temp=last_term(pre, debug=debug)
     ff=temp$fromt[1,1]
     if( ff > 1) pre0=substr(pre,1,ff-1)
     else if( ff == 0 ) pre0=pre
     pre=temp$term
     # ( in pre
     if( !( substr(pre,nchar(pre),nchar(pre)) %in%
            c("*","/","$","&","~","`","~") ) ){       #  20200520
      # no *
      pre0=paste(pre0,pre,sep=""); pre=""
     }
    }
    presign="+"
    if( substr(pre,1,1) %in% c("+","-","!","#") ){
     presign=substr(pre,1,1)
     pre=substr(pre,2,nchar(pre))
    }
    if( debug ) Print("2nd pre",pre0,presign,pre)

    # decompose post
    post0=""
    nt=0
    if( post != "" ){
     res=first_term(post, debug=debug)
     f0=res$fromto[1,2]
     if( f0 < nchar(post) ) post0=substr(post,f0+1,nchar(post))
     post=res$term
    }

    if( debug ) Print("\nfinal", oi, pre0, presign, pre, body, op1, post
                      , post0)

    # decompose body into terms
    res=find_terms( body, markso=0 )
    operi=res$oper; termi=res$term
    operi[operi == ""]="+"
    ntermsi=res$nterms

    if( debug ) Print("\nbody",ntermsi,operi,termi)


    #new=paste(oi,pre0,sep="") ############################################
    new=NULL
    if( pre0 != "" ) new=paste(oi,pre0,sep="")
    else if( oi == "-" ){
     if( presign == "+" ) presign="-"
     else presign="+"
    }
    else if( oi == "#" ){
     if( presign == "!" ) presign="#"
     else presign="!"
    }
    if( substr(pre,nchar(pre),nchar(pre)) != "/" ){

     # for each term in body, distribute pre and op,
     # Skip the division by multiple terms in paren.
     for( j in 1:ntermsi ){

      oij=operi[j]; tij=termi[j]
      cond1=( op == "/"  &  nterms(tij) > 1 )             # not used!
      if( cond1 )  tij=paste( "(", tij, ")" ,sep="" )
      if(  presign == "-" ){
       if( oij == "-" ) oij="+"
       else oij="-"
      }
      if(  presign == "#" ){
       if( oij == "#" ) oij="!"
       else oij="#"
      }
      if( debug ) Print(j, new, oij,pre,tij,op, post)
      new=paste( new, oij, pre, tij, op, post, sep="")

     } # end of j loop

     new=paste(new,post0,sep="")

     new=gsub("+-","-",new, fixed=1)
     new=gsub("--","+",new, fixed=1)
     new=gsub("++","+",new, fixed=1)
     new=gsub("-+","-",new, fixed=1)
     new=gsub("!#","#",new, fixed=1)
     new=gsub("##","!",new, fixed=1)
     new=gsub("!!","!",new, fixed=1)
     new=gsub("#!","#",new, fixed=1)

     term[i]=new; oper[i]=""       ############### <----------- ??????

     # if( debug ) Print("**concat\n", i,oper[i],substr(term[i],1,1))

    } # not division

    if( debug ) Print("new\n",new,"\n")

   } # (...)*a found


  } # end of i loop for terms

  expr=paste( paste( oper, term, sep="" ), collapse="" )

 } # end of method = 2



 # final
 expr=gsub("+-","-",expr, fixed=1)
 expr=gsub("--","+",expr, fixed=1)
 expr=gsub("++","+",expr, fixed=1)
 expr=gsub("-+","-",expr, fixed=1)
 expr=gsub("!#","#",expr, fixed=1)
 expr=gsub("##","!",expr, fixed=1)
 expr=gsub("!!","!",expr, fixed=1)
 expr=gsub("#!","#",expr, fixed=1)

 # remove paren
 if( substr(expr,1,1) == "("  &&
     locbalpar(expr,1) == nchar(expr) ) expr=substr(expr,2,nchar(expr)-1)


 # verify
 if( verify ){
  valid=is_valid( expr )
  if( !valid[[1]] ){
     Print(valid)
  }
 }

 if( simplify > 0 )
  for( i in 1:simplify) {
   expr=Simplify(expr, markso=0)
  }


 # unmark special operators
 expr=unmark_soper( expr )


 if( debug ) Print(expr)

 if( length(expr) == 1 ) expr=c(expr)

 return( expr )


} # end of Expand


