#' Matrix Inversion
#'
#' Matrix Inversion by Sweep or Gaussian Elimination
#'
#' @param A A square matrix
#' @param method "Sweep" or "GE"
#' @param loc vector of pivot locations
#'
#' @param simplify = 0 not to use Simplify nor remove_paren \cr
#' = 1 to use Simplify
#'
#' @param eps small value to check if the pivot is 0.
#' @param nopermute =1 to avoid permuting rows/cols when inverting A.
#' @param chkginv = 1 to check if the resulting g-inverse is genuin.
#' @param history = 1 to return the permutation matrix D
#'
#' @param debug = 1 to print intermediate result.
#'
#' @details
#' If \code{A} is a numeric matrix, \code{matSwp} function
#' in lazy.mat package will be used with the following parameters: \cr
#'  \code{A, loc, eps, nopermute, chkginv, history, debug}.
#'
#' If \code{A} is a semi symbolic character matrix, \code{matSweep} function
#' or \code{invGE} function in lazy.symbolic package will be used
#' with the following parameters \cr
#'  \code{A, loc, simplify, debug}.
#'
#' If \code{A} is a semi symbolic character matrix,
#' if \code{method="Sweep"}, \code{matSweep} will be used. (Sweep Operator)
#' if \code{method="Ge"}, \code{invGE} will be used. (Gaussian Elimination)
#'
#'
#' @examples
#' # numeric A
#' An=demomat(2,2)
#' Inv(An)
#'
#' # character A
#' A=demomat(2,2, root="a")
#' Inv(A)
#' 
#' # # by Gaussian Elimination
#' # Inv(A, method="GE" )
#'
#' # character A as a partitioned matrix
#' Ap=demomat(2,2, root="A", fullsymb=1)
#' Inv(Ap)
#'
#' @export
#'
Inv <- function( A, loc=1:nrow(A), method="Sweep", ... ){
 # Inversion of a matrix
 # Shin-ichi Mayekawa
 # 20200629
 # method: 20250302
 # method and loc switched: 20250315
 # no invGE exa,ple: 20250316
 # ...: 20250316
 #

 if( is.numeric(A) ){
  invA=matSwp( A=A, loc=loc, ... )
 }
 else{
  if( method == "Sweep" ||  !is.null(attributes(A)$fullsymb) )
   invA=matSweep( A=A, loc=loc, ... )
  else
   invA=invGE( A=A, loc=loc, ... )[,-c(1:ncol(A))]
 }

 return( invA )

} # end of Inv



