#' Symbolic Expression of the i-j element of the Kronecker Product
#'
#' @param A A matrix
#' @param B B matrix
#' @param row row index
#' @param col col index
#' @param eval = 1 to evaluate the result
#' @param ntogoback  n parameter of sys.parent function
#' @param print = 1 to print the result
#'
#' @details
#' The i-j element of A@@B is \cr
#' \code{
#' A[(i-(i-1)\%\%nrow(B)+nrow(B)-1)/nrow(B)
#' ,(j-(j-1)\%\%ncol(B)+ncol(B)-1)/ncol(B)]
#' *
#' B[(i-1)\%\%nrow(B)+1,(j-1)\%\%ncol(B)+1]
#' }
#' \cr
#' \cr
#' When eval=1, the  A and B matrices must be defined.
#' Also, the numeric value of indices i and j must be specified.
#'
#' @examples
#' \dontrun{
#' KP(A,B)
#' KP(D,E, row="p", col="q")
#'
#' # with evaluation
#' A <- demomat(2,2,root="a"); B <- demomat(3,3,root="b")
#' i <- 3; j <-4
#' KP(A,B,eval=1,print=1)
#' }
#'
#' @return
#' When eval=0, the symbolic expression will be returned. \cr
#' When eval=1, a list of ABij, value, i, and j will be returned.
#'
#' @export
#'

KP <- function( A, B,  row="i", col="j", eval=0, ntogoback=1, print=0 ){
 # The i-j element of the Kronecker product of A and B matrices
 # Shin-ichi Mayekawa
 # 20151002
 #

 Amat=as.character(substitute(A))
 Bmat=as.character(substitute(B))
 nrB=paste("nrow(",Bmat,")",sep="")
 ncB=paste("ncol(",Bmat,")",sep="")
 subai=paste( "(", row, "-(", row, "-1)%%", nrB, "+", nrB,"-1)/", nrB, sep="" )
 subaj=paste( "(", col, "-(", col, "-1)%%", ncB, "+", ncB,"-1)/", ncB, sep="" )
 subbi=paste( "(",row,"-1)%%",nrB,"+1", sep="" )
 subbj=paste( "(",col,"-1)%%",ncB,"+1", sep="" )

 ABij=paste(Amat,"[",subai,",",subaj,"] * ",Bmat,"[",subbi,",",subbj,"]",sep="")

 res=ABij

 if( print ){
  cat("\n\nThe ", row, "-", col, " element of ", Amat, "@", Bmat, " is\n\n"
      , ABij, "\n")
 }

 if( eval == 1 ){
  eA=exists( Amat, where=sys.frame(sys.parent(n=ntogoback)) )
  eB=exists( Bmat, where=sys.frame(sys.parent(n=ntogoback)) )
  ei=exists( row, where=sys.frame(sys.parent(n=ntogoback)) )
  ej=exists( col, where=sys.frame(sys.parent(n=ntogoback)) )
  if( eA*eB*ei*ej ){
   evOK=1
   Amat1=get(Amat,pos=parent.frame(n=ntogoback))
   Bmat1=get(Bmat,pos=parent.frame(n=ntogoback))
   row1=get(row,pos=parent.frame(n=ntogoback))
   col1=get(col,pos=parent.frame(n=ntogoback))
   if( row1 > nrow(Amat1)*nrow(Bmat1) |  col1 > ncol(Amat1)*ncol(Bmat1) ){
    cat("\nerror1:(KP) The row/col index out of range.\n")
    return( res )
   }
   # Print(Amat1,Bmat1,row1,col1)
   # Print(i,j,nrow(A),ncol(A))
   # subai1=(i-(i-1)%%nrow(A)+nrow(A)-1)/nrow(A)
   # subaj1=(j-(j-1)%%ncol(A)+ncol(A)-1)/ncol(A)
   # subbi1=(i-1)%%nrow(A)+1
   # subbj1=(j-1)%%ncol(A)+1
   # Print(subai1,subaj1,subbi1,subbj1)
   ev=eval( parse(text=ABij), envir=parent.frame(n=ntogoback) )
   if( print ){
    cat("\nwhich evaluates to ", ev, ",\n")
    cat("\nwith ", row, " = ", row1, ", ", col, " = ", col1, ", and\n\n")
    eval(parse(text=paste(Amat,"=Amat1",sep="")))
    eval(parse(text=paste(Bmat,"=Bmat1",sep="")))
    eval(parse(text=paste("Print(",Amat,",",Bmat,")",sep="")))
    cat("\nThe Kronecker Product", Amat, "@", Bmat,  "is\n\n")
    print(matKTimes(Amat1,Bmat1))
   }
   res=list(ABij=ABij, value=ev, i=row1, j=col1)
  }
  else{
   evOK=0
   cat("\nerror1:(KP) Some of the indices/matrices are not defined.\n")
  }
 }

 return( res )

} # end of KP




