#' Sum of Two or More Symbolic Expression
#'
#' @param ... Symbolic matrices or vectors to be added.
#' @param simplify = 0 not to use Simplify nor remove_paren \cr
#' = 1 to use Simplify
#' @param debug = 1 to print intermediate result
#'
#' @details
#' Tow matrices of the same dimensions can be added or subtraced. \cr
#' A scalar can be added to or subtracted from a matrix \cr
#' \cr
#' Numeric matrices will be treated as character ones.\cr\cr
#' When you want to chage the sign of all the elements of a matrix, use \cr
#' A1 <- Minus( 0, A )   or   A1 <- matTimes( -1, A )
#'
#'
#' If the arguments are  super matrices with SMdim attributes,
#' the final result will have an appropriate SMdim attribute.
#'
#' @return A character matrix
#'
#' @examples
#' A <- demomat(3,3,root="a")
#' A[2,2] <- paste("-",A[2,2],sep="")
#' B <- demomat(3,3,root="b")
#' B[3,3] <- 0
#' Plus( A, B )
#' Plus( A, 1 )
#'
#' @export
#'

Plus <- function( ..., simplify=1, debug=0 ){
 # symbolic matrix addition
 # Shin-ichi Mayekawa
 # 20150725
 # names: 20150803
 # Simplify: 20150806
 # bugfix: 20150806
 # matPlus renamed as Plus: 20150824
 # ...: 20150904
 # simplify=1: 20150906
 # SMdim: 20200525
 # fullsymb: 20200608
 #
 #  matrix + matrix
 #  vector + matrix R
 #  scalar + matrix
 #
 #  vector + vector
 #  scalar + vector
 #
 #  scalar + scalar
 #

 # analyse ...
 name=name <- as.character(substitute(list(...)))[-1L]
 val=list(...)
 locnull=which(unlist(lapply(val,is.null)))
 if( length(locnull) > 0 ){
  name=name[-locnull]
  val=val[-locnull]
 }
 narg=length(val)

 if( debug ) Print(narg,name,val)

 if( narg >= 3 ){
  # recursive call to Plus
  C=Recall( val[[1]], val[[2]], simplify=simplify, debug=debug )
  for( i in 3:narg ){
   C=Recall( C, val[[i]], simplify=simplify, debug=debug )
  }
  return( C )
 }


 ######################################################################
 # Simple case starts here.
 A=val[[1]]; B=val[[2]]

 # SMdim attribute
 SMdim=attributes(A)$SMdim
 if( debug ) Print( SMdim1,SMdim2,SMdim )

 # check if A or B are partitioned matrix
 fsA=attributes(A)$fullsymb
 fsB=attributes(B)$fullsymb


 # size of matrix
 nr1=nrow(A); nc1=ncol(A)
 nr2=nrow(B); nc2=ncol(B)
 len1=0; len2=0
 ismatA=1; ismatB=1

 # First, make non matrix input a scalar
 if( is.null(nr1) ){
  # A is a vector
  ismatA=0
  nr1=1; nc1=1
  len1=length(A)
 }
 if( is.null(nr2) ){
  # B is a vector
  ismatB=0
  nr2=1; nc2=1
  len2=length(B)
 }


 # Both A and B are vectors of the same size -> just add
 if( is.vector(A)*is.vector(B)*(length(A)>1)*(length(B)>1)*
     (length(A)==length(B))  ){
  A=matrix(A,,1); B=matrix(B,,1)
 }
 else if( nr1*nr2*nc1*nc2 == 1 ){
  # Both A and B are scalar
  A=matrix(A,1,1); B=matrix(B,1,1)
 }
 else if( nr1*nc1 == 1 )
  A=matrix(A,nr2,nc2)
 else if( nr2*nc2 == 1 )
  B=matrix(B,nr1,nc1)

 nr1=nrow(A); nc1=ncol(A)
 nr2=nrow(B); nc2=ncol(B)


 rn=rownames(A); cn=colnames(A)
 if(is.null(rn) & !is.null(rownames(B)) ) rn=rownames(B)
 if(is.null(cn) & !is.null(colnames(B)) ) cn=colnames(B)

 if( debug ) Print(nr1,nc1,nr2,nc2, len1,len2, ismatA, ismatB, rn, cn)


 if( !( nr1 == nr2  &  nc1 == nc2 ) ){
  cat("\n error1:(Plus) Two matrices do not conform. \n")
  return()
 }
 C=matrix("",nr1,nc2)
 for( i in 1:nr1 ){
  for( j in 1:nc2 ){

   a=A[i,j]; if( a == "0" ) a=""
   b=B[i,j]; if( b == "0" ) b=""
   if( substr(b,1,1) == "-" ){
    if( b == "" )
     C[i,j]=a
    else
     C[i,j]=paste( a, "-", substr(b,2,nchar(b)), sep=" " )
   }
   else{
    if( b == "" )
     C[i,j]=a
    else
     C[i,j]=paste( a, "+", b, sep=" " )
   }
   if( C[i,j] == "" ) C[i,j]=0

  } # j
 } # i

 # names
 dimnames(C)=list(rn,cn)

 # final treatment
 if( simplify >=1  ){
  C=Simplify(C)
  if( simplify >= 2){
   C=Simplify(C)
  }
 }

 if( !is.null(SMdim$nrows) ) attributes(C)$SMdim=SMdim
 if( !is.null(fsA) |  !is.null(fsB) ) attributes(C)$fullsymb=1

 return( C )

} # end of Plus



#' Difference of Two or More Symbolic Expressions
#'
#' @param ... Symbolic matrices or vectors to be subtracted.
#' @param simplify = 0 not to use Simplify nor remove_paren \cr
#' = 1 to use Simplify
#' @param debug = 1 to print intermediate result
#'
#' @details
#' Tow matrices of the same dimensions can be added or subtraced. \cr
#' A scalar can be added to or subtracted from a matrix \cr
#' \cr
#' Numeric matrices will be treated as character ones. \cr\cr
#' When you want to chage the sign of all the elements of a matrix, use \cr
#' A1 <- Minus( 0, A )   or   A1 <- Times( -1, A )
#' \cr
#' Minus( A, B, C, ...) -> A-B-C-...
#'
#'
#' If the arguments are  super matrices with SMdim attributes,
#' the final result will have an appropriate SMdim attribute.
#'

#' @return A character matrix
#'
#' @examples
#' Minus( 3,2,1 )
#' A <- demomat(3,3,root="a")
#' A[2,2] <- paste("-",A[2,2],sep="")
#' B <- demomat(3,3,root="b")
#' B[3,3] <- 0
#' Minus( A, B )
#' Minus( A, 1 )
#'
#' @export
#'

Minus <- function( ..., simplify=1, debug=0 ){
 # symbolic matrix addition
 # Shin-ichi Mayekawa
 # 20150725
 # names: 20150803
 # Simplify: 20150806
 # bugfix: 20150806
 # matMinus renamed as Minus: 20150824
 # ...: 20150904
 # simplify=1: 20150906
 # SMdim: 20200525
 # fullsymb: 20200608
 #
 #  matrix + matrix
 #  vector + matrix R
 #  scalar + matrix
 #
 #  vector + vector
 #  scalar + vector
 #
 #  scalar + scalar
 #

 # analyse ...
 name=name <- as.character(substitute(list(...)))[-1L]
 val=list(...)
 locnull=which(unlist(lapply(val,is.null)))
 if( length(locnull) > 0 ){
  name=name[-locnull]
  val=val[-locnull]
 }
 narg=length(val)

 if( debug ) Print(narg,name,val)

 if( narg >= 3 ){
  # recursive call to Minus
  C=Recall( val[[1]], val[[2]], simplify=simplify, debug=debug )
  for( i in 3:narg ){
   C=Recall( C, val[[i]], simplify=simplify, debug=debug )
  }
  return( C )
 }


 ######################################################################
 # Simple case starts here.
 A=val[[1]]; B=val[[2]]

 # SMdim attribute
 SMdim=attributes(A)$SMdim
 if( debug ) Print( SMdim1,SMdim2,SMdim )

 # check if A or B are partitioned matrix
 fsA=attributes(A)$fullsymb
 fsB=attributes(B)$fullsymb


 # size of matrix
 nr1=nrow(A); nc1=ncol(A)
 nr2=nrow(B); nc2=ncol(B)
 len1=0; len2=0
 ismatA=1; ismatB=1

 # First, make non matrix input a scalar
 if( is.null(nr1) ){
  # A is a vector
  ismatA=0
  nr1=1; nc1=1
  len1=length(A)
 }
 if( is.null(nr2) ){
  # B is a vector
  ismatB=0
  nr2=1; nc2=1
  len2=length(B)
 }


 # Both A and B are vectors of the same size -> just add
 if( is.vector(A)*is.vector(B)*(length(A)>1)*(length(B)>1)*
     (length(A)==length(B))  ){
  A=matrix(A,,1); B=matrix(B,,1)
 }
 else if( nr1*nr2*nc1*nc2 == 1 ){
  # Both A and B are scalar
  A=matrix(A,1,1); B=matrix(B,1,1)
 }
 else if( nr1*nc1 == 1 )
  A=matrix(A,nr2,nc2)
 else if( nr2*nc2 == 1 )
  B=matrix(B,nr1,nc1)

 nr1=nrow(A); nc1=ncol(A)
 nr2=nrow(B); nc2=ncol(B)

 rn=rownames(A); cn=colnames(A)
 if(is.null(rn) & !is.null(rownames(B)) ) rn=rownames(B)
 if(is.null(cn) & !is.null(colnames(B)) ) cn=colnames(B)

 if( debug ) Print(nr1,nc1,nr2,nc2, len1,len2, ismatA, ismatB, rn, cn)


 if( !( nr1 == nr2  &  nc1 == nc2 ) ){
  cat("\n error1:(Munus) Two matrices do not conform. \n")
  return()
 }
 C=matrix("",nr1,nc2)
 for( i in 1:nr1 ){
  for( j in 1:nc2 ){

   a=A[i,j]; if( a == "0" ) a=""
   b=B[i,j]; if( b == "0" ) b=""

   if( regexpr("[+-]", b) == -1 ) single=1
   else single=0

   if( a == b ) C[i,j]=""
   else if( b == "" )  C[i,j]=a
   else if( single == 0 ){
    C[i,j]=paste( a, "-(", b, ")", sep=" " )
   }
   else{
    C[i,j]=paste( a, "-", b, sep=" " )
   }
   if( C[i,j] == "" ) C[i,j]=0

  } # j
 } # i

 # names
 dimnames(C)=list(rn,cn)

 # final treatment
 if( simplify >=1  ){
  C=Simplify(C)
  if( simplify >= 2){
   C=Simplify(C)
  }
 }

 if( !is.null(SMdim$nrows) ) attributes(C)$SMdim=SMdim
 if( !is.null(fsA) |  !is.null(fsB) ) attributes(C)$fullsymb=1

 return( C )

} # end of Minus



