#' Simplify Symbolic Expressions
#' (very primitive version)
#'
#' @param expr Expression (vector or matrix)
#' @param method How to simplify: default method is 1:3. \cr
#' = 1 : varname*num -> num*varname \cr
#' = 2 : processing 1s and 0s, a/a -> 1, etc
#' w/o using \code{find_terms} \cr
#' = 3 : a-a or -a+a -> 0  or  0*func(x) -> 0 \cr
#' = 4 : to print the result of analyze_expr( expr )
#' @param func_level Level of function argument simplification \cr
#' = 0 not to simplify function arguments \cr
#' = 1 to simplify the top level function arguments \cr
#' = 2 to simplify the 2nd level function arguments
#' (the arguments of the functions in functions.)
#' @param expandfuncarg = 0 not to expand the function argument. \cr
#' = 1 to expand function argument with function name \cr
#' = 2 to expand function argument (body)
#' @param markso = 0 not to mark the special opeartors
#' @param all = 1 to apply Simplify repeatedly.
#' @param verify = 0 not to check the result by \code{is_valid}
#' @param debug = 1 to print details
#'
#' @details
#' This is a very primitive version. \cr
#' This function applies some mathematical rules such as \cr
#' \code{0*a -> 0} or \code{a*1 -> a} described as regular expressions
#' to an input expression. \cr\cr
#' In method=1 and 2, \code{gsub} function is used to replace
#' the patterns such as \code{0*a} or \code{-((a))} with their simplified
#' mathematically equivalent forms. \cr
#' In method=3, input expression is decomposed into terms and the patterns
#' such as \code{func(a)*0} or \code{X\%*\%inv(X)} will be simplified.
#' \cr
#' Can handle fully symbolic matrix expressions.
#' \cr
#' \cr
#' Repeated use of this function may simplify the result further. \cr
#' Simplify may work better if you \code{Expand} the expression
#' prior to submission. \cr
#' \cr
#' When called from remove_paren, use method=1:3. \cr
#'
#' @return Simplified expression
#'
#' @examples
#' expr="-((a))-b/b+0*c+d*0+1*e-e*2+f-3*f+0/(g+h) - (x-y)*(((z+1)))"
#' Simplify( expr, all=1 ) |> print() |> Expand(all=1)
#'
#' ( res <- matTimes( demomat(2,2,root="a"), demomat(2,1), simplify=0 ) )
#' Simplify(res)
#'
#' Simplify(c("1*b","1*(b)","1*f(b)"))
#' Simplify(c("0*b","0*(b)","0*f(b)"))
#'
#' dd <- Det(demomat(2,2,root="a"))
#' Simplify(Expand(dd))
#'
#' expr1=c("-(X+A)%*%inv(X+A)%*%(X+A)%*%inv(X+A)"
#'    , "-(X+A1)%*%inv(X+A)%*%(X+A)%*%inv(X+A)")
#' Simplify(expr1)
#' Simplify(expr1, all=1)
#'
#' ( res <- Simplify( "a+b+2*a+2*b-3*a+3*b" ) )
#' Simplify(res)
#' # Sorry.
#'
#' # semi symbolic matrix
#' A=demomat(2,2,root="a")
#' AinvA=matTimes( A, Inv(A) )
#' Simplify( AinvA, all=1 )
#' Simplify( Expand(AinvA, all=1), all=1 )
#' # Sorry, cannot simplify \code{AinvA[2,1]} and \code{AinvA[2,2]}.
#'
#' # fully symbolic matrix
#' A=demomat(2,2,root="A", fullsymb=1)
#' A1=matSweep(A,1)
#' A11=matSweep(A1,1)
#' A11s=Simplify(A11)
#' Expand(A11s)
#'
#'
#' @export
#'

Simplify <- function( expr, all=0, method=1:3, func_level=0, expandfuncarg=1
          , markso=1, verify=1, debug=0 ){
 # simplify expression
 # Shin-ichi Mayekawa
 # 20150803,04
 # alpha*num: 20150804
 # 1.0 -> 1: 20150806
 # bugfix: 20150814,15
 # new version: 20150816granda,17
 # / added to v3 and v31: 20150823Maiko
 # bugfix for alpha*num: 20150904
 # remove_paren: 20150905,06
 # recurisive call with a-a -> 0: 20150906
 # a/a -> 1: 20150906
 # (+....) -> (...) from remove_paren: 20150906
 # method 3 w/o and with remove_paren: 20150907
 # ( added: 20150912
 # find_factors: 20150915
 # markso option: 20150922
 # (Use Simplify9.R if this version does not work.)
 # 0*0* -> 0: 20150924
 # bugfix: 20151006
 # [ added 20151006
 # all: 20151006
 # bugfix for (0): 20151007
 # bugfix for %*% and %@%; ` and ~ added: 20200520,28
 # a+a etc: 202006115
 # debug comment: 20200615
 # regular expr simplified: 20200616
 # mark_soper used: 20200616
 # renew all the regular expressions: 22200616,17,18
 # method 3.9: 20200626
 # Identity: 20200626
 # inverse: 20200626,27
 # method renumbered: 20200627
 # remove paren: 20200627
 # bugfix for 3.9: 20200628
 # bugfix for 3.5: 20200629
 # bugfix for 3.9: 20200702,03
 # bugfix locbalpar: 20200715
 # inv(inv(A))=A: 20200715
 # bugfix for inv(X)%*%X: 20200716
 # bugfix for method=2 of "(B)-0" for HS: 20230428
 # numeric input: 20230503
 # trim at the beginning: 20230513
 # inv(inv()) corrected: 20230513
 # t(t()) added: 20230513
 # exp -> expr: 20230513
 # inv(I) etc: 20230623,24
 # func_level: 20230623,24
 # lv() removed from examples for version 4.2: 20230625
 # bugfix for t(t(-x)): 20230627
 # find_farfunc$func -> is.function: 20230628
 # noexpand with matInv: 20240731
 # noexpand bugfix: 20240803
 # noexpand --> expandfuncarg: 20240804
 # .Last.value --> pipe: 20250314
 #

 # a number
 # num0 <- "[0-9]*\\.{0,1}[0-9]*" <- wrong?
 # a variable name \\2 contains a variable name
 # var0 <- "(^|[-/+*^=`~]| )([[:alpha:]][[:alnum:]_]*)($|[-/+*^=`~]| )*"
 # mathematical operator
 # mathop <- "[-/+*^=`~]"


 # is numeric?
 if( all(is.numeric(expr)) ) return(expr)

 # is expr a scalar?
 if( is.vector(expr) ) lenexp=length(expr)
 else lenexp=nrow(expr)*ncol(expr)

 # treatment of inv(I) ant t(I): 20230624
 expr=gsub( "inv(I)", "I", expr, fixed=1 )
 expr=gsub( "inv(inv(I))", "I", expr, fixed=1 )
 expr=gsub( "t(I)", "I", expr, fixed=1 )
 expr=gsub( "t(t(I))", "I", expr, fixed=1 )
 expr=gsub( "t(0)", "0", expr, fixed=1 )
 expr=gsub( "t(t(0))", "0", expr, fixed=1 )

 # expr=Expand(expr,simplify=0)



 ############# Fixed Point Iteration  (recursive call)   ####################

 # fixed point iteration
 if( all == 1 ){
 res0=expr
 for( llll in 1:10 ){
  if( debug ) cat("\n\n####### Top of fixed point iteration ", llll,"...\n")
  res=Recall( res0, all=0, method=method, markso=markso
        , func_level=func_level, debug=debug )
  if( all(res0 == res) ) break
  res0=res
 }
 if( length(res) == 1 ) res=c(res)
 return( res )
 }




 ############# for each element of c(exp)   (recursive call)  ################

 # if not scalar, use recursive call
 if( lenexp > 1 | is.matrix(expr) ){
 expr[]=unlist(lapply( expr, Simplify, method=method, markso=markso
           , func_level=func_level, debug=debug ))
 return( expr )
 }




 ############# for each function found in expr ## 20230623 #################
 #############   (recursive call)  #########################################
 ############# Must be improved!!!            ##############################
 if( func_level > 0 ){

  if( debug ) Print("######## Top of Function Argument Evaluation", expr)

  # does the expression contain functions?
  ff=find_varfunc( expr )
  locfunc=which(ff$is.function==1)
  lenlocf=length(locfunc)
  # if not scalar, use recursive call
  if( lenlocf > 0 ){

   from=1

   nfunc=0
   repeat{

    nfunc=nfunc+1
    locfunc=locfunc[from]
    flist=ff$varlist[locfunc]
    ft=ff$fromto[locfunc,,drop=0]
    bp=ff$bp[locfunc,,drop=0]
    if( debug ) Print("######## top of repeat", nfunc, flist,ft,bp)
    top=substr(expr,1,bp[1,1])
    body=substr(expr,bp[1,1]+1,bp[1,2]-1)

    # This was added on 20240803
    if( expandfuncarg >= 2 ) body=Expand(body)

    tail=substring(expr,bp[1,2])
    # Print(top,body,tail)
    if( regexpr("[-/+*^`~]", body) > 0 ){
    if( debug ) Print("######## Start simplifying a function argument.")
     body=Simplify( body, method=method, markso=markso
                  , debug=debug, func_level=0, all=all )
    }
    if( debug ) Print("######## End of one function argument.", nfunc, body)

    # treatment of inv(I): 20230624
    expr=gsub( "inv(I)", "I", body, fixed=1 )
    expr=gsub( "inv(inv(I))", "I", body, fixed=1 )
    expr=gsub( "t(I)", "I", body, fixed=1 )
    expr=gsub( "t(t(I))", "I", body, fixed=1 )
    expr=gsub( "t(0)", "0", expr, fixed=1 )
    expr=gsub( "t(t(0))", "0", expr, fixed=1 )

    expr[]=paste(top,body,tail, sep="")

    # This was added on 20240731
    if( expandfuncarg >= 1 ) expr=Expand(expr,simplify=0)
    if( debug ) Print("######## New expression .", expr)


    ff=find_varfunc( expr )
    locfunc=which(ff$is.function==1)
    lenlocf=length(locfunc)
    if( debug ) Print("########", nfunc,lenlocf)
    if( nfunc > 50 || lenlocf <= nfunc ) break
    locfunc=locfunc[nfunc+1]

   } # end of repeat

   if( debug ) Print("****** end of repeat", expr)

   return( expr )

  } # end of one func found

 } # end of function argument





 ##########################################################################
 ############# core part starts here. #####################################
 ##########################################################################

 # old parameter
 collect=0



 if( debug ) Print("\nTop of Simplify", expr)

 # compress all blanks
 expr=gsub( " +", "", expr )
 if( debug ) Print("After compress", expr)


 # remove balanced parens at the top and the tail
 if( substr(expr,1,1) == "(" &&
   locbalpar(expr,1) == nchar(expr) ) expr=substr(expr,2,nchar(expr)-1)


 # mark special operators
 if( markso == 1 ){
  expr=mark_soper( expr )
  if( debug ) Print("After mark_soper", expr)
 }



 if( 1 %in% method ){

  if( debug ){
   cat("\nStarting Method=1\n")
 }


  # repeat
  expr0=expr
  for( i in 1:10 ){


   if( debug ) cat("\nmethod 1, loop ",i,"\n")


   # (+....) -> (...)
   expr=gsub( "\\(\\+", "\\(", expr )
   if( debug ) cat("1.1; (+....) -> (...)",expr, "\n", sep=" :: ")

   # ((x)) -> x and (x) -> x
   expr=gsub(
   "(^|[-+*/`~])\\(\\(([0-9]*\\.{0,1}[0-9]*|[[:alpha:].][[:alnum:]_]*)\\)\\)"
   ,"\\1\\2", expr )
   expr=gsub(
   "(^|[-+*/`~])\\(([0-9]*\\.{0,1}[0-9]*|[[:alpha:].][[:alnum:]_]*)\\)"
   ,"\\1\\2", expr )
   if( debug ) cat("1.2; (x) -> x",expr, "\n", sep=" :: ")


   # alpha*num -> num*alpha
   #  1    2           3           4
   #  (top | *)(variable     ) * (number       )($ or */ )
   v3=
   "(^|[-+*`~])([[:alpha:]][[:alnum:]_]*)\\*([0-9]*\\.{0,1}[0-9]*)($|[-+*/`~])"
   expr=gsub( v3, "\\1\\3\\*\\2\\4", expr )
   if( debug ) cat("1.3; alpha*num -> num*alpha",expr, "\n", sep=" :: ")

   # alpha*(1/num) -> (1/num)*alpha
   # 1    2           3                4
   # (top | *)(variable     ) * ((1/number)           )($ or */ )
   v3=
"(^|[-+*`~])([[:alpha:]][[:alnum:]_]*)\\*(\\(1\\/[0-9]*\\.{0,1}[0-9]*\\))($|[-+*/`~])"
   expr=gsub( v3, "\\1\\3\\*\\2\\4", expr )
   if( debug ) cat("1.4; a*(1/num) -> (1/num)*a",expr, "\n", sep=" :: ")

   # alpha/num -> (1/num)*alpha
   #  1    2           3           4
   #  (top | *)(variable     ) * (number       )($ or */ )
  v2="(^|[-+*`~])([[:alpha:]][[:alnum:]_]*)\\/([0-9]*\\.{0,1}[0-9]*)($|[-+*/`~])"
   expr=gsub( v2, "\\1(1/\\3)\\*\\2\\4", expr )
   if( debug ) cat("1.5; a/num -> (1/num)*a ",expr, "\n", sep=" :: ")

   if(regexpr("inv(I)", expr, fixed=1) > 0 ) Print("++++++++++++++++++++++")

   # inv(I): 20230623
   expr=gsub( "inv(I)", "I", expr, fixed=1 )
   expr=gsub( "inv(inv(I))", "I", expr, fixed=1 )
   if( debug ) cat("1.6; inv(I) -> I ",expr, "\n", sep=" :: ")

   # check if converged
   if( all(expr==expr0) ) break
   expr0=expr


  } # end of inner fixed point loop


 } # end of method 1




 if( 2 %in% method ){

 if( debug ){
  cat("\nStarting Method=2\n")
 }


  # 1.0 -> 1
  expr=gsub( "(^|[^[:alnum:]])1\\.0*($|[^[:alnum:]])", "\\11\\2", expr  )
  if( debug ) cat("2.1; 1.0 -> 1",expr, "\n", sep=" :: ")

  # 0.0 -> 0
  expr=gsub( "(^|[^[:alnum:]])(0*\\.*0+|0+\\.*0*)($|[^[:alnum:]])"
       , "\\10\\3", expr  )
  if( debug ) cat("2.2; 0.0 -> 0",expr, "\n", sep=" :: ")


  # until the fixed point is reached.?
  expr0=expr
  for( i in 1:10){

   if( debug ) cat("\nmethod 2, loop ",i,"\n")

   # 0*0 -> 0
   expr=gsub("0([`~*]0)+","0",expr)
   if( debug ) cat("2.2.2; 0*0*0... -> 0",expr, "\n", sep=" :: ")

   # -0 as a term
   # -0 ->
   expr=gsub( "(^|[-+*`~]|[\\)\\(])-0($|[-+]|[\\)])"
        , "\\1\\2", expr  )
   if( debug ) cat("2.2.5; -0 -> 0",expr, "\n", sep=" :: ")

   # comment out: 20230428
   #  # -0 -> 0
   #  expr=gsub( "(^|[-+*`~]|[\\)\\(])-0($|[-+*`~/]|[\\)])"
   #        , "\\10\\2", expr  )
   #  if( debug ) cat("2.3; -0 -> 0",expr, "\n", sep=" :: ")


   # 0*0*0 -> 0
   expr=gsub("0([`~*]0)+","0",expr)
   if( debug ) cat("2.4; 0*0*0... -> 0",expr, "\n", sep=" :: ")

   # 0*x -> 0
   expr=gsub(
  "((^|[-/+*^=`~]|\\()0+)[/*`~]([[:alpha:]][[:alnum:]_]*|[[:digit:]]+)($|[/+*^=`~-]|\\))"
        , "\\20\\4", expr )
   if( debug ) cat("2.5; 0*x, 0/x -> 0", expr, "\n", sep=" :: ")

   # x*0 -> 0
   expr=gsub(
  "(^|[-/+*^=`~]|\\()([[:alpha:]][[:alnum:]_]*|[[:digit:]]+)[*`~]0+($|[-)/+*^=`~]| )"
       , "\\10\\3", expr )
   if( debug ) cat("2.6; x*0 -> 0", expr, "\n", sep=" :: ")

   # a+0 or a-0 -> a
   expr=gsub( "(\\.+)[-+]0($| )", "\\1", expr )
   if( debug ) cat("2.7; a+0",expr, "\n", sep=" :: ")

   # +0+ etc -> NULL
   expr=gsub( "(^|[-+])0($|[-+\\)])", "\\2", expr )
   if( debug ) cat("2.8; +0+ etc -> NULL",expr, "\n", sep=" :: ")


   # 1*x -> x
   expr=gsub(
   "(^|[-/+*=`~])1\\*([[:alpha:]][[:alnum:]_]*|[[:digit:]]+)($|[-/+*=`\\(\\[])"
       , "\\1\\2\\3", expr )
   if( debug ) cat("2.9; 1*x, 1*f(x) -> x",expr, "\n", sep=" :: ")

   # x*1 -> x
   expr=gsub(
  "(^|[-/+*=`~])([[:alpha:]][[:alnum:]_]*|[[:digit:]]+|.+\\))\\*1($|[-/+*=`~])"
       , "\\1\\2\\3", expr )
   if( debug ) cat("2.10; x*1, f(x)*1 -> x",expr, "\n", sep=" :: ")

   # (...)*1 -> (...)
   # expr=gsub( "(.+\\))\\*1($|[-/+*=`~])", "\\1\\2", expr )
   # if( debug ) cat("2.11.1; (...)*1 -> (...)",expr, "\n", sep=" :: ")


   # I*x -> x
   expr=gsub(
   "(^|[-/+*=`~])I`([[:alpha:]][[:alnum:]_]*|[[:digit:]]+)($|[-/+*=`\\(\\[])"
       , "\\1\\2\\3", expr )
   if( debug ) cat("2.12; I*x, I*f(x) -> x",expr, "\n", sep=" :: ")

   # x*I -> x
   expr=gsub(
   "(^|[-/+*=`~])([[:alpha:]][[:alnum:]_]*|[[:digit:]]+|.+\\))`I($|[-/+*=`~])"
       , "\\1\\2\\3", expr )
   if( debug ) cat("2.13; x*I, f(x)*I -> x",expr, "\n", sep=" :: ")




   # Below uses a back reference to \\2.
   # x/x -> 1
   expr=gsub(
    "(^|[-/+*=`~])([[:alpha:]][[:alnum:]_]*|[[:digit:]]+)/\\2($|[-/+*=`~])"
       , "\\11\\3", expr )
   if( debug ) cat("2.14; x/x -> 1",expr, "\n", sep=" :: ")

   # f(x)/f(x)  may not be correct???
   expr=gsub(
    "(^|[-/+*=`~])([[:alpha:]][[:alnum:]_]*\\(.+\\))/\\2($|[-/+*=`~])"
    , "\\11\\3", expr )
   if( debug ) cat("2.15; f(x)/f(x) -> 1",expr, "\n", sep=" :: ")

   # (...) /(...)
   expr=gsub( "(^|[-/+*=`~])(\\(.+\\))/\\2($|[-/+*=`~]| )"
        , "\\11\\3", expr )
   if( debug ) cat("2.16; (...) /(...)",expr, "\n", sep=" :: ")


   # inv(A)%*%A --> I
   expr=gsub(
"(^|[-/+*=`~])inv\\(([[:alpha:]][[:alnum:]_]*|[[:digit:]]+)\\)`\\2($|[-/+*=`~])"
       , "\\1I\\3", expr )
   if( debug ) cat("2.17; x/x -> 1",expr, "\n", sep=" :: ")


   # A%*%inv(A) --> I
   expr=gsub(
"(^|[-/+*=`~])([[:alpha:]][[:alnum:]_]*|[[:digit:]]+)`inv\\(\\2\\)($|[-/+*=`~])"
       , "\\1I\\3", expr )
   if( debug ) cat("2.18; x/x -> 1",expr, "\n", sep=" :: ")


   # inv(inv(x))
   loc=regexpr( "(^|[^[:alnum:]._])inv\\(inv\\(", expr, fixed=0 )
   if( loc > 0 ){
    loc=loc+regexpr("inv(inv(", substring(expr,loc), fixed=1)-1
    top=""
    if( loc > 1 ) top=substr(expr,1,loc-1)
    end1=locbalpar( expr, loc+3 )
    end2=locbalpar( expr, loc+7 )
    if( end1 == end2+1){
     tail=""
     if( end1 < nchar(expr) ) tail=substr( expr,end1+1,nchar(expr) )
     inside=substr(expr,loc+8,end2-1)
     if( nterms(inside) > 1 ) inside=paste("(",inside,")",sep="")
     expr=paste(top,inside,tail,sep="")
     expr=gsub( "+-", "-", expr, fixed=1 )
     expr=gsub( "--", "+", expr, fixed=1 )
     expr=gsub( "-+", "-", expr, fixed=1 )
     expr=gsub( "++", "+", expr, fixed=1 )
     if( debug ) cat("2.19; inv(inv(x)) -> x",expr, "\n", sep=" :: ")
    }
   }

   # t(t(x))  bugfix: 20230627
   loc=regexpr( "(^|[^[:alnum:]._])t\\(t\\([^-]", expr, fixed=0 )
   if( loc > 0 ){
    loc=loc+regexpr("t(t(", substring(expr,loc), fixed=1)-1
    top=""
    if( loc > 1 ) top=substr(expr,1,loc-1)
    end1=locbalpar( expr, loc+1 )
    end2=locbalpar( expr, loc+3 )
    if( end1 == end2+1){
     tail=""
     if( end1 < nchar(expr) ) tail=substr( expr,end1+1,nchar(expr) )
     inside=substr(expr,loc+4,end2-1)
     if( nterms(inside) > 1 ) inside=paste("(",inside,")",sep="")
     expr=paste(top,inside,tail,sep="")
     expr=gsub( "+-", "-", expr, fixed=1 )
     expr=gsub( "--", "+", expr, fixed=1 )
     expr=gsub( "-+", "-", expr, fixed=1 )
     expr=gsub( "++", "+", expr, fixed=1 )
     if( debug ) cat("2.20; t(t(x)) -> x",expr, "\n", sep=" :: ")
    }
   }


   # leftovers
   expr=gsub( "^\\+","", expr )


   # check if converged
   if( all(expr==expr0) ) break
   expr0=expr

  } # end of inner fixed point loop


  # final
  if( any(expr == "") ) expr[expr == ""]=0

  if( debug ) Print(" end of method=2", expr)

 } # end of method 2




 if( 3 %in% method ){

  # a-a or -a+a -> 0
  # for each term 20150905
  # This part requires that expr is a scalar
  res=find_terms( expr, markso=0 )
  oper=res$oper; term=res$term
  nterms=res$nterms
  if( debug ){
   cat("\nStarting Method=3\n")
   Print(nterms,oper,term)
  }

  # Skip this part: 20230624
  if(0){

  # check if a term has enclosing matching parens. 20230624
  premoved=0
  for( i in 1:nterms ){
   ti=term[i]
   if( locbalpar(ti,1) == nchar(ti) && oper[i] %in% c("+","") ){
   premoved=1
   terms[i]=substr(ti,2,nchar(ti)-1)
   if( oper[i] == "+" && substr(ti,2,2) == "+" ){
    term[i]=substr(ti,3,nchar(ti)-1)
   }
   else if( oper[i] == "+" && substr(ti,2,2) == "-" ){
    term[i]=substr(ti,3,nchar(ti)-1)
    oper[i]="-"
   }
   }
  }
  if( premoved != 0 ){
   expr=paste( c(t(cbind(oper,term))), collapse="" )
   res=find_terms( expr, markso=0 )
   oper=res$oper; term=res$term
   nterms=res$nterms
   if( debug ){
    cat("\nStarting Method=3:  Enclosing Parens Removed.\n")
    Print(nterms,oper,term)
   }
  }

 # term == "*0"    20150924  # ` added 20200520
 for( i in 1:nterms){
   if( term[i] == "*0" | term[i] == "`0" | term[i] == "~0" ){
    if( debug ) Print("3.1; *0 -> NULL",i,oper[i],term[i])
    term[i]=""; oper[i]=""
   }
   if( term[i] == "+0" | term[i] == "-0" ){
    if( debug ) Print("3.2; +0 -> NULL",i,oper[i],term[i])
    term[i]=""; oper[i]=""
   }
 }

 } # end of if(0)


  # For each term
  for( i in 1:nterms ){

   if( term[i] == "" ) next

   # decompose the i-th term into factors
   res=find_factors(term[i], markso=0)
   factor=res$factor
   foper=res$oper
   nf=res$nfactors
   if( debug ) Print(i,oper[i], foper,factor)

   ######################################################
   #  Must handle (0)*a etc here!!!!!!!!!! 20151006

   # 0*func(a): "`" added: 20230623, bugfix for div-by-zero
   # for each factor of the i-th term containing 0.
   if( debug ) Print("**** 0*func ****",nf,term[i],foper,factor)
   zero=0
   if( any( factor == "0" ) ){
    loc=which( factor == "0" )
    if( debug ) Print(loc)
    for( j in 1:length(loc) ){
     if( nf > 1 & foper[loc[j]+1] %in% c("%*%","*","/","`") ) zero=1
     else
     if( foper[loc[j]] %in% c("%*%","*","`") | foper[loc[j]] != "/" ) zero=1
    }
   }
   if( zero == 1 ){
    if( debug ) Print( "3.3; 0*func found: ",oper[i], term[i] )
    term[i]=""; oper[i]=""
    next
   } # end of 0*func(a)


   # 1*func(a): 20230623`
   # for each factor of the i-th term containing 0.
   if( debug ) Print("**** 1*func ****",nf,term[i],foper,factor)
   one=0
   if( any( factor %in% c("1","I") ) ){
   loc=which( factor %in% c("1","I") )
   if( debug ) Print(loc)
    for( j in 1:length(loc) ){
     if( nf > 1 & foper[loc[j]+1] %in% c("%*%","*","`") ) one=1
    else
     if( foper[loc[j]] %in% c("%*%","*","`") ) one=2 # ?????
    }
   }
   if( one == 1 ){
    if( debug ) Print( "3.3.1; 1*func found: ", one, oper[i], term[i] )
    foper[loc[j]+1]=""
    factor=factor[-loc[j]]; foper=foper[-loc[j]]
    temp=c(t(cbind(foper,factor)))
    temp=paste(temp,collapse="")
    term[i]=temp
    next
   } # end of 1*func(a)


   # a/a ?????????????????????????????????????
   match=0
   if( any( foper == "/" ) & length(unique(factor)) < nf ){
   loc=which( foper == "/" ) # This cannot be 1 or nf.
   if( debug ) Print(loc)
   for( j in 1:length(loc) ){

    den=factor[loc[j]]
    loc1=setdiff( which(factor == den), loc[j] )
    if( debug ) Print(j, loc, loc1)
    if( length(loc1) > 0 ){
     for( k in 1:length(loc1) ){
      if( foper[loc1[k]] != "/" ){
       match=1
       if( debug ) Print( "3.4; a.../a found: ",foper[loc1[k]]
                  ,factor[loc1[k]], den )
       factor[loc1[k]]="1"; foper[loc[j]]=""; factor[loc[j]]=""
       break
      }
     } # end of k
     if( debug ) Print(foper,factor)
     term[i]=paste( paste( foper, factor, sep="" ), collapse="" )
    }
   } # end of j
  } # a/a


   # inv(x)
   resreg=unlist( regexpr("inv\\(.+", factor) )
   # Print(resreg)
   if( any( resreg > 0 ) ){
   loc=which( resreg > 0 )
   if( debug ) Print("inv found", loc, resreg[loc])
   for( j in 1:length(loc) ){
    locj=loc[j]
    fj=factor[locj]

   if( locbalpar(fj,resreg[locj]+3) == nchar(fj) ){  # 20200716

   # 20200716
   if( fj != "" )
     inside=substr(fj,resreg[locj]+4,locbalpar(fj,resreg[locj]+3)-1)
   else inside=""
   if( debug ) Print(j,locj,inside)
    if( locj < nf ){
     # inv(x)%*%x
     if( foper[locj+1] == "`" ){
      fj1=factor[locj+1]
      # Print(foper[locj+1], fj1, paste("(",inside,")",sep=""))
      if( ( inside == fj1 || paste("(",inside,")",sep="") == fj1 ) ){
       if( debug ) Print( "3.5; inv(x)%*%x: ", inside, fj1 )
       factor[locj]="I"
       inside=""
       factor[locj+1]=""; foper[locj+1]=""
      }
     }
    }
    if( locj > 1 ){
     # x %*% inv(x)
     if( foper[locj] == "`" ){
      fj1=factor[locj-1]
      # if( debug )
      # Print(foper[locj], fj1, paste("(",inside,")",sep=""))
      if( ( inside == fj1 || paste("(",inside,")",sep="") == fj1 ) ){
       if( debug ) Print( "3.6; inv(x)%*%x: ", inside, fj1 )
        factor[locj]="I"; foper[locj]=""
        factor[locj-1]=""
      }
     }
    }

   } # locbaplar

  } # end of j

  # Below was moved in from outside of end of inv(x)  20230634
  # update term i with new factors
  term[i]=paste(foper,factor,collapse="", sep="")

 } # inv(x) found



 } # end of i

 # expr=paste( mapply( paste, oper, term, sep="" ), collapse="" )


 if( nterms >= 2 ){

  if( debug ){
  cat("For each pair of terms...\n")
  }

  # for each pair of terms
  for( i in 1:(nterms-1) ){    # range of i and j were switched. 20200615
   for( j in (i+1):nterms ){

   if( term[i] == term[j] ){

   # Here, term[i] = term[j]
   if( (oper[i] %in% c("","+") & oper[j] == "-") |
     (oper[i] == "-" & oper[j] %in% c("","+") ) ){
    # a-a or -a+a found
    if( debug )
    Print( "3.7; a-a: ",oper[i], term[i], " and ", oper[j], term[j] )
    term[i]=""; oper[i]=""; term[j]=""; oper[j]=""
    break
   }
   else if( (oper[i] %in% c("","+") & oper[j] == "+") |
         (oper[i] == "-" & oper[j] == "-" ) ){
    # a+a or -a-a found
    if( debug)
    Print( "3.8; a+a or -a-a: ",oper[i],term[i]," and ", oper[j], term[j] )
    term[i]=paste("2*", term[i], sep="")
    term[j]=""; oper[j]=""
    break
   }

   } # end of term[i] == term[j]
   else{

   # Here, term[i] != term[j]   # 20200626

   if( term[i] == "" || term[j] == "" ) next

   # decompose both terms into factors:  redundant!!
   #
   # term i: cannot move this ouside of j-loop!!
   res=find_factors(term[i], markso=0)
   factori=res$factor
   # if factor i's first element is not numeric, do nothing.
   if( regexpr("[^[:digit:]]",factori[1]) > 0 ) next
   foperi=res$oper
   if( length(factori) == 1 ){
    factori=c(1,factori)
    foperi=c(foperi,"*")
   }
   nfi=length(factori)
   if( factori[1] == "" ) factori[1]="1"

   # term j
   res=find_factors(term[j], markso=0)
   factorj=res$factor
   foperj=res$oper
   if( length(factorj) == 1 ){
    factorj=c(1,factorj)
    foperj=c(foperj,"*")
   }
   nfj=length(factorj)
   if( factorj[1] == "" ) factorj[1]="1"

   if( debug ){
     Print(i, oper[i], foperi,factori)
     Print(j, oper[j], foperj,factorj)
   }

   # num1*a + num2*a:  num has been moved to the left of alpha by 1.3
   if( nfi == nfj && all(factori[2:nfi] == factorj[2:nfj]) &&
     all(foperi[2:nfi] == foperj[2:nfj]) &&
     foperi[2] == "*" && foperj[2] == "*" ){

    if( debug){
     Print( "3.9; num1*a + num2*a: ", "/"
         ,oper[i], term[i], " and ", oper[j], term[j] )
    }

    # attach oper[i]=="-" to the factor so that it can be take care of
    # when evaluating collected numeric factors.
    if( oper[i] == "-" ){
    numfact=paste("(",oper[i], factori[1], oper[j], factorj[1],")",sep="")
    oper[i]="+"
    }
    else{
    numfact=paste("(", factori[1], oper[j], factorj[1],")",sep="")
    }

    if( debug ){
     Print(numfact,foperi[2:nfi], factori[2:nfi])
    }

    # evaluate numeric factor
    if( regexpr("[^[:digit:]()+-]", numfact) <= 0 ){
     code=paste("numfact=", numfact, sep="")
     eval(parse(text=code))
     numfact=format(numfact)
     operi=substr(numfact,1,1)
     if( operi %in% c("+","-") ){
      oper[i]=operi
      numfact=substr(numfact,2,nchar(numfact))
     }
     else{
     oper[i]="+"
     }
    }
    else if( collect == 0 ) next

    term[i]=paste (numfact, paste( foperi[2:nfi], factori[2:nfi]
           , sep="", collapse=""), sep="" )

    # term j has been absorbed in term i, i < j
    term[j]=""; oper[j]=""

   } # end of processing the terms containing num*varmane

   } # end of term[i] != term[j]

  } # end of j loop
  } # end of i loop

  # expr=paste( mapply( paste, oper, term, sep="" ), collapse="" )
 } # end of pair of terms


  expr=paste( mapply( paste, oper, term, sep="" ), collapse="" )


 if( debug ){
   Print(" end of method=3", oper, term)
   Print(" end of method=3", expr)
 }


 } # end of method 3




 if( 4 %in% method ){
  if( debug ){
   cat("\nStarting Method=4\n")
  }

  # resa=analyze_expr( expr, print=1 )

  if( debug ) Print(" end of method=4", expr)

 } # end of method 4



 # unmask special operators
 if( markso == 1 ) expr=unmark_soper(expr)


 # final
 if( any(expr == "") ) expr[expr == ""]=0

 expr=trim( expr, all=2 )

 if( length(expr) == 1 ) expr=c(expr)

 # verify
 if( verify ){
  valid=is_valid( expr )
  if( !valid[[1]] ){
   Print(valid$ok)
  }
 }


 # return
 return( expr )


} # end of Simplify


