#' Elementwise Multiplication of Two or More Symbolic  Vectors/Matrices
#'
#' @param ...  Symbolic matrices or vectors to be multiplied elementwise.
#' @param simplify = 0 not to use Simplify nor remove_paren \cr
#' = 1 to use Simplify
#' @param debug = 1 to print intermediate result
#'
#' @details
#' Let v be a vector of length lv, and M, a matrix of the size nr x nc. \cr
#' If lv <= nr*nc, v*M and M*v exist. \cr
#' When (nr*nc \%\% lv) == 0, there will be no warning but
#' otherwise, you will be warnded. \cr
#' If lv > nr*nc, v*M nor M*v do not exits. It causes error. \cr
#' \cr
#' When exits, v*M = M*v = matrix( c(M)*v, nrow(M), ncol(M) ).
#' \cr\cr
#' v\%*\%M and M\%*\%v exist if v can be promoted to either a row or column
#'  vector to make the two arguments conformable. \cr
#' \cr
#' vec1*vec2 always exists and
#' length(vec1*vec2)=max(length(v1),length(v2)) .
#'
#'
#' If the arguments are  super matrices with SMdim attributes,
#' the final result will have an appropriate SMdim attribute.
#'
#' @return A character matrix
#'
#' @examples
#' Times("a",2)
#' Times("a+b",2)
#' Times("c","a+b")
#'
#' M32 <- demomat(3,2, root="a")
#' v2 <- 1:2; v3 <- 1:3; v4 <- 1:4; v9 <- 1:9
#' vm2 <- matrix(v2,2)
#' vm3 <- matrix(v3,3)
#'
#' # matrix times vector with conformable length
#' # This seems to be what we want.
#' Times(v3,M32)
#' Times(M32,v3)
#'
#' # matrix times vector with conformable length
#' # This may not be what we want.
#' Times(v2,M32)
#' Times(M32,v2)
#'
#' # %*% operator promotes vectors to matrices.
#' Print(v3%*%demomat(3,2), demomat(3,2)%*%v2)
#'
#'
#' # When langth(vec) < length(c(mat)),
#' # with warning
#' Times(v4,M32)
#' Times(M32,v4)
#'
#' # When langth(vec) > length(c(mat)),
#' # ERROR
#' # Times(v9,M32)
#' # Times(M32,v9)
#'
#' @export
#'

Times <- function( ..., simplify=1, debug=0 ){
 # symbolic matrix elementwise multiplication
 # Shin-ichi Mayekawa
 # 20150817,22,23
 # times renamed as times2: 20150823
 # mattimes renamed as Times: 20150823
 # nterms: 20150824
 # ...: 20150904
 # simplify=1: 20150906
 # SMdim: 20200525
 # fullsymb: 20200608
 #

 # trim
 # A=trim(A,all=1)
 # B=trim(B,all=1)

 #
 #
 # This function requires time2 function which is defined in matTimes.
 #
 #

 # analyse ...
 name=name <- as.character(substitute(list(...)))[-1L]
 val=list(...)
 locnull=which(unlist(lapply(val,is.null)))
 if( length(locnull) > 0 ){
  name=name[-locnull]
  val=val[-locnull]
 }
 narg=length(val)

 if( debug ) Print(narg,name,val)

 if( narg >= 3 ){
  # recursive call to Times
  C=Recall( val[[1]], val[[2]], simplify=simplify, debug=debug )
  for( i in 3:narg ){
   C=Recall( C, val[[i]], simplify=simplify, debug=debug )
  }
  return( C )
 }


 ######################################################################
 # Simple case starts here.
 A=val[[1]]; B=val[[2]]

 # SMdim attribute
 SMdim=attributes(A)$SMdim
 if( debug ) Print( SMdim1,SMdim2,SMdim )

 # check if A or B are partitioned matrix
 fsA=attributes(A)$fullsymb
 fsB=attributes(B)$fullsymb


 # What are they?
 isvecA=-1
 if( is.vector(A) ) isvecA=1
 else if( is.matrix(A) ) isvecA=0
 isvecB=-1
 if( is.vector(B) ) isvecB=1
 else if( is.matrix(B) ) isvecB=0

 if( isvecA == -1  |  isvecB == -1 ){
  cat("\n error1:(Times) The arguments must be either vector or matrix. \n")
  return()
 }

 nam=NULL

 if( isvecA+isvecB == 2 ){
  # both vector
  #       C=A*B
  len=max(length(A),length(B))
  AA=A; BB=B
  if( length(A) < length(B) ){
   AA=c(matrix(A,length(A)*length(B)))[1:length(B)]
  }
  else if( length(A) > length(B) ){
   BB=c(matrix(B,length(A)*length(B)))[1:length(A)]
  }
  C=character(len)
  for( i in 1:len ){
   C[i]=times2( AA[i], BB[i], prefix="" )
  }
  names(C)=names(A)
 }
 else if( isvecA+isvecB == 0 ){
  # both matrix
  if( all( dim(A) == dim(B) ) ){
   C=matrix("", nrow(A),ncol(A))
   for( i in 1:nrow(A) )
    for( j in 1:ncol(A) ){
     C[i,j]=times2( A[i,j], B[i,j], prefix="" )
    }
   if( !is.null(nam<-dimnames(A)) ) nam=nam
   else if( !is.null(nam<-dimnames(B)) ) nam=nam
   if( !is.null(nam) ) dimnames(C)=nam
  }
  else{
   cat("\n error1:(Times) Matrices do not conform.\n")
   return()
  }
 }
 else{
  # vec * mat  or   mat * vec
  if( isvecA == 1 ){
   # vec A * mat B
   nrB=nrow(B); ncB=ncol(B); lenMat=nrB*ncB
   nrC=nrB
   if( !is.null(nam<-dimnames(B)) ) nam=nam
   if( length(A) <= lenMat ){
    #       C=matrix(A*c(B),nrow(B))
    if( length(A) < lenMat )
     A=c(matrix(A,length(A)*lenMat))[1:lenMat]
    B=c(B)
   }
   else{
    cat("\n error1:(Times) Cannot multiply vector A and matrix B.\n")
    cat(" length(A) is too large.\n")
    return()
   }
  }
  else{
   # mat A * vec B
   nrA=nrow(A); ncA=ncol(A); lenMat=nrA*ncA
   nrC=nrA
   if( !is.null(nam<-dimnames(A)) ) nam=nam
   if( length(B) <= lenMat ){
    #       C=matrix(A*c(A),nrow(A))
    if( length(B) < lenMat )
     B=c(matrix(B,length(B)*lenMat))[1:lenMat]
    A=c(A)
   }
   else{
    cat("\n error1:(Times) Cannot multiply vector A and matrix B.\n")
    cat(" length(B) is too large.\n")
    return()
   }
  }

  # multiply
  C=character(lenMat)
  for( i in 1:lenMat )
   C[i]=times2( A[i], B[i], prefix="" )
  C=matrix(C,nrC)
  if( !is.null(nam) ) dimnames(C)=nam

 }

 # names


 # final treatment
 if( simplify >=1  ){
  C=Simplify(C)
  if( simplify >= 2){
   C=Simplify(C)
  }
 }

 if( !is.null(SMdim$nrows) ) attributes(C)$SMdim=SMdim
 if( !is.null(fsA) |  !is.null(fsB) ) attributes(C)$fullsymb=1

 return( C )

} # end of Times


