#' Decompose an Expression into Additive Terms and Multiplicative Factors
#'
#' @param expr An input expression
#' @param sort = 1 to sort the result by variable name
#' @param print = 0 not to print the result
#' @param debug = 1 to print intermediate info
#'
#' @details
#'  The result of the analysis is stored in a character matrix named \code{tab}
#'  .  \cr
#'  If the input expression is \cr
#'   \code{"-inv(A)+(B+C)+D\%*\%E+F\%*\%(G-(H)-(I)+J)"} \cr
#'  it will be decomposed into the terms and factors in the matrix shown below:
#'   \preformatted{
#'        topers   o1  factor_1    o2     factor_2
#'   -----------------------------------------------------
#'   term_1    -       inv(A)
#'   term_2    +       (B+C)
#'   term_3    +       D           \%*\%    E
#'   term_4    +       F           \%*\%    (G-(H)-(I)+J)
#' }
#'
#'
#' \cr
#'   Each row of the tab matrix corresponds to a term in \code{expr} and
#'   each term is decomposed into factors and operators. \cr
#'
#'   \code{topers} and \code{o1} are the operators to be attached to factor1.\cr
#'   (\code{o1} may be an empty string vector as shown in the example above.)\cr
#'   \code{o2} is the operator to be attached to \code{factor2}, etc. \cr
#'
#'   \code{topers} and \code{o1} are additive operators, + or -
#'   ,  and \code{op2, op3,...}, are multiplicative operators, *, \%*\%, etc.
#'
#'  From this, the original expression can be recovered as:
#'  \preformatted{
#'   paste( apply(tab,1, function(x){paste(x,collapse="")} ), collapse="")
#'   }
#'
#'
#' @return
#'  A list of tab, valid and identical, where \cr
#'  tab is the table shown above, \cr
#'  valid is the flag to see if the expression is valid, \cr
#'  identical is the flag to see if the input and the output are identical.
#'
#' @examples
#' analyze_expr( "-inv(A)+(B+C)+D%*%E+F%*%(G-(H)-(I)+J)" )
#' analyze_expr( "-inv(A)+(B+C)+D%*%E+F%*%(G-(H)-(I)+J)", sort=1 )
#'
#' # recovery
#' expr <- "-inv(A)+2*(B-C)+Psi(X)"
#' tab <- analyze_expr( expr )$tab
#' expr2 <- paste(apply(tab,1, function(x){paste(x,collapse="")}), collapse="")
#' identical(expr,expr2)
#'
#' # scalar expressions
#' set.seed(1702)
#'
#' nexpr=20
#' for( i in 1:nexpr ){
#'
#'  # generate expression
#'  natom=10
#'  expr=gen_random_expr( natom=natom, nparen=4, oper=c("*","+","-") )
#'  if( runif(1) < 0.2 ) expr=sub( "a", "F(a)", expr )
#'  if( runif(1) < 0.3 ) expr=sub( "d", "F1(d)", expr )
#'  if( runif(1) < 0.2 ) expr=sub( letters[natom], "Ff2(x)", expr )
#'
#'  # analyze expression
#'  res=analyze_expr( expr, print=1 )
#'
#' }
#'
#'
#' # matrix expressions
#' set.seed(1702)
#'
#' nexpr=20
#' for( i in 1:nexpr ){
#'
#'   # generate expression
#'   natom=10
#'   expr=gen_random_expr( natom=natom, nparen=4
#'                         , alpha=LETTERS[1:natom], oper=c("%*%","+","-") )
#'   if( runif(1) < 0.3 ) expr=sub( "A", "inv(A)", expr )
#'   if( runif(1) < 0.3 ) expr=sub( "C", "diag(C)", expr )
#'   if( runif(1) < 0.3 ) expr=sub( LETTERS[natom], "ginv(X)", expr )
#'
#'   # analyze expression
#'   res=analyze_expr( expr, print=1 )
#'
#' }
#'
#' @export
#'

analyze_expr <- function( expr, sort=0, print=0, debug=0 ){
 # decompose expression into additive terms and their multiplicative factors
 # Shin-ichi Mayekawa
 # 20200529
 # check if expr is a vector: 20200629
 # roxygen2: 20220412dnc
 # check validity before and after: 20230428
 # sort: 20230504cot
 # roxygen2: 20230513,20250313
 # bugfix due to new get_valLL: 20250313
 #

 # Args:
 #  exp   input valid expression
 #
 # Values:
 #  A list of tab, valid and identical, where
 #   tab is a matrix containing the result of the analysis shown below,
 #   identical = 1 if the input expression can be recovered from tab.
 #
 #  tab is a matrix of term x factor:
 #  If the input expression is
 #   "-inv(A)+(B+C)+D%*%E+F%*%(G-(H)-(I)+J)"
 #  it will be decomposed into the terms and factoers in the matrix show below:
 #
 # \preformatted{
 #        topers   o1 factor_1    o2     factor_2
 #   -----------------------------------------------------
 #   term_1    -      inv(A)
 #   term_2    +      (B+C)
 #   term_3    +      D           %*%    E
 #   term_4    +      F           %*%    (G-(H)-(I)+J)
 # }
 #
 #   Each row of the tab matrix corresponds to a term in expr and
 #   each term is decomposed into factors and operators.
 #
 #   topers and o1 are the operators to be attached to factor1,
 #   o2 is the operator to be attached to factor2, etc.
 #
 #   o1 is an additive operator (+ or -) and op2, op3,..., are
 #   multiplicative operators (*, %*%, etc).
 #
 #  From this, the original expression can be recovered as:
 #   paste( apply(tab,1, function(x){paste(x,collapse="")} ), collapse="")
 #

 if( print ) cat( "\ninput expression =", expr, "\n" )

 if( is.matrix(expr) || length(expr) > 1 ){
   cat("\n\nerror1:(analyze_expr) Cannot handle vector input.\n\n")
   return(NULL)
 }

 # check if valid
 res=is_valid( expr )
 if( !res$valid ){
   cat("\nerror1(analyse_expr): input expresion is not valid.\n")
   print(res)
   return( NULL )
 }

 # decompoosese the expression into termss
 res_terms=find_terms( expr, sort=sort )

 if( debug ){
  printm(res_terms, .lj.=1)
 }

 topers=res_terms$oper
 terms=res_terms$term

 topers[is.na(topers)]=""

 if( debug ) printm(topers,terms, .lj.=1)

 # decompose each term into factors: result is a list of list
 factors=lapply( terms, find_factors, sort=sort )

 if( debug ){
  cat("str(factor)\n", str(factors))
  printm( factors )
 }

 # collect factors
 tf_table=get_valLL(factors,"factor", method=2)
 if( is.vector(tf_table) ) tf_table=matrix(tf_table,,1)
 tf_table[is.na(tf_table)]=""
 rownames(tf_table)=paste("term_",1:nrow(tf_table),sep="")
 colnames(tf_table)=paste("factor_",1:ncol(tf_table),sep="")

 # collect factor-operators
 op_table=get_valLL(factors,"oper", method=2)
 if( is.vector(op_table) ) op_table=matrix(op_table,,1)
 op_table[is.na(op_table)]=""
 rownames(op_table)=paste("term_",1:nrow(op_table),sep="")
 colnames(op_table)=paste("o",1:ncol(op_table),sep="")

 if( debug ){
  printm(tf_table, .lj.=1)
  printm(op_table, .lj.=1)
 }

 # for each term, interleave factor-operators and factors
 tab=if( !all(op_table=="") ) interleave( op_table, tf_table)
 else tab=tf_table

 if( debug ){
  printm(tab)
  # printm(interleave( op_table[,-1,drop=FALSE], tf_table))
 }

 # attach term-operators
 tab=cbind(topers,tab)


 # checking the result by recovering the original expression from tab
 res=paste( apply(tab,1, function(x){paste(x,collapse="")} ), collapse="")
 if( identical(expr,res) ) identical=1 else identical=0

 # check if valid
 res2=is_valid( res )
 valid=res2$valid
 if( !valid ){
   cat("\nerror1(analyse_expr): Big Trouble!! Result is not valid.\n")
   print(res2)
 }

 if( debug ){
  printm(expr, .lj.=1)
  printm(res, .lj.=1)
 }

 if( print ){
  # cat("Structure of Expression:\n")
  print(tab, quote=0)
  if( !identical ) printm(identical,valid, res)
 }

 return( list(tab=tab, identical=identical, valid=valid) )

} # end of analyze_expr


