#' Find a term separated by a multiplicative operator
#'
#' @param expr A string containing an expression
#' @param pattern A pattern to be used to identify a variable \cr
#' The default pattern is \code{"([[:digit:]]+|[[:alpha:]][[:alnum:]_]*)"}.
#' @param complete = 0 to find the first variable, not the first term
#'
#' @details
#' This function finds the first term of expr.
#' The term is defined as the product (including division) of variables.
#' A variable is defined as the string consisting of valid root characters
#' trailed by subscripts in [ and ], if any.
#' Those characters in a balanced parens can be regarded as a variable
#' as well as a function reference such as f(args).
#' Therefore, \code{expr} such as \code{A[a,b]} or \code{f(args)}
#'  are treated equally.
#'
#' When \code{complete=0}, the function returns the first variable.
#' WHen \code{complete=1}, the function returns the first term.
#'
#' This function searches for a variable in \code{expr} from left to right.
#' If a variable is found and it is trailed by a multiplicative operator,
#' the search continues to the right again.
#'
#' @return
#' If complete=1, a list of \cr
#' \preformatted{
#'   term  The term found w/o trailing and preceeding signs.
#'   loc   The starting location of the term in expr.
#'   len   The length of the term
#'   pre   The preceeding string to the left of the term, if any.
#'   post  The trailing string to the right of the term, if any.
#' }
#'
#' \code{term} will be a valid expression but
#' depending on the input, \code{pre} and \code{post} may not be
#' valid expressions.
#'
#' From \code{expr} the term can be obtained as \cr
#' \preformatted{
#' res <- find_a_term( expr )
#' term <- substr(expr,res$loc,res$loc+res$len-1)
#' }
#' See the example below.
#'
#' @examples
#' find_a_term( "-a*b+c")
#' find_a_term( "-f(a*b)+c")
#' find_a_term( "a*F(b+c)")
#'
#' find_a_term( "-a*b+c", complete=0)
#'
#' expr="-b[i,j]*a+c(a,b))-f(x)+a2"
#' res=find_a_term( expr )
#' Print(res)
#' Print(substr(expr,res$loc,res$loc+res$len-1), res$term)
#'
#' @export
#'

find_a_term <- function( expr, pattern=NULL, complete=1 ){
 # find a term separated by a multiplicative operator
 # Shin-ichi Mayekawa
 # 20200817,18,19,20
 #
 # This function finds the first term of expr.
 # The term is defined as the product (including division) of variables.
 # A variable is defined as the string consisting of valid root characters
 # trailed by subscripts in [ and ], if any.
 # Those characters in a balanced parens can be regarded as a variable
 # as well as a fuction reference such as f(args).
 # Therefore, expr such as A[a,b] or f(args) are treated equally.
 #
 # When complete=0, the function returns the first variable.
 # WHen complete=1, the function returns the first term.
 #
 # This function searches for a variable in expr from left to right.
 # If a variable is found and it is trailed by a multiplicative operator,
 # the search continues to the right again.
 #

 # pattern to identify the variable name
 if( is.null(pattern) ){
  sign="([+-]*)"
  varnum="([[:digit:]]+|[[:alpha:]][[:alnum:]_]*)"
  paren="[[({]"
  pattern=paste(varnum,sep="")
 }

 res=list(term="", loc=-1, len=0, pre="", post="")
 lenexpr=nchar(expr)
 term=NULL
 len=0
 len0=0

 if( complete ==  1 ){
  # recursive use to find all the elements of a term

  for(llll in 1:999 ){

   # get a variable or balancing parens
   res1=Recall( expr, complete=0 )
   v=res1$term
   if( v == "" ) return( res )

   term=paste(term,v,sep="")
   len=len+res1$len

   if( llll == 1 ){
    loc=res1$loc+len0
    pre=res1$pre
   }

   # if post is not a multiplicative operator, return.
   post1=res1$post
   if( length(post1) == 0  ||
     ( length(post1) > 0  &&  !(substr(post1,1,1) %in% c("*","~","`","/")) ) ){
    res=named_list(term, loc, len, pre, post=post1)
    return( res )
   }

   # append operator
   term=paste(term,substr(post1,1,1),sep="")
   len=len+1
   len0=len0+res1$loc+res1$len-1

   # next
   expr=res1$post

  } # end of repeat

 } # end of complete



 ######################################################################
 # main body starts here.
 #
 # returns a variable or a variable with subscripts in [ and ],
 # or function with ( and  ).
 #
 # The variable found is: substr( expr, loc, loc+len-1 )
 #
 # loc is the location of the variable or -1,
 # len is the length of the variable or 0,
 # pre is the preceeding chars before variable name,
 # post is the trailing chars after variable name or ] or ).
 ######################################################################

 # location of a variable or left paren
 loc=regexpr( pattern, expr )
 locp=regexpr( paren, expr )

 if( 0 < locp && locp < loc ){
  # left paren is located to the left of first variable name (w/o subs)
  locbp2=locbalpar(expr,locp)
  term=substr(expr,locp,locbp2)
  loc=locp
  len=locbp2-loc+1
 }
 else{
  # variable found
  len=attributes(loc)$match.length
  attributes(loc)=NULL
  term=substr(expr,loc,loc+len-1)
  # Print(expr,pattern,loc,lenexpr)
  if( loc > 0 ){
   if( loc < lenexpr ){
    # variable is trailed by something
    locnext=loc+len
     nextc=substr(expr, locnext, locnext)
    if( nextc %in% c("(","[","{") ){
     # subscript found
     locbp=locbalpar(expr,locnext)
     term=substr(expr,loc,locbp)
     len=locbp-loc+1
    } # end of subscript
   }
  }
 }

 pre=""
 if( loc > 1 ) pre=substr(expr,1,loc-1)
 post=""
 if( loc+len <= lenexpr ) post=substr(expr,loc+len,lenexpr)

 res=named_list(term,loc,len,pre,post)
 # Print(res)
 return( res )

} #   find_a_term


