#' Decompose a Term into Multiplicative Factors separated by
#' \code{ *, /,  \%*\%,  \%\@\%}.
#'
#'
#' @param expr An expression string (scalar, vector or matrix) \cr
#' Each element of expr must not contain multiple terms.
#' @param markso = 0 not to mark the special opeartors
#' @param sort = 1 to sort the factors according to their names
#' @param debug = 1 to print the intermediate result
#'
#' @return
#' A list consisting of \cr
#'  # of factors (nfactors), \cr
#'  a vector of operators (oper), \cr
#'  a vector of factors (factor), \cr
#'  location of each oper+factor (fromto), \cr
#'  a matrix of operators and terms (tab) ,  \cr
#'  a vector of validity flag of each term (valid), \cr
#'  ok flag (ok=1 if the input expr is equal to the reconstructed one.)
#'
#'
#' @details
#' It is assumed that there is just one character in between \% and \%.
#' \cr
#' The i-th factor with the preceding operator of expr is
#' substr( expr0, fromto[i,1], fromto[i,2] ) where
#' expr0 is the blank compressed expr. \cr
#' The prefix operator associated with the i-th factor is oper[i]. \cr\cr
#' The original expression can be recovered as:
#' \preformatted{
#' res <- find_factors( expr )
#' expr <- paste( mapply( paste, res$oper, res$factor, sep="" ), collapse="" )
#' }
#' or
#' \preformatted{
#' expr <- paste( paste( res$oper, res$factor, sep="" ), collapse="" )
#' }
#' or \cr
#' #' \code{
#' expr <- paste( apply(res$tab,1, function(x){paste(x,collapse="")} )
#' , collapse="")
#' }
#' \cr
#' The unbalanced parens, if any,  will be neglected but recorded in
#' \code{valid} flag.
#' \cr
#' \cr
#' The factors can be sorted according to their names as follows: \cr
#' \preformatted{
#' res <- find_factors("A*d/a/c", sort=1)
#' exprs <- paste( paste( res$oper, res$factor, sep="" ), collapse="" )
#'
#' res <- find_factors("A*c/(x+y)/d", sort=1)
#' exprs <- paste( paste( res$oper, res$factor, sep="" ), collapse="" )
#' }
#'
#'
#' @examples
#' find_factors( "a*b/c*(d-e)*f(n)" )
#' find_factors( c("a*b/c*(d-e)*f(n)", "-2.13*a") )
#' find_factors( "a%*%b/c%*%(d-e)%*%f(n)" )
#' find_factors( "ab%*%cd*(b%p%q+a%m%c)/c*d/a%T%b" )
#' # error
#' find_factors( "a+b" )
#'
#' @export
#'

find_factors <- function( expr, markso=1, sort=0, debug=0 ){
 # count the # of terms in an expression
 # Shin-ichi Mayekawa
 # 20150824
 # nterms modified: 20150904,05
 # 20150914dnc,15
 # made sure range1[,1]<=range1[,2]: 20150915
 # "" input: 20150916
 # last * or /: 20150917
 # special operators: 20150919
 # markso option: 20150922
 # sort: 20150923Granda
 # tab and is_valid output: 20230625
 # exp -> expr: 20230625
 #

 # mark
 MARK=";"

 # is expr a scalar?
 if( is.vector(expr) ) lenexpr=length(expr)
 else lenexpr=nrow(expr)*ncol(expr)


 ##################### recursive call ################################
 # if not scalar, use recursive call
 if( lenexpr > 1  |  is.matrix(expr) ){
  res=lapply( expr, find_factors, debug=debug )
  return( res )
 }


 ############# main starts here. #####################################

 # compress expr
 expr=gsub( " ", "", expr )

 if( expr == "" ){
  cat("\nerror1:(find_factors) Input expr is empty.\n\n")
  res=list(nfactorsoper=NULL, factor=NULL, fromto=NULL,ok=0)
  return(res)
 }


 if( markso == 1 ){
  # mark special operators
  expr=mark_soper( expr )
 }

 lenexpr=nchar(expr)
 expr0=expr
 expr00=expr

 # mark the last * or / by ?.
 if( substr(expr,lenexpr,lenexpr) %in% c("*","/","$","&","~","`") ){
  expr=paste(expr,"?",sep="")
  lenexpr=lenexpr+1
 }

 # nterms
 nterms=find_terms(expr)$nterms
 if( nterms( expr ) > 1 ){
  cat("\nerror1:(find_factors)  The elements of expr must not contain"
      , " multiple terms.\n")
  return( NULL )
 }



 if( debug ) Print(expr,expr0)


 # Mark the inside of top level parens
 # find the location of left parens
 temp=unlist(strsplit(expr,""))
 loc1=which( unlist( gregexpr( "\\(", temp ) ) == 1 )
 np=length(loc1)

 range1=0
 if( np > 0 ){

  # find the matching parens
  loc2=loc1
  for( i in 1:np ){
   loc2[i]=locbalpar( expr, loc1[i] )
  }

  # range=mapply( ":", loc1, loc2 )
  range=cbind(loc1,loc2)
  if( debug ) Print(temp,range, printrownames=1)

  # location of top-level parens
  k=1
  loc=1
  for( i in 1:np ){
   ll=which( range[,1] > range[k,2] )
   if( length(ll) == 0 ) break
   loc=c( loc, k <- ll[1] )
  }
  range1=range[loc,,drop=0]
  np1=nrow(range1)
  if( debug ) Print(np1,loc,range1,range)

  # mark the inside of parens
  for( i in 1:np1 ){
   MARKs=paste(rep(MARK,range1[i,2]-range1[i,1]+1),collapse="")
   # mark the inside of parens
   if( range1[i,1] <= range1[i,2] ){
    substr( expr, range1[i,1], range1[i,2] )=
     substr(MARKs,1,range1[i,2]-range1[i,1]+1)
   }
  }

 } # end of marking


 if( debug ) Print(expr, range1)



 # get factors

 split=strsplit(expr,"[$&~`*/]")[[1]]
 split=split[split != ""]
 nfactors=length(split)

 oper=""
 fromto=get_range(nchar(split)+1)
 if( oper != "" ) fromto=fromto+1
 # operator at the end of each factor
 oper=c(oper,unlist(lapply(fromto[,2], function(x){ substr(expr0,x,x) })))
 oper=oper[1:nfactors]
 # i-th factor is in substr(expr0,fromto[i,1],fromto[i,2]-1)
 factor=unlist(apply(fromto,1,function(x){substr(expr0,x[1],x[2]-1)}))

 fromto=fromto-1
 fromto[1,1]=1


 if( debug ) Print(nfactors, split, fromto, oper,factor)


 # last * or /
 if( substr(expr,lenexpr,lenexpr) == "?" ){
  substr(expr,lenexpr,lenexpr)=""
  factor[nfactors]=""
  fromto[nfactors,2]=fromto[nfactors,2]-1
 }



 # reconstruction
 expr1 <- paste( paste( oper, factor, sep="" ), collapse="" )
 if( debug ) Print(expr1,expr00, nchar(expr1),nchar(expr00))
 if( expr1 == expr00 ) ok=1
 else ok=0

 if( markso == 1 ){
  # unmark special operators
  oper=unmark_soper( oper )
  factor=unmark_soper( factor )
 }

 if( debug ) Print(expr,expr0)
 if( debug ) Print(expr1,expr00,ok)
 if( debug ) Print(oper,factor)

 if( sort ){
  # sort factors according to their names
  od=order(factor)
  factor=factor[od]
  oper=oper[od]
  oper[oper==""]="*"
  if( oper[1] == "/" ) factor[1]=paste("1/",factor[1], sep="" )
  oper[1]=""
  if( debug ) Print("sorted", oper,factor)
 }

 # validity of each term found
 valid=as.numeric( c( lapply( lapply( factor, is_valid )
                              ,  function(x) x[["valid"]] ) ) )

 # tab
 tab=interleave(matrix(oper,nfactors,1), matrix(factor,nfactors,1))
 rownames(tab)=paste("factor_",1:nfactors,sep="")
 colnames(tab)=c("oper","factor")

 res=named_list( nfactors, oper, factor, fromto, tab, ok, valid )

 return( res )


} # end of find_factors


