#' Decompose an Expression into Additive Terms separated by \code{+} or \code{-}
#'
#'
#' @param expr An expression string (scalar, vector or matrix)
#' @param markso = 0 not to mark the special opeartors
#' @param sort = 1 to sort the factors according to their names
#' @param debug = 1 to print the intermediate result
#'
#' @return
#' A list consisting of \cr
#' # of terms (nterms), \cr
#' a vector of operators (oper), \cr
#' a vector of terms (term), \cr
#' location of each oper+term (fromto), \cr
#' a matrix of operators and terms (tab),  \cr
#' a vector of validity flag of each term (valid), \cr
#' ok flag (ok=1 if the input expr is equal to the reconstructed one.)
#'
#' @details
#' The i-th term with the preceding operator of expr is
#' substr( expr0, fromto[i,1], fromto[i,2] ) where
#' expr0 is the blank compressed expr. \cr
#' The prefix operator associated with the i-th term is oper[i]. \cr\cr
#' The original expression can be recovered as:
#' \preformatted{
#' res <- find_terms( expr )
#' expr <- paste( mapply( paste, res$oper, res$term, sep="" ), collapse="" )
#' } or \cr
#' \code{
#' expr <- paste( paste( res$oper, res$term, sep="" ), collapse="" )
#' } or \cr
#' \code{
#' expr <- paste( apply(res$tab,1, function(x){paste(x,collapse="")} )
#' , collapse="")
#' }
#' \cr
#' \cr
#' The unbalanced parens, if any,  will be neglected but recorded in
#' \code{valid} flag.
#' \cr
#' \cr
#' The terms can be sorted according to their names as follows: \cr
#' \preformatted{
#' res <- find_terms("c-a+d-b", sort=1)
#' exprs <- paste( paste( res$oper, res$term, sep="" ), collapse="" )
#' }
#'
#'
#' @examples
#' res=find_terms( "-((a)) - c + ((a)+(b)) + (a-b) + 1 - (x-y)*(((z+1)))" )
#' paste( apply(res$tab,1, function(x){paste(x,collapse="")} ), collapse="")
#' find_terms("-a")
#' find_terms("x*a/(1-y)")
#' find_terms("-a+b-0")
#' find_terms("(A%p%B)%*%C-D%*%diag(X) %m% E")
#' find_terms(Simplify("-a+b-0"))
#'
#' find_terms("(((b)+c")
#' find_terms(")*c+5")
#'
#'
#' @export
#'

find_terms <- function( expr, markso=1, sort=0, debug=0 ){
 # count the # of terms in an expression
 # Shin-ichi Mayekawa
 # 20150824
 # nterms modified: 20150904,05
 # fromto output: 20150914
 # made sure range1[,1]<=range1[,2]: 20150915
 # "" input: 20150916
 # last + or -: 20150917
 # special operators: 20150919
 # markso option: 20150922
 # sort: 20150923Granda
 # roxygen2 factor->term: 20200529
 # tab and is_valid output: 20230625
 # exp -> expr: 20230625
 #

 # is expr a scalar?
 if( is.vector(expr) ) lenexpr=length(expr)
 else lenexpr=nrow(expr)*ncol(expr)

 ##################### recursive call ################################
 # if not scalar, use recursive call
 if( lenexpr > 1  |  is.matrix(expr) ){
  res=lapply( expr, find_terms, debug=debug )
  return( res )
 }


 ############# main starts here. #####################################

 # blackout mark
 MARK=";"

 # compress expr
 expr=gsub( " ", "", expr )

 if( expr == "" ){
  cat("\nerror1:(find_terms) Input expr is empty.\n\n")
  res=list(nterms=0, oper=NULL, term=NULL, fromto=NULL,ok=0)
  return(res)
 }


 if( markso == 1 ){
  # mark special operators
  expr=mark_soper( expr )
 }


 lenexpr=nchar(expr)
 expr0=expr

 # mark the last + or - by ?.
 if( substr(expr,lenexpr,lenexpr) %in% c("+","-","!","#") ){
  expr=paste(expr,"?",sep="")
  lenexpr=lenexpr+1
 }
 MARKs=paste(rep(MARK,lenexpr),collapse="")


 # find the location of left parens
 temp=unlist(strsplit(expr,""))
 loc1=which( unlist( gregexpr( "\\(", temp ) ) == 1 )
 np=length(loc1)

 if( np > 0 ){

  # find the matching parens
  loc2=loc1
  for( i in 1:np ){
   loc2[i]=locbalpar( expr, loc1[i] )
  }

  # range=mapply( ":", loc1, loc2 )
  range=cbind(loc1,loc2)
  if( debug ) Print(temp,range, printrownames=1)

  # location of top-level parens
  k=1
  loc=1
  for( i in 1:np ){
   ll=which( range[,1] > range[k,2] )
   if( length(ll) == 0 ) break
   loc=c( loc, k <- ll[1] )
  }
  range1=range[loc,,drop=0]
  np1=nrow(range1)
  if( debug ) Print(np1,loc,range1,range)

  # mark the inside of parens
  for( i in 1:np1 ){
   if( range1[i,1] <= range1[i,2] ){
    substr( expr, range1[i,1], range1[i,2] )=
     substr(MARKs,1,range1[i,2]-range1[i,1]+1)
   }
  }

 }

 if( debug ) Print(expr0,"/",expr)

 # count the terms
 oper=""
 if( substr(expr,1,1) %in% c("+","-","!","#")  ){
  oper=substr(expr,1,1)
  expr=substr(expr,2,nchar(expr))
 }
 nterms=length(split <- strsplit(expr,"[-+!#]")[[1]])

 # get terms
 fromto=get_range(nchar(split)+1)
 if( oper != "" ) fromto=fromto+1
 # operator at the end of each term
 oper=c(oper,unlist(lapply(fromto[,2], function(x){ substr(expr0,x,x) })))
 oper=oper[1:nterms]
 # i-th term is in substr(expr0,fromto[i,1],fromto[i,2]-1)
 term=unlist(apply(fromto,1,function(x){substr(expr0,x[1],x[2]-1)}))

 # adjust fromto
 fromto=fromto-1
 fromto[1,1]=1

 # last + or -
 if( substr(expr,lenexpr,lenexpr) == "?" ){
  substr(expr,lenexpr,lenexpr)=""
  term[nterms]=""
  fromto[nterms,2]=fromto[nterms,2]-1
 }

 if( debug ) Print(nterms, split, fromto, oper,term)

 # reconstruction
 expr1=paste(oper[1],term[1],sep="")
 if( nterms >= 2 ){
  for( i in 2:nterms ){
   expr1=paste(expr1,oper[i],term[i],sep="")
  }
 }
 if( debug ) Print(expr1,expr0, nchar(expr1),nchar(expr0))
 if( expr1 == expr0 ) ok=1
 else ok=0

 if( markso == 1 ){
  # unmark special operators
  oper=unmark_soper( oper )
  term=unmark_soper( term )
 }

 if( sort ){
  # sort factors according to their names
  od=order(term)
  term=term[od]
  oper=oper[od]
  oper[oper==""]="+"
  if( debug ) Print("sorted", oper,term)
 }

 # validity of each term found
 valid=as.numeric( c( lapply( lapply( term, is_valid )
                           ,  function(x) x[["valid"]] ) ) )

 # tab
 tab=cbind(oper,term)
 rownames(tab)=paste("term_",1:nterms,sep="")

 res=named_list( nterms, oper, term, fromto, tab, ok, valid )
 return( res )

} # end of find_terms


