#' Find Variables and Functions in an Expression
#'
#' @param expr An input expression as a (scalar) string.
#' @param func = 1 to return the functions with nesting level info
#' @param mark = 1 to mark the found ones.
#' @param markchar A character vector of length=2,
#' The first one is for variable, the 2nd, function.
#' @param print = 1 to print the result.
#' @param debug = 1 to print the intermediate results.
#'
#' @details
#' Given a valid expression, decompose it into variables and functions.
#'
#' @return
#' A list of\cr
#' \code{expr} (marked) input expression \cr
#' \code{varlist} a list or variables/functions in expr \cr
#' \code{is.function} indicator vector: \code{is.function[i]=1}
#'  if \code{varlist[i]} is a function \cr
#' \code{fromto} the range of the variables/functions \cr
#' \code{bp} the location of the left and the right parens for functions \cr
#' \cr
#' In addition, if \code{func=1}, \cr
#' \code{fname} function name. \cr
#' \code{fpnum} parent function number. \cr
#' \code{fpname} parent function name. \cr
#' \code{flevel} function level: (flevel=1 means the top level.) \cr
#' \code{fbody} function body (with enclosing parens).
#'
#'
#' @examples
#' find_varfunc("-a(10)+b1*c2-d(x)", mark=0, print=1)
#' find_varfunc("-a(10)+b1*c2-d(x)", mark=1, print=1)
#' find_varfunc("a-b+1*(c-d-1*(e+f)-g)+1*(h1(x+g(y))+(i+j))", mark=1, print=1)
#'
#' expr="-BB+f(C1-D)-a*gg(h1(EE)+P+h2(XX-q(y)))-H"
#' res=find_varfunc( expr, print=1, func=1)
#'
#' @export
#'

find_varfunc <- function( expr, func=0, mark=0, markchar=c("V","F")
                        , debug=0, print=0 ){
 # find variables or functions
 # Shin-ichi Mayekawa
 # 20230430cot,0502,03
 # all numeric input: 20230503
 # rownames to bp and fromto: 20230623
 # bugfix for marking funcs: 20230623
 # check if the parens are matched: 20230623
 # type option removed: 20230623
 # func: 20230627dnc
 # old func renamed as is.func: 20230628
 # roxgen details added: 20250301
 #

 # delimiters: Below can preceed a variable or function
 # oper=c("+","-","*","/","%","`","!","#","$","&","~","(")
 # oper="\\+\\*\\/\\$\\(%`!#&~ -"
 #

 if( length(expr) > 1 ){
  cat("\nerror1(find_varfunc): expr must be a scalar.\n")
  return( NULL )
 }
 lexpr=nchar(expr)

 # check if the parens are matched.
 loc_paren=gregexpr( "(", expr, fixed=1 )[[1]]
 if( loc_paren[1] > 0 ){
  locp=numeric(length(loc_paren))
  for( i in 1:length(loc_paren) ){
    locp[i]=locbalpar( expr, loc_paren[i] )
  }
  if( any(locp == 0) ){
    cat("error1:(find_varfunc)  Missing", sum(locp==0), "right parns.\n")
    return( NULL )
  }
 }
 loc_paren=gregexpr( ")", expr, fixed=1 )[[1]]
 if( loc_paren[1] > 0 ){
   locp=numeric(length(loc_paren))
   for( i in 1:length(loc_paren) ){
     locp[i]=locbalpar( expr, loc_paren[i] )
   }
   if( any(locp == 0) ){
     cat("error1:(find_varfunc)  Missing", sum(locp==0), "left parns.\n")
     return( NULL )
   }
 }



 # variables: delim + alpha + alnum
 locvf=gregexpr(
    "(^|[\\+\\*\\/\\$\\(%`!#&~ -]+)([[:alpha:].]+[[:alnum:]._]*)"
                     , expr )[[1]]


 nvar=length(locvf)
 mlen=attributes(locvf)$match.length
 attributes(locvf)=NULL
 if( debug ) Print(expr,lexpr,locvf,mlen)

 varlist=character(nvar)
 fromto=matrix(0,nvar,2)
 bp=matrix(0,nvar,2)
 is.function=rep(0,nvar)
 if( nvar == 0 ) return( valist )

 for( i in 1:nvar ){

  fromto[i,1]=locvf[i]
  fromto[i,2]=locvf[i]+mlen[i]-1

  # remove delimiter before varfunc name
  loc=regexpr("[[:alpha:].]",substr(expr,fromto[i,1],fromto[i,2]))
  if( loc > 1 ){
   fromto[i,1]=locvf[i]+loc-1
  }

  if( lexpr > fromto[i,2] && substr(expr,fromto[i,2]+1,fromto[i,2]+1)=="(" ){
    # function name found
    is.function[i]=1
    locL=fromto[i,2]+1
    locR=locbalpar( expr, locL)
    bp[i,1]=locL; bp[i,2]=locR
  }

  if( debug ){
   Print(fromto[i,1],fromto[i,2])
   Print(substr(expr,fromto[i,1],fromto[i,2]))
  }

  varlist[i]=substr(expr,fromto[i,1],fromto[i,2])

 } # end of i loop

 rownames(fromto)=varlist
 rownames(bp)=varlist

 if( func == 1 ){
  locfunc=which(is.function == 1)
  nfunc=length(locfunc)
  varlist=varlist[locfunc]
  fname=varlist
  fromto=fromto[locfunc,,drop=0]
  bp=bp[locfunc,,drop=0]
  is.function=is.function[locfunc]
  if( nfunc > 0 ){

   # function levels
   fpnum=numeric(nfunc); flevel=fpnum;
   fbody=character(nfunc); fpname=fbody
   fpnum[1]=0; flevel[1]=1
   fbody[1]=substr(expr,bp[1,1],bp[1,2])

   if( nfunc >= 2 ){
    for( i in 2:nfunc ){

     # index
     lbp=bp[i,1]
     fbody[i]=substr(expr,bp[i,1],bp[i,2])
     if( debug ) Print(i,varlist[i], fbody[i])

     # parent etc
     # is this function in other ones?
     loc4=which( bp[1:(i-1),2] > bp[i,2] )
     if( length(loc4) > 0 ){
       # parent = nearest function above
       # level = parent's level + 1
       fpnum[i]=loc4[length(loc4)]
       flevel[i]=flevel[fpnum[i]]+1
     }
     else flevel[i]=1

    } # end of i loop
   }

   fpname=c(".",fname)[fpnum+1]
   names(fpnum)=fname; names(fpname)=fname
   names(flevel)=fname; names(fbody)=fname

  }
 } # end of func
 else{
   fname=NA; fpnum=NA; fpname=NA; flevel=NA;  fbody=NA
 }


 if( print ){
   Print("input", expr)
   Print( fromto, varlist, is.function, bp)
   if( func == 1 ) Print(fname,fpnum,fpname,flevel, fbody)
 }


 if( mark && any(varlist != "") ){

  for( i in 1:nvar ){
   # mark functions
   if( is.function[i] == 0 ){
      substr( expr, fromto[i,1], fromto[i,2] )=
        paste(rep(markchar[1],fromto[i,2]-fromto[i,1]+1), collapse="")
    }
  }
  for( i in 1:nvar ){
   # mark variables
   if( is.function[i] == 1 ){
     substr( expr, fromto[i,1], bp[i,2] )=
       paste(rep(markchar[2], bp[i,2]-fromto[i,1]+1), collapse="")
   }
  }

  if( print ){
    Print("marked", expr)
  }

 } # end of mark

 # return
 res=named_list( expr, varlist, is.function, fromto, bp
                 , fname, fpnum, fpname, flevel, fbody)

 return( res )

} # end of find_varfunc



