#' Generate Random Expression
#'
#' @param natom  # of atmic elements in an expression
#' @param nparen  # of parentheses in an expression
#' @param nparen2  # of higher level parentheses in an expression
#' @param random = 1 to randomize the roots of the atomic elements
#' @param maxlena   # of alphabets in the root
#' @param maxlennum  # of numerics in the root
#' @param oper  a vector consisting of the binary operators to be used
#' @param alpha a vector of characters to be used as the root
#' @param num a vector of the numbers to be used as the root
#' @param print = 1 to print the result \cr
#'  = 2 to print the expanded result.
#'
#' @details
#' This function generates a valid expression according to the specification
#' given by the input parameters
#'
#' @return an expression as a character vector of length one
#'
#' @examples
#' set.seed(1701)
#' gen_random_expr(  )
#' gen_random_expr( natom=9, random=1, nparen=3 )
#' set.seed(1701)
#' gen_random_expr( natom=9, random=1, nparen=3, nparen2=3 )
#' expr <- gen_random_expr( natom=9, random=0, nparen=3, nparen2=1
#'          , oper=c("%*%","+","-"), alpha=LETTERS[-9] )
#' expr <- sub( "D", "diag(D)", expr, fixed=1 )
#'
#' @export
#'

gen_random_expr <- function( natom=2, nparen=0, nparen2=0
               , random=0, maxlena=1, maxlennum=0
               , oper=c("+","-","*")
               , alpha=c(letters,LETTERS)
               , num=formatC(0:9,digits=0,format="d")
               , print=0, debug=0 ){
 # generate random expressions
 # Shin-ichi Mayekawa
 # 20200517MMC,18,19
 # simple version: 20200519
 # debug: 20200603
 # nparen2: 20200603
 # insert_paren: 20200603,04
 # nparen3: 2020060   Does NOT work!!!!!!
 # rosygen2 example added: 20200820
 #
 #
 #  Args:
 #     natom      # of atomic elements in an expression
 #     nparen     # of parenteses in each of expression
 #     random     = 0 to use the elements of alpha sequentially as the root
 #     maxlena    max length of the alpha root
 #     maxlennum  max length of the postfix numbers
 #     oper       a vector consisting of the operators to be used
 #                Use oper=c("%*%","+","-") for a matrix expression.
 #     alpha      a vector of alphabetic root candidates
 #                Use alpha=LETTERS to restrict the roots to upper case.
 #     num        a vector of numeric postfix candidates
 #     print      = 1 to print the result
 #                = 2 to print the expanded result
 #
 #
 # Details
 #
 #   By dafault, the variables such as a, b1, c, d2, ... will be generated.
 #   The dafault operatos are: +, -, and *.
 #
 #   expression = oper1  &  alpha1   &   num1
 #                oper2  &  (        &   alpha2   &   num2     &      )
 #
 #
 # Values:
 #
 #   expr   the expression generated
 #



 insert_paren <- function( nparen2, df, debug=0 ){
  # insert parentheses
  # Shin-ichi Mayekawa
  # This belonged to the main part of gen_random_expr
  #

  # from df
  if( debug ){
   cat("******* Starting insert_paren  with nparen2=", nparen2, "*******\n")
   Print(df)
  }
  lenexpr=nrow(df)
  from=df$from
  to=c(df$to,max(df$to)+1)
  lp=df$lp.x
  atoms=df$atoms
  rp=df$rp.x
  opers=df$opers

  expr=paste( paste( df$opers, df$lp.x, df$lp.y, df$atoms
                     , df$rp.x, df$rp.y, sep="" ), collapse="" )

  # The first operator cannot be empty because of locparens.
  if( opers[1] == "" ) opers[1]="?"

  # location of operators in from
  # at which the left parens must be inserted.
  locopers=which( opers != "" )
  lenlocopers=length(locopers)      #  +     1   ###############

  #  ( is to the right of an operator
  #  ) is to the left  of an operator

  #
  if( debug ){
   cat("expr w/o parens\n",expr,"\n")
   Print(from,to,opers,atoms)
  }

  ft=matrix(0,nparen2,2)
  used=NULL
  candidate=1:lenlocopers
  for( j in 1:nparen2 ){
   ######################################
   if( j == 1 ) candidate=1:lenlocopers
   else candidate=(1:lenlocopers)[-used]
   ######################################
   #Print(j,candidate)
   #   if( length(candidate) == 1 ) locparen=candidate
   #   else locparen=sample(x=candidate,size=1)
   locparen=candidate[sample.int(length(candidate),1)]
   ft1=locopers[locparen]
   used=c(used,locparen)
   rest=(locparen+1):(lenlocopers+1)
   #   if( length(rest) == 1 ) locparen2=rest
   #   else locparen2=sample(x=rest,size=1)
   locparen2=rest[sample.int(length(rest),1)]
   ft2=locopers[ locparen2 ]  # 20200603
   if( debug >= 2 ) Print(j,used,rest,locparen, locparen2)
   ft[j,1]=from[ft1]
   ft[j,2]=to[ft2]
  } # end of j loop

  # remove the first operator dummy mark
  if( opers[1] == "?" ) opers[1]=""

  ft=ft[order(ft[,1]),,drop=0]
  # Print(ft)
  # Print(from,opers,lp,atoms,rp)
  df1=data.frame( from,opers,lp,atoms,rp, stringsAsFactors=0 )
  df2=data.frame( from=ft[,1], lp="(", stringsAsFactors=0 )
  df3=data.frame( from=ft[,2], rp=")", stringsAsFactors=0 )

  if( debug >= 2 ){
   Print(df1,df2,df3)
  }

  df=merge( df2, df1, by="from", all=1, stringsAsFactors=0 )
  if( debug >= 2 ) Print("d12",df)
  df=merge( df,  df3, by="from", all=1, stringsAsFactors=0 )
  if( debug >= 2 ) Print("d123",df)

  #  ( is to the right of an operator
  #  ) is to the left  of an operator
  df[is.na(df)]=""
  df=df[,c("from","opers","lp.x","lp.y","atoms","rp.x","rp.y")]

  if( debug ){
   Print(df)
   expr=paste( paste( df$opers, df$lp.x, df$lp.y, df$atoms
                      , df$rp.x, df$rp.y, sep="" ), collapse="" )
   valid=is_valid(expr, print=0)$valid
   Print(expr,valid)
   expr1=Expand(expr)
   Print(expr1)
   expr2=Expand(expr1)
   Print(expr2)
   # return(expr2)
  }


  nchar=apply(df[,c("opers","lp.x","lp.y","atoms","rp.x","rp.y")]
              ,1,function(x){ sum(nchar(x))})
  df=data.frame( df, nchar, stringsAsFactors=0
                 , from1=c(1,cumsum(nchar)+1)[1:nrow(df)], to1=cumsum(nchar) )

  df=data.frame( from=df$from1, to=df$to1, stringsAsFactors=0
                 , df[,c("opers","lp.x","lp.y","atoms","rp.x","rp.y")] )
  if( debug ) Print(df)

  # move paren from y to x
  for( i in 1:nrow(df) ){
   if( df$lp.x[i] == ""  &  df$lp.y[i] == "(" ){
    df$lp.x[i]="("; df$lp.y[i]=""
   }
  }
  for( i in 1:nrow(df) ){
   if( df$rp.x[i] == ""  &  df$rp.y[i] == ")" ){
    df$rp.x[i]=")"; df$rp.y[i]=""
   }
  }

  return( df )

 } # end of insert_paren




 ########################################################################

 # error
 if( nparen > natom | nparen2 > natom ){
  cat("\nerror1(gen_random_expr): nparen or nparens too large!!\n")
  return( NULL )
 }

 if( debug ){
  Print(natom,nparen,random,maxlena,maxlennum)
 }

 # expand alpha
 if( length(alpha) < natom ){
  alpha=c(alpha,rep(alpha,natom))[1:natom]
 }

 # operators
 num=formatC(0:9,digits=0,format="d")
 oper1=c("+","-","*")
 oper2=c("%*%","%@%")
 oper_used=oper

 # Print(natom,maxlena,maxlennum,random,nparen)
 # Print(alpha,num,oper_used)

 # storage
 fromto=matrix(0,natom,2)
 opers=character(natom)
 atoms=character(natom)

 # length of alpha root and trailing numbers
 lena=sample(1:maxlena,1)
 lennum=sample(0:maxlennum,1)

 # first atom
 atom0=alpha[1]
 if( random ){
  atom0=paste( paste(sample(alpha,lena,replace=1), sep="",collapse="")
         , paste(sample(num,lennum,replace=1), sep="",collapse="")
         , sep="" )
 }
 oper0=sample(c("","+","-"),1)
 expr=paste(oper0,atom0,sep="")

 fromto[1,]=c(1,nchar(expr))
 opers[1]=oper0
 atoms[1]=atom0

 # 2nd atom and later
 if( natom > 1 ){
  for( j in 2:natom ){

   # length of alpha root and trailing numbers
   lena=sample(1:maxlena,1)
   lennum=sample(0:maxlennum,1)
   operj=sample(oper_used,1)
   atomj=alpha[j]
   if( random ){
    atomj=paste( paste(sample(alpha,lena,replace=1), sep="",collapse="")
           , paste(sample(num,lennum,replace=1), sep="",collapse="")
           , sep="" )
   }
   expr1=paste(operj,atomj,sep="")
   expr=paste(expr,expr1,sep="")

   # Print(i,j,expr,expr1, nchar(expr),nchar(expr1))

   fromto[j,]=fromto[j-1,2]+c(1,nchar(expr1))
   opers[j]=operj
   atoms[j]=atomj

  } # end of j loop
 }

 df=data.frame( from=fromto[,1],to=fromto[,2], opers, lp.x="", lp.y=""
         , atoms, rp.x="", rp.y="", stringsAsFactors=0 )
 if( debug ){
  cat("End of No Paren\n")
  Print(expr)
  Print(df)
 }



 if( nparen >= 1 ){

  if( debug ) Print("\n\nFirst Round")
  df=insert_paren( nparen, df, debug=debug )

 } # end of nparen > 0 (round 1)


 if( nparen2 > 0 ){

  if( debug ) Print("\n\nSecond Round")
  df=insert_paren( nparen2, df, debug=debug )

 } # end of nparen2 > 0


  # Print(df)
 expr=paste( paste( df$opers, df$lp.x, df$lp.y, df$atoms
           , df$rp.x, df$rp.y, sep="" ), collapse="" )


 if( print ){
  Print(expr)
  if( print >= 2 ) Print(Expand(expr))
 }

 # ff=find_factors(expr)
 # ft=find_terms(expr)
 # Print(ft$nterms)

 return( expr )

} # gen_random_expr



