#' Matrix Inversion by Gaussian Elimination
#'
#'
#' @param A Input Matrix (nueric or symbolic)
#' @param loc vector of pivot locations
#' @param nopermute = 0 to find the best pivot when A is numeric.
#' @param eps eps for zero.
#' @param simplify = 0 not to implify the result when A is character.
#' @param print = 1 to print the intermediate results.
#'
#' @details
#'  This function performs the Gaussian elimination on \code{A} matrix.
#'  When \code{loc=1:nrow(A)},
#'  the elimination starts from column 1 to column \code{n=ncol(A)} as:
#'  \preformatted{
#'   B0 <- cbind( A, I )
#'   B1 <- T1 \%*\% B0
#'   B2 <- T2 \%*\% B1
#'    :        :
#'   Bk <- Tk \%*\% Bk-1
#'    :        :
#'   Bn <- Tn \%*\% Bn-1
#'  }
#'  where Tk matrix makes the k-th column of Bk matrix a unit vector
#'  whose k-th element is 1 and 0 elsewhere.
#'  \code{B[i,i]} is used as the pivot.
#'
#'  At the end of the iteration, we have
#'  \preformatted{
#'    Bn = cbind( I, inv(A) )
#'  }
#'  In other words
#'  \preformatted{
#'   inv(A) = Tn \%*\% Tn-1 \%*\% ... \%*\% T2 \%*\% T1
#'  }
#'
#' When \code{loc != 1:nrow(A)}, at each step, \code{B[loc[i],loc[i]]} is used
#' as the pivot.
#'
#' When \code{length(loc) < nrow(A)}, the intermediate result will be returned.
#'
#' @return
#' \code{ cbind(I, inv(A) } \cr
#' The inverse of \code{A} is given as \code{cbind(I,inv{A})[,-c(1:nrow(A))]}.
#'
#' @examples
#'
#' # Symbolic matrix A.
#'
#' n <- 2
#' Aa <- demomat(n,n,root="a")
#'
#' set.seed(1701)
#' An <- matrix(rnorm(n*n),n,n)
#'
#' res <- invGE(Aa, print=1)
#' II <- res[,c(1:n)]
#' invAa <- res[,-c(1:n)]
#'
#' # numerical checking
#' Eval(II, Aa=An)
#' # Eval(Aa%*%invAa, Aa=An)
#' Eval(matTimes(Aa,invAa), Aa=An)
#'
#'
#' # Numeric matrix A.
#'
#' set.seed(1702)
#' n <- 4
#' A <- matrix( sample(0:9,n*n,replace=1), n,n )
#'
#' # result
#' res <- invGE(A, print=1)
#' invA <- res[,-(1:n)]
#' printm(A, invA, A%*%invA, fmt=".2")
#'
#'
#'
#' # Example of permutation
#' # needs permutation
#' set.seed(1701)
#' n=4
#' A=matrix( sample(0:9,n*n,replace=1), n,n )
#'
#' # wrong result
#' invA=invGE(A, print=1)[,c(1:n)]
#' printm( A%*%invA )
#'
#' # correct result by nopermute=0 option
#' invA=invGE(A, print=1, nopermute=0)[,-c(1:n)]
#' printm( A%*%invA, fmt="7.3"  )
#'
#' # correct by specifying loc argument
#' invA=invGE(A, print=1, loc=c(1,4,3,2))[,-c(1:n)]
#' printm( A%*%invA, fmt="6.3" )
#'
#'
#'
#' # non full rank matrix A.
#' set.seed(1701)
#' n <- 4
#' A <- matrix( sample(0:9,n*n,replace=1), n,n )
#' A[,4] <- A[,1]+A[,2]
#'
#' res <- invGE(A, print=1, nopermute=0)
#' invA <- res[,-c(1:n)]
#' round( A%*%invA ,6 )
#' chkginv(A,invA)
#'

#'
#'
#' @export
#'

invGE <- function( A, loc=1:nrow(A), nopermute=1, eps=1e-9
                   , simplify=1,  print=0  ){
 # inverse matrix by Gaussian Elimination
 # Shin-ichi Mayekawa
 # 20230420
 # symbolic A: 20250201
 # length(loc): 20250303
 # use loc[i] everywhere: 20250303
 # modify_math_operators: 20250315,16
 #



 #
 # Seems the we cannot use modify_math_operators() here,
 # possibly because Simnplify is sued?
 #
 
 
 if(0){
 ##########################################################################
 ################ Infix Operators from lazy.symbplic2 package #############
 ##########################################################################
 `+` <- function( A, B=NULL ){
  # + operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   if( is.null(B) ) return( Plus(0,A))
   else return( Plus(A,B) )
  }
  else{
   if( is.null(B) )  return(  .Primitive("+")(0,A) )
   else return(  .Primitive("+")(A,B) )
  }
 } # end of +

 `-` <- function( A, B=NULL ){
  # - operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   if( is.null(B) ) return( Minus(0,A))
   else return( Minus(A,B) )
  }
  else{
   if( is.null(B) )  return(  .Primitive("-")(0,A) )
   else return(  .Primitive("-")(A,B) )
  }
 } # end of -

 `*` <- function( A, B ){
  # * operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   return( Times(A,B) )
  }
  else{
   return(   .Primitive("*")(A, B) )
  }
 } # end of *
 
  `/` <- function( A, B ){
  # / operator
  # Shin-ichi Mayekawa
  # 20250301

  if( is.character(A) | is.character(B) ){
   return( paste("(", A, ")/(", B, ")", sep="") )
  }
  else{
   return(   .Primitive("/")(A, B) )
  }
 } # end of /
 
 `%*%` <- function( A, B ){
  # %*% operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   return( matTimes(A,B) )
  }
  else{
   return(   .Primitive("%*%")(A, B) )
  }
 } # end of %*%
 ##########################################################################
 ##########################################################################
 ##########################################################################
 } # if0
 else{
  modify_math_operators(print=0)
 }


 # abs
 abs <- function( x ){
  if( is.character(x) ){
   return( paste("(", x, ")", sep="") )
  }
  else{
   return(   .Primitive("abs")(x) )
  }
 }
 
 n=nrow(A)

 # no permute when A is character or loc is given.
 if( is.character(A) ) nopermute=1
 if( !identical( loc, 1:n ) ) nopermute=1

 if( !is.matrix(A) ){
  cat("\n error1:(invGE) Input must be a matrix. \n")
  stop()
 }


 # SMdim and fullsymb attribute
 SMdim=attributes(A)$SMdim
 fullsymb=attributes(A)$fullsymb

 if( !is.null(fullsymb) ){
  cat("\n error1:(invGE) Cannot handle fully symbolic matrix yet. \n")
  stop()
 } # end of full symbolic A matrix


 AI=cbind(A,diag(n))
 rownames(AI)=paste("r",1:n,sep="")
 colnames(AI)=c(paste("c",1:n,sep=""),paste("I",1:n,sep=""))

 if( print ){
  cat("\nInitial augumented matrix\n")
  Printm(AI, fmt="5.2")
 }

 loc1=1:n

 for( i in 1:length(loc) ){

  if( nopermute != 1 ){
   # partial pivoting
   # find the row with max value at column i of AI.
   locmax=which.max(abs(AI[i:n,i])) + i-1
   if( locmax != i ){
    AI[c(i,locmax),]=AI[c(locmax,i),]
    if( print ){
     cat("\nafter permutation of rows ", i, " and ", locmax, "\n")
     Printm(AI, fmt="5.2")
    }
    loc1[i]=locmax
    loc1[locmax]=i
   }
  } # end of partial pivoting

  # pivot
  li=loc[i]
  pivot=AI[li,li]

  # check if A is singular
  if( is.numeric(A)  &&  abs(pivot) <= eps ){
   cat("\n*** Input matrix seems to be singular. ***\n")
   return( AI )
  }

  # T matrix
  T=diag(n)
  T[,li]=-AI[,li, drop=FALSE]/pivot
  T[li,li]=1/pivot

  # accumulate
  AI=T%*%AI

  if( simplify && is.character(A) ){
   T=Expand(T)
   T=Simplify(T, all=1)
   AI=Expand(AI)
   AI=Simplify(AI, all=1)
  }

  if( print ){
   if( nopermute == 0 )
    cat("\nafter pivoting at i = ", i, " (",loc1[i],")", "\n")
   else
    cat("\nafter pivoting at i = ", i, " (",loc[i],")", "\n")
   Printm(T, AI[,1:n], AI[,-(1:n)], fmt="5.2")
  }

 } # end of i loop

 # reorder according to loc
 # if( length(loc) == n ){
 #  AI=AI[loc,]
 # }

 # return the inverse
 return( AI )

} # end of invGE



