#' Check If the Input Expression is Valid
#'
#' @param expr An expression as a string
#' @param print = 1 to print the result
#'
#' @details
#' Use \code{find_varfunc(expr)} to list variables/functions in expr.
#'
#' @return A list of checks\cr
#' \code{valid} = TRUE if the input expression is valid \cr
#' meaning the following ok's are all zero.
#' \code{errlist}  A list of error names if any. \cr
#' \code{ok} A vector of the following check flags: \cr\cr
#' \code{ubp} : unbalanced parentheses \cr
#' \code{)a} : no operators after ) \cr
#' \code{+)|(*} : parenthesis after or before an operator \cr
#' \code{*+} : operators w/o variables \cr
#' \code{/+} : operators w/o variables \cr
#' \code{1a} : number followed by alpha or left paren \cr
#' \code{**} : multiplicative operators at start or any oprs. at end \cr
#' \cr
#' \code{errflag} return values from regexpr
#'
#'
#' @examples
#' is_valid( "a+b/c" )
#' is_valid( "a*(b+c)d +9f(a)" )
#' is_valid("a-b(c-2d-1*(e+f)-g+))a+*(h1a(x+g(y)/+4)+(i+j))", print=1)
#' is_valid("(v+v+(v+(+v+v)+v)+((+v*(+((+(v)*v))))+v))+(v)+v+v*v)")
#'
#' @export
#'

is_valid <- function( expr, print=0 ){
 # check if the input expr is valid
 # Shin-ichi Mayekawa
 # 20150920,21,22
 # bugfix for balpar: 20200606
 # print=0 as default: 20200606
 # simplified: 20200606,07,08
 # bugfix: 20200716
 # exp -> expr: 20230430cot
 # function name: 20230430cot,0503
 # numeric: 20230503
 # bugfix for e5: 20230627
 # starting or ending with opers e6: 20250313

 # numeric?
 if( is.numeric(expr) ) return( list(valid=TRUE) )


 # compress
 expr=gsub( " ", "", expr )


 # check if the parentheses are balanced
 locparen=gregexpr( "[()]", expr )[[1]]
 if( any(locparen > 0) ) nparen=length(locparen)
 else nparen=0
 locubp=-1
 if( nparen > 0 ){
  locbp=numeric(nparen)
  attributes(locparen)=NULL
  for( i in 1:nparen ){
   locbp[i]=locbalpar( expr, locparen[i] )
  }
  locubp=which(locbp==0)
  if( length(locubp) == 0 ){
   locubp=-1
  }
  else locubp=999
 }

 # after right paren:  )a, )(
 e1=regexpr( "\\)([[:alnum:]]|\\()", expr )[[1]]

 # before right paren:  *), %), -), +)
 # and after left paren:  (*, (/, (%, ()
 e2=regexpr( "([\\%\\*\\(+-]\\))|(\\([\\*\\%\\)/])", expr )[[1]]

 # *+ or *- or -* or -/
 e3=regexpr( "([\\%\\*/+-][+-])|([+-][\\%\\*/+-])", expr )[[1]]

 # /+ or +/
 e4=regexpr( "/[\\*\\%+-]|[\\*\\%+-]/", expr )[[1]]

 # 1a or 1():  math oper + digit + (alpha or ( )
 e5=regexpr( "(^|[\\+\\*\\/\\$\\(%`!#&~ -]+)[[:digit:]]+([[:alpha:]]|\\()"
             , expr )[[1]]

 # )* etc at the end
 e6=regexpr("[-+*/%]$|^[*/%]", expr )[[1]]

 errname=c( " ubp ",   " )a ", " +)|(* ", " *+ ", " /+ ", " 1a ", " ** " )
 errflag=c( locubp, e1, e2, e3, e4, e5, e6 )
 names(errflag)=errname

 ok=as.numeric(errflag==-1)
 names(ok)=errname

 valid=all(ok == 1 )
 errlist=errname[which(!ok)]
 if( length(errlist) == 0 ) errlist=""

 if( print ){
   Print(valid, errlist, errflag)
   print(ok)
 }

 res=named_list( valid, errlist, ok )

 return( res )

} # end of is_valid


