#' Mark Special Operator by Single Alpha
#'
#' @param exp Input expression which contains special operators
#' @param MARK Alternative definition of MARKs as a string
#' @param debug = 1 tp print intermediate result
#'
#' @details
#' The following marking rules will be used. \cr
#'  \code{MARKs="`"}  for \code{\%*\%} \cr
#'  \code{MARKp="!"}  for \code{\%p\%} \cr
#'  \code{MARKm="#"}  for \code{\%m\%} \cr
#'  \code{MARKt="$"}  for \code{\%t\%} \cr
#'  \code{MARKT="&"}  for \code{\%T\%} \cr
#'  \code{MARKK="~"}  for \code{\%@@\%} \cr
#'  \cr
#'  MARK parameter has the following format: \cr
#'   MARK=' MARKp="p"; MARKm="m" '
#'
#' @return
#' Expression with the spacial operators marked.
#'
#' @examples
#' exp="(A %p% B %m% C) %t% (D %T% E) %@@%  F %*% G"
#' mark_soper( exp )
#' unmark_soper( exp )
#'
#' @export
#'

mark_soper <- function( exp, MARK=NULL, debug=0 ){
 # mark special operators
 # Shin-ichi Mayekawa
 # 20150919
 #
 # MARK='MARKp="p"; MARKm="m"'
 #

 # mark
 MARKs="`" # for %*%
 MARKp="!" # for %p%
 MARKm="#" # for %m%
 MARKt="$" # for %t%
 MARKT="&" # for %T%
 MARKK="~" # for %@%


 # change mark if necessary
 if( !is.null(MARK) ){
  eval(parse(text=MARK))
 }

 if( debug ) Print( MARKp,MARKm,MARKt,MARKT,MARKK )

 # replace
 exp[]=gsub( "%*%", MARKs, exp, fixed=1 )
 exp[]=gsub( "%p%", MARKp, exp, fixed=1 )
 exp[]=gsub( "%m%", MARKm, exp, fixed=1 )
 exp[]=gsub( "%t%", MARKt, exp, fixed=1 )
 exp[]=gsub( "%T%", MARKT, exp, fixed=1 )
 exp[]=gsub( "%@%", MARKK, exp, fixed=1 )

 return( exp )

} # end of mark_soper



#' Recover the Marked Special Operators
#'
#' @param exp Input expression which contains marked special operators
#' @param MARK Alternative definition of MARKs as a string
#' @param debug = 1 tp print intermediate result
#'
#' @details
#' The following marking rules will be used. \cr
#'  \code{MARKs="`"}  for \code{\%*\%} \cr
#'  \code{MARKp="!"}  for \code{\%p\%} \cr
#'  \code{MARKm="#"}  for \code{\%m\%} \cr
#'  \code{MARKt="$"}  for \code{\%t\%} \cr
#'  \code{MARKT="&"}  for \code{\%T\%} \cr
#'  \code{MARKK="~"}  for \code{\%@@\%} \cr
#'  \cr
#'  MARK parameter has the following format: \cr
#'   MARK=' MARKp="p"; MARKm="m" '
#'
#' @return
#' Expression with the spacial operators recovered.
#'
#' @examples
#' exp="(A %p% B %m% C) %t% (D %T% E) %@@%  F %*% G"
#' mark_soper( exp )
#' unmark_soper( exp )
#'
#' @export
#'

unmark_soper <- function( exp, MARK=NULL, debug=0 ){
 # unmark special operators
 # Shin-ichi Mayekawa
 # 20150919
 #
 # MARK='MARKp="p"; MARKm="m"'
 #

 # mark
 MARKs="`" # for %*%
 MARKp="!" # for %p%
 MARKm="#" # for %m%
 MARKt="$" # for %t%
 MARKT="&" # for %T%
 MARKK="~" # for %@%


 # change mark if necessary
 if( !is.null(MARK) ){
  eval(parse(text=MARK))
 }

 if( debug ) Print( MARKp,MARKm,MARKt,MARKT,MARKK )

 # replace
 exp[]=gsub( MARKs, "%*%", exp, fixed=1 )
 exp[]=gsub( MARKp, "%p%", exp, fixed=1 )
 exp[]=gsub( MARKm, "%m%", exp, fixed=1 )
 exp[]=gsub( MARKt, "%t%", exp, fixed=1 )
 exp[]=gsub( MARKT, "%T%", exp, fixed=1 )
 exp[]=gsub( MARKK, "%@%", exp, fixed=1 )

 return( exp )

} # end of unmark_soper




