#' Symbolic Kronecker Product
#'
#' @param ...  Symbolic matrices or vectors to be multiplied.
#' @param simplify = 1 to simplify the result
#' @param add_paren = 1 to enclose the left matrix by parens
#' @param ra = 1 to remove the SMdim attributes
#' @param remove_attributes same as the ra parameter above.
#' @param debug = 1 to print intermediate result
#'
#' @return
#' A nrow(A)*nrow(B) x ncol(A)*ncol(B) matrix
#'
#' @details
#' If A or B is a vector, it will be converted to a column vector. \cr
#' With ra=0, the result can be printed with separator by \code{Printb}.
#'
#' Do NOT use \code{t} function to transpose a Kronecker product. \cr
#' It may destroy the SMdim attributes and causes error in \code{Printb}.
#'
#'
#' @examples
#' # symbolic example
#' A <- demomat(2,2,root="a"); B <- demomat(3,3,root="b")
#' AkB=matKTimes(A,B)
#' Printb(AkB)
#'
#' Printb(matKTimes(diag(2),A))
#' Printb(matKTimes(1:2,A))
#' Printb(matKTimes(matrix(1:2,1),A))
#' Printb(matKTimes(A,1:2))
#' Printb(matKTimes(A,matrix(1:2,1)))
#'
#' tAktB=matKTimes(t(A),t(B))
#' Printb(tAktB)
#' Printb(Minus(transp(AkB),tAktB))
#' # Printb(t(AktB))
#'
#' @export
#'

matKTimes <- function( ..., add_paren=0, simplify=0, ra=0
                       , remove_attributes=ra, debug=0 ){
 # symbolic Kronecker product
 # Shin-ichi Mayekawa
 # 20150908
 # supermat and getSM removed: 20200525
 # ra parameter: 20200525
 # error for fullsymb attributes: 20200608
 #

 # analyse ...
 name=name <- as.character(substitute(list(...)))[-1L]
 val=list(...)
 locnull=which(unlist(lapply(val,is.null)))
 if( length(locnull) > 0 ){
  name=name[-locnull]
  val=val[-locnull]
 }
 narg=length(val)

 if( debug ) Print(narg,name,val)

 attc=unlist( lapply( val, attributes ) )
 # Print(att,attc)
 if( any(names(attc) == "fullsymb") ){
   cat("\nerror1:(matKTimes) Cannot use fuly symbolic matrices.\n")
   return()
 } 
 
 if( !is.null(ra) ) remove_attributes=ra

 if( narg >= 3 ){
  # recursive call to matTimes
  C=Recall( val[[1]], val[[2]], simplify=simplify, add_paren=add_paren
            , debug=debug )
  for( i in 3:narg ){
   C=Recall( C, val[[i]], simplify=simplify, add_paren=add_paren
             , debug=debug )
  }
  return( C )
 }


 ######################################################################
 # Simple case starts here.
 A=val[[1]]; B=val[[2]]

 # compress all blanks
 A[]=gsub( " +", "", A )
 B[]=gsub( " +", "", B )

 # shape
 if( !is.matrix(A) ) A=matrix(A,,1)
 if( !is.matrix(B) ) B=matrix(B,,1)

 # size of matrix
 nr1=nrow(A); nc1=ncol(A)
 nr2=nrow(B); nc2=ncol(B)

 if( debug ) Print(nr1,nc1,nr2,nc2, A, B)

 # error
 if( !is.matrix(A)  |  !is.matrix(B) ){
  cat("\n error1:(matKTimes) Input must be matrices. \n")
  return()
 }

 # add paren to A?
 if( add_paren ) A=add_paren(A)

 C=supermat( rep(nr2,nr1), rep(nc2,nc1) )
 for( i in 1:nr1){
  for( j in 1:nc1 ){
   C[ getSM(C,i,j,1)[[1]], getSM(C,i,j,1)[[2]] ]=Times( A[i,j], B, simplify=0 )
  }
 }

 if( remove_attributes ) attributes(C)$SMdim=NULL
 return( C )

} # end of matKTimes


