#' Replace the Elements of a Symbolic Matrix
#'
#'
#' @param exp A symbolic matrix
#' @param ... Replacement of the form pat1=pat2 \cr
#' where the size of pat1 must be equal to pat2.
#' @param values A character string consisting of several replacements
#' separated by semicolons.
#' @param fullsymb = 1 to treat \code{exp} as a full symbolic expression.\cr
#' This has priority over the attributes of  \code{exp}.
#' @param ntogoback n parameter of sys.parent function
#' @param print = 2 to show the details of the replacement
#' = 1 to print the result
#' @param debug = 1 to print the additional info
#'
#' @return exp A symbolic expression of the same size as exp
#' with the desired replacement.
#'
#' @details
#'
#' This function replaces the elements of pat1 matrix included in \code{exp}
#' with the corresponding elements of pat2 matrix. \cr
#' A symbolic matrix containing the elements of symbolic matrix \code{pat1}
#' can be converted  to a numeric one by specifying a numeric \code{pat2} matrix.
#' \cr
#'
#' Consider the following example:
#' \preformatted{
#'   A <- demomat(3,2,root="a"); An <- demomat(3,2)
#'   B <- demomat(2,3,root="b"); Bn <- demomat(2,3) + 100
#'   AB <- matTimes(A,B)
#'   ABn <- matReplace( AB, A=An, B=Bn )
#' }
#' Here, the elements of AB matrix are symbolic, e.g., \cr
#'  \code{AB[1,2]="a11*b12+a12*b22"} \cr
#' and we would like to replace the i-j element of \code{A} matrix,
#' \code{aij}, with the corresponding i-j element of \code{An} matrix,
#' 10*i+j, and the i-j elements of \code{B} matrix with the corresponding
#' element of \code{Bn} matrix, 100+10*i+j, and so on. \cr
#' The finall result is in \code{ABn} whose 1-2 element, for example, is \cr
#' \code{"11*112+12*122"}.
#'
#' When \code{exp} is a fully symbolic matrix as seen in the last example below,
#' \code{exp} contain an expression consisting of the matrices which appear
#'  to the left of equal signs as \code{pat1} in the replacement specification
#'  such as \code{A11}, \code{A12}. \cr
#' In this case, those matrices will be replaced by the corresponding
#' matrices to the right of equal sign of the specification, \code{pat2}.
#'
#' When \code{exp} is a fully symbolic expression w/o fullsymb attribute,
#' specifying fullsymb=1 will treat the expression as fully symbolic one.
#' \cr
#'
#' Details of semi symbolic case. \cr
#'
#' If ... consists of \code{A=An, B=Bn}  \cr
#' the following code: \cr
#' \preformatted{
#'   atd <- analyze_three_dots(...)
#'   pat2 <- atd$val
#'   name <- as.list(atd$name2)
#'   names(name) <- atd$name
#' }
#' results in the following: \cr
#' \preformatted{
#'  pat2: list consisting of the value of \code{An} and \code{Bn}
#'         in the parent environment
#'  name: list of characters "An" and "Bn"
#'  names(name): vector consisting of "A" and "B"
#'  }
#' However, the value of \code{A} and \code{B}
#'  in the parent environment is missing. \cr
#'
#' For example, ... argument of the above example
#' will be decomposed into the following lists of length two: \cr
#' \preformatted{
#' name:
#' $A
#' "An"
#'
#' $B
#' "Bn"
#'
#'
#' val:
#' $A
#'    c1 c2
#' r1 11 12
#' r2 21 22
#' r3 31 32
#'
#' $B
#'    c1 c2 c3
#' r1 11 12 13
#' r2 21 22 23
#' }
#' \cr
#'
#' Since \code{exp} consists of the elements of \code{A} and \code{B}
#' defined in the parent environment, in order to replace those by
#' the elements of \code{An} and \code{Bn}, we must retrieve the
#' values of f \code{A} and \code{B} from the parent environment,
#'
#'
#' The replacement can be given in two ways. \cr
#' In the first method, in order to replace the elements of A matrix by the
#' corresponding elements of An matrix, and, B by Bn, \cr
#' we write \cr
#'  \code{ matReplace( exp, A=An, B=Bn ) } \cr
#' Another way is to write \cr
#'  \code{ matReplace( exp, values="A=An; B=Bn" ) } \cr
#' The second method is easy to use when there are many similar replacements.
#' \cr
#' The second method has the proprity over the first method. \cr
#'
#' If the replacements such as values="a=1; b=2;..." are given, \cr
#' and the values of a, b etc in the parent environment are scalars,
#' or not defined, \cr
#' the occurrence of literal "a" is replaced by 1, "b" by 2, and so on.
#'
#'
#' @examples
#'
#' # matrix replacement
#' A <- demomat(3,2,root="a"); An <- demomat(3,2)
#' B <- demomat(2,3,root="b"); Bn <- demomat(2,3)
#' AB <- matTimes(A,B)
#' ABn <- matReplace( AB, A=An, B=Bn, print=2 )
#' Print( Eval(ABn)-An%*%Bn )
#'
#' val <- "A=An; B=Bn"
#' ABn2 <- matReplace(AB, values=val)
#' Print( Eval(ABn2)-An%*%Bn )
#'
#' # scalar replacement
#' val <- "a=1; b=2; c=3; d=4; e=5"
#' exp <- "(a+b*c-d)/e"
#' matReplace(exp,values=val)
#'
#' sexp <- "s( a[i,j]*v[k], {k,1,ncol(A)})"
#' matReplace(sexp,values="ncol(A)=2")
#'
#'
#' # Below gives unexpected result.
#' matReplace( "cos(theta)*sin(a)", a=1, theta=0 )
#' matReplace( "cos(theta)*sin(a)", values="a=1; theta=0" )
#' # This works as expected.
#' matReplace( "cos(theta)*sin(a)", theta=0, a=1 )
#' matReplace( "cos(theta)*sin(a)", values="theta=0; a=1" )
#'
#' # Treating exp as a fully symbolic expression with fullsymb=1.
#' matReplace("A", A=demomat(3,3), fullsymb=1)
#' matReplace("matTimes(t(A),B)"
#'      , A=demomat(2,2, root="a"), B=demomat(2,2, root="b"), fullsymb=1)
#'
#'
#' \dontrun{
#' library(lazy.sumbolic2)
#'
#' # Fully symbolic matrix
#' # define a fully symbolic matrix A
#' A <- demomat( 2,2, root="A", fullsymb=1 )
#' tA <- tp(A)
#' expr=matTimes(tA,A)
#'
#' A11s <- demomat(2,2,root="a11_")
#' A21s <- demomat(2,2,root="a21_")
#' A12s <- demomat(2,2,root="a12_")
#' A22s <- demomat(2,2,root="a22_")
#'
#' valuelist <- "A11=A11s; A21=A21s; A12=A12s; A22=A22s"
#' matReplace( expr, values=valuelist )
#' matReplace( expr, A11=A11s, A21=A21s, A12=A12s, A22=A22s )
#'
#' }
#'
#' @export
#'

matReplace <- function( ..exp, ..., values=NULL
                        , fullsymb=NULL, ntogoback=1, print=0, debug=0 ){
 # replace pat1 by pat2 in exp
 # Shin-ichi Mayekawa
 # 20150821dnc,22
 # ntogoback: 20150907
 # values: 20150918
 # scalar: 20150919
 # bugfix for values: 20150924
 # shape etc: 20150924
 # non strict check of the sizes: 20150926
 # fixed option: 20151006
 # cannot have #' in main body: 20171007
 # roxygen2 bad example: 20200516
 # exp -> ..exp to avoid partial matching: 20200608,09
 # new analyze_three_dots: 20200610
 # fully symbolic: 20200610
 # fullsymb option: 20200630
 # modify_math_operators: 20250315
 #
 #  Note:
 #
 #  If the i-th element of ... contains  pat1=pat2,
 #  when ... is decomposed into name and val by
 #    name <- as.list(substitute(list(...)))[-1L]
 #    val <- list(...)
 #  name[[i]] will contain "pat2" and val[[i]] will contain the value of pat2,
 #  and names(name) will contain "pat1".
 #  That is, the value of pat1 is missing.
 #  For example, ... argument of
 #    A=demomat(3,2,root="a"); An=demomat(3,2)
 #    B=demomat(2,3,root="b"); Bn=demomat(2,3)
 #    matReplace( exp, A=An, B=Bn )
 #  will be decomposed into the following lists of length two:
 #
 #  name:
 #  $A
 #  An
 #
 #  $B
 #  Bn
 #
 #
 #  val:
 #  $A
 #     c1 c2
 #  r1 11 12
 #  r2 21 22
 #  r3 31 32
 #
 #  $B
 #     c1 c2 c3
 #  r1 11 12 13
 #  r2 21 22 23
 #
 #
 #  Therefore, the value of A, which is the A matrix,
 #  is missing, and we must retrived the original value of A
 #  in the parent frame from which matReplace function was executed.
 #  In order to do so, use get function with pos=parent.frame(n=1).
 #
 #
 #

 # modified operators
 modify_math_operators( where="here", print=0 )

 # original
 exp=..exp
 exp0=exp

 atd=analyze_three_dots(...)
 pat2=atd$val
 name=as.list(atd$name2)
 names(name)=atd$name


 # store ...
 # name <- as.list(substitute(list(...)))[-1L]
 # pat2 <- list(...)


 # parent env
 sysparent=sys.parent(n=ntogoback)
 if( debug ){
  Print(sysparent)
  Print(name,pat2,names(name))
  Print(atd$name,atd$name2, atd$val)
 }



 if( !is.null(values) ){

  # values  given
  values=gsub(" ","",values)
  res=strsplit(values,";")[[1]]
  res=res[res != ""]
  nv=strsplit(res,"=")
  name=NULL; pat2=NULL
  for( i in 1:length(nv) ){
   name=append(name,list(nv[[i]][2]))
   pat2=append(pat2,list(nv[[i]][1]))
  }
  names(name)=unlist(pat2)
  names(pat2)=unlist(pat2)
  nname=length(name)

  if( debug ) Print(name,pat2)
  # get the values of pat2
  for( i in 1:nname ){
   temp=name[[i]]
   if( debug ) Print(i,temp)
   if( exists( temp, where=sys.frame(sys.parent(n=ntogoback)) ) ){
    if( debug ) Print(i,pat2[[i]])
    pat2[[i]]=get(temp,pos=parent.frame(n=ntogoback))
   } # end of exists
   else{
    # Here, temp does not exist in the parent frame.
    # cat("\nerror1:(matReplace) Object ", temp, " does not exist.\n")
    pat2[[i]]=temp
   } # end of does not exist
  } # end of i

 } # end of values


 # This stores "pat1"
 namename=names(name)
 nname=length(namename)

 if( debug ){
  cat("\n(matReplace): name:\n"); print(name)
  cat("\n(matReplace): pat2:\n"); print(pat2)
  cat("\n(matReplace): namename:\n"); print(namename)
  cat("\n(matReplace): nname:\n"); print(nname)
 }

 # dimensions of pat2 elements
 dim2=list()
 for( i in 1:nname ){
  if( is.vector(pat2[[i]]) ) dim2[[i]]=length(pat2[[i]])
  else dim2[[i]]=dim(pat2[[i]])
 } # end of i



 # SMdim and fullsymb attribute
 SMdim=attributes(exp)$SMdim

 # fullsymb option has priority over attributes
 if( is.null(fullsymb) )  fullsymb=attributes(exp)$fullsymb  # NULL or 1
 else fullsymb=2

 #
 # symbolic or fully symbolic
 #
 if( !is.null(fullsymb) ){

   # define the matrices to be used
   for( i in 1:nname ){
     code=paste(namename[i],"=pat2[[i]]",sep="")
     if( debug) Print(i,code)
     eval(parse(text=code))
   }

   # input must be a matrix
   if( !is.matrix(exp) ){
     exp=matrix(exp,,1)
   }

   # arrange exp as a real matrix:
   # If exp=matrix(c("A11", "A21", "A12", "A22"),2,2), convert it to
   # exp2="rbind( cbind(A11,A12), cbind(A21,A22) )"
   # and then evaluate exp.
   #
   exp1=apply(exp,1
              ,function(x){ paste("cbind(",paste(x,collapse=","),")", sep="")})
   exp2=paste("rbind(", paste(exp1,collapse=","),")")

   if( debug ) Print(exp1,"/", exp2)

   exp3=eval(parse(text=exp2))

   # inherit the attributes
   if( !is.numeric(exp3) ){
     if( !is.null(fullsymb) && fullsymb == 1 )
       attributes(exp3)$fullsymb=fullsymb
   }
   if( !is.null(SMdim) ){
     attributes(exp3)$SMdim=SMdim
   }


   if( print ){
     cat("\nResult\n")
     Print(exp0, exp3)
   }

   return( exp3 )

 } # end of fully symbolic expression


 #
 #
 # Here, a semi-symbolic expression is given.
 #
 #

 # retrieve the value of pat1 from the parent frame and replace

 pat1=list()
 dim1=list()

 for( i in 1:nname ){

  # retrieve the value
  temp=namename[i]
  if( exists( temp, where=sys.frame(sys.parent(n=ntogoback)) ) ){
   pat1[[i]]=get(temp,pos=parent.frame(n=ntogoback))
   # Here, pat1[[i]] contains the value of namename[i] in the parent frame.
   if( is.vector(pat1[[i]]) ) dim1[[i]]=length(pat1[[i]])
   else if( is.array(pat1[[i]]) ) dim1[[i]]=dim(pat1[[i]])
   else{
    pat1[[i]]=temp; dim1[[i]]=1
   }
   if( all(dim1[[i]] == 1) ) pat1[[i]]=temp  # 20150924
  } # end of exists
  else{
   # Here, temp does not exist in the parent frame.
   # cat("\nerror1:(matReplace) Object ", temp, " does not exist.\n")
   pat1[[i]]=temp; dim1[[i]]=1
  } # end of does not exist


  # replace pat1 with pat2
  if( print >= 2 ){
   cat("\nReplacing pat1(",namename[[i]], ") with pat2(", name[[i]], ")\n")
   Print(pat1[[i]],pat2[[i]])
  }

  if( debug ){
   Print(dim1[[i]], dim2[[i]])
  }

  if( all(pat1[[i]] != "")   &  all( dim1[[i]] <= dim2[[i]] ) ){ # 20150926
   if( length(dim1[[i]]) == 1 ){
    # vec
    for( ii in 1:dim1[[i]] )
     if( regexpr( "[[:alpha:]]", pat1[[i]][ii] )[[1]] > 0 )
      exp=gsub( pat1[[i]][ii], pat2[[i]][ii], exp, fixed=1 )
   }
   else{
    # mat
    for( ii in 1:dim1[[i]][1] )
     for( jj in 1:dim1[[i]][2] ){
      if( regexpr( "[[:alpha:]]", pat1[[i]][ii,jj] )[[1]] > 0 )
       exp=gsub( pat1[[i]][ii,jj], pat2[[i]][ii,jj], exp, fixed=1 )
      # Print(ii,jj,exp)
     }
   }
  }

 } # end of i loop


 if( print ){
  cat("\nResult\n")
  Print(exp0, exp)
 }

 return( exp )

} # end of matReplace


