#' Symbolic Matrix Sweep Operator
#'
#' @param A A square matrix
#' @param loc vector of pivot locations
#' @param simplify = 0 not to use Simplify or Expand \cr
#' = 1 to use Simplify once before exit.\cr
#' = 2 to use Simplify at the bottom of each pivot loop.
#' @param all parameter for Simplify
#' @param func_level parameter for Simplify
#' @param debug = 1 to print details
#'
#' @details
#' Definition of Sweep \cr
#' C <- matSweep( A, p ) means: \cr
#'  c[i,j] <- c[i,j] - c[i,p]*c[p,j]/c[p,p] \cr
#'  c[p,] <- c[p,]/c[p,p] \cr
#'  c[,p] <- - c[,p]/c[p,p] \cr
#'
#' # Inversion of A.  \cr
#' matSweep( A, 1:ncol(A) ) == solve(A) \cr
#' This is equivalent to Inv(A).
#' \cr
#' # Sweep is reversible. \cr
#' matSweep( matSweep( A, loc ), loc ) == A \cr
#'
#' Use matSwp for numeric matrix. \cr
#'
#' @return
#' A matrix of the same size as A.
#'
#' @examples
#'
#' # inversion of 2 x 2 A
#' A <- demomat(2,2,root="a")
#' Ainv <- matSweep(A)
#' AAinv <- matTimes(A,Ainv)
#' Print( Eval(AAinv, A=demomat(2,2)), fuzz=1e-9 )
#' AinvA <- matTimes(Ainv,A)
#' Print( Eval(AinvA, A=demomat(2,2)), fuzz=1e-9 )
#'
#' # inversion of 3 x 3 A
#' A <- demomat(3,3,root="a")
#' An <- demomat(3,3); An[1,3] <- 11  # make sure it it fullrank
#' # An <- demomat(3,3,shape="uppert")
#' Ainv <- matSweep(A)
#' AAinv <- matTimes(A,Ainv)
#' AAinvn <- matReplace(AAinv,A=An)
#' Print( Eval(AAinvn), fuzz=1e-9 )
#' AinvA <- matTimes(Ainv,A)
#' AinvAn <- matReplace(AinvA,A=An)
#' Print( Eval(AinvAn), fuzz=1e-9 )
#'
#'
#' # See if it is reversible.
#' A <- demomat(2,2,root="a")
#' A1 <- matSweep(A,1)
#' A11 <- matSweep(A1,1)
#' Print( Eval(A11,A=demomat(2,2)) )
#'
#' # See if it is reversible.
#' A <- demomat(3,3,root="a")
#' A1 <- matSweep(A,1:2)
#' A11 <- matSweep(A1,1:2)
#' Print( Eval(A11,A=demomat(3,3)) )
#'
#'
#' # Fully Symbolic Partitioned Matrix
#' Ap=demomat(2,2,root="A", fullsymb=1)
#' Print(Ap)
#' invAp=matSweep(Ap)
#' Print(invAp)
#'
#' # numeric checking
#' A11n=matrix(c(1,2,0,2),2,2); A12n=matrix(c(1,0,1,1),2,2)
#' A21n=matrix(c(2,0,1,1),2,2); A22n=matrix(c(1,1,0,2),2,2)
#' Apn=rbind(cbind(A11n,A12n),cbind(A21n,A22n))
#' invApn=solve(Apn)
#' # change inv to solve
#' invAps=gsub("inv","solve", invAp, fixed=1)
#' invApn2=matReplace( invAps, A11=A11n, A12=A12n, A21=A21n, A22=A22n )
#' Print(invApn-invApn2, fmt="9.6")
#'
#' @export
#'

matSweep <- function( A, loc=1:nrow(A), simplify=2
                      , all=0, func_level=0, debug=0 ){
  # matrix sweep operator
  # Shin-ichi Mayekawa
  # 20150823
  # 20150823 Maiko
  # simplify: 20150823,24
  # simplify=1: 20150906
  # error check: 20150908
  # not simplify at all: 20151006
  # default simplify=2: 20151006
  # SMdim and fullsymb attribute: 20200527
  # bugfix: 20200715???????????????????
  # all and func_level of Simplify: 20230624

  # errors
  if( !is.matrix(A) ){
    cat("\n error1:(matSweep) Input must be a matrix! \n")
    return()
  }

  nr=nrow(A); nc=ncol(A)
  if( nr != nc ){
    cat("\n error1:(matSweep) Input must be square! \n")
    return()
  }

  lenp=length(loc)

  # SMdim and fullsymb attribute
  SMdim=attributes(A)$SMdim
  fullsymb=attributes(A)$fullsymb

  if( !is.null(fullsymb) ){

    #
    # Here, A is a fully symbolic matrix.
    #

    C=A
    for( pv in loc[1:lenp] ){

      cpivot=C[pv,pv]

      r1=NULL; if( pv > 1 ) r1=1:(pv-1)
      r2=NULL; if( pv < nc ) r2=(pv+1):nc
      r12=c(r1,r2)

      if( debug ) Print(pv,r12, cpivot)

      cpvrow=C[pv,]; cpvcol=C[,pv]

      for( j in r12 ){
        for( k in r12 ){
          # c(jk)=c(jk)-c(jp)*c(pk)/c(pp)
          C[j,k]=
          paste( C[j,k], "-(", C[j,pv], ")%*%inv(", cpivot, ")%*%(", C[pv,k]
                 , ")", sep="" )
        }
      }

      C[pv,]=paste( "inv(", cpivot, ")%*%(", cpvrow, ")", sep="" )
      C[,pv]=paste( "-(", cpvcol, ")", "%*%inv(", cpivot, ")" , sep="" )
      C[pv,pv]=paste( "inv(", cpivot, ")", sep="" )

      if( simplify >= 2 ){
        C=Expand(C)
        C=Simplify(C, func_level=func_level, all=all) # 20151006, 20230623
      }

    } # end of pivot loop


    # final treatment
    if( simplify >=1  ){
      C=Simplify(C, func_level=func_level, all=all)
    }

    # inherit the attributes
    attributes(C)$fullsymb=fullsymb
    if( !is.null(SMdim) ){
      attributes(C)$SMdim=SMdim
    }

    return( C )

  } # end of full symbolic A matrix



  #
  # Here, A is a semi-symbolic matrix.: no fullsymb attribute
  #

  C=A
  for( pv in loc[1:lenp] ){

    cpivot=C[pv,pv]

    r1=NULL; if( pv > 1 ) r1=1:(pv-1)
    r2=NULL; if( pv < nc ) r2=(pv+1):nc
    r12=c(r1,r2)

    if( debug ) Print(pv,r12, cpivot)

    cpvrow=C[pv,]; cpvcol=C[,pv]

    for( j in r12 ){
      for( k in r12 ){
        # c(jk)=c(jk)-c(jp)*c(pk)/c(pp)
        C[j,k]=paste( C[j,k], "-(", C[j,pv], ")*(", C[pv,k], ")/(", cpivot, ")"
                      , sep="" )
      }
    }

    C[pv,]=paste( "(", cpvrow, ")", "/(", cpivot, ")" , sep="" )
    C[,pv]=paste( "-(", cpvcol, ")", "/(", cpivot, ")" , sep="" )
    C[pv,pv]=paste( "1/(", cpivot, ")", sep="" )

    if( simplify >= 2 ) C=Simplify(C, func_level=func_level, all=all) # 20151006

  } # end of pivot loop

  # final treatment
  if( simplify >=1  ){
    C=Simplify(C, func_level=func_level, all=all)
  }

  if( !is.null(SMdim) ){
    attributes(C)$SMdim=SMdim
  }

  return(C)

} # end of matSweep


