#' Symbolic Multiplication of Two or More Matrices
#'
#' @param ...  Symbolic matrices or vectors to be multiplied.
#' @param method = 2 to use educational method
#' @param padd = 0 not to add blank to both sides of + or -
#' @param simplify = 0 not to simplify the result
#' @param debug = 1 to print intermediate result
#'
#' @details
#' In order to be able to multiply A from the left of  B,
#' ncol(A) == nrow(B). \cr
#' A matrix and a vector can be multiplied as long as the condition above
#' holds. \cr
#' If one argument is a vector, it will be promoted to either a row or column
#'  vector (matrix) to make the two arguments conformable. \cr
#' If both are vectors of the same length, it will return the inner product.\cr
#' A scalar can be multiplied to a matrix \cr
#' \cr
#' Numeric matrices will be treated as character ones.
#' \cr
#' When you want to chage the sign of all the elements of a matrix, use \cr
#' A1 <- matMinus( 0, A )   or   A1 <- matTimes( -1, A )
#'
#'
#' If the arguments are  super matrices with SMdim attributes,
#' the final result will have an appropriate SMdim attribute.
#'
#' Can handle fully symbolic partitioned matrices.
#'
#' @return A character matrix
#'
#' @examples
#' A <- demomat(3,2,root="a")
#' B <- demomat(2,3,root="b")
#' matTimes( A, B )
#' C <- demomat(2,3,root="c")
#' matTimes( A, Plus(B,C) )
#'
#' A[2,2] <- paste("-",A[2,2],sep="")
#' B[1,1] <- 0
#' matTimes( A, B, simplify=0, method=2 )
#' matTimes( A, B, simplify=0 )
#' matTimes( A, B )
#' matTimes( A, B, padd=0 )
#' matTimes( A, -2 )
#' matTimes( matTimes(-1,A), B )
#' matTimes( A, B[,2] )
#' matTimes( A[1,], B )
#' matTimes( A, 1:2 )
#'
#' @export
#'

matTimes <- function( ..., method=1, padd=0, simplify=1, debug=0 ){
 # symbolic matrix multiplication
 # Shin-ichi Mayekawa
 # 20150725,26
 # zero: 20150803
 # names: 20150803
 # simplify: 20150803
 # padd: 20150803
 # bugfix for nr2==1 & nc2==1 & ismatA: 20150804
 # simplify option: 20150817
 # bugfix: 20150817
 # times renamed as times2: 20150823
 # nterms in times2: 20150824
 # ...: 20150904
 # simplify=1: 20150906
 # simplify only once: 20151006
 # SMdim: 20200525
 # symbol for partitioned matrix: 20200527
 #
 #  matrix x matrix
 #  vector x matrix
 #  scalar x matrix
 #
 #  vector x vector
 #  scalar x vector
 #
 #  scalar x scalar
 #


 # analyse ...
 name=name <- as.character(substitute(list(...)))[-1L]
 val=list(...)
 locnull=which(unlist(lapply(val,is.null)))
 if( length(locnull) > 0 ){
  name=name[-locnull]
  val=val[-locnull]
 }
 narg=length(val)

 if( debug ) Print(narg,name,val)

 if( narg >= 3 ){
  # recursive call to matTimes
  C=Recall( val[[1]], val[[2]], simplify=simplify, debug=debug )
  for( i in 3:narg ){
   C=Recall( C, val[[i]], simplify=simplify, debug=debug )
  }
  return( C )
 }


 ######################################################################
 # Simple case starts here.
 A=val[[1]]; B=val[[2]]

 # check if A or B are partitioned matrix
 fsA=attributes(A)$fullsymb
 fsB=attributes(B)$fullsymb
 if( !is.null(fsA) | !is.null(fsB) ){
   symbol="%*%"
   resfs=1
 }
 else{
   symbol="*"
   resfs=0
 }

 # SMdim attribute
 SMdim1=attributes(A)$SMdim
 SMdim2=attributes(B)$SMdim
 SMdim=list(nrows=SMdim1$nrows,ncols=SMdim2$ncols)
 if( debug ) Print( SMdim1,SMdim2,SMdim )

 # trim
 A=trim(A,all=1)
 B=trim(B,all=1)

 # genuine R vector?
 isvecA=0; if( is.vector(A) ) isvecA=1
 isvecB=0; if( is.vector(B) ) isvecB=1

 if( isvecA & isvecB & length(A) != length(B) ){
  cat("\n error1:(matTimes) Tow vectors of different sizes"
       , " cannot be multiplied.\n")
  return()
 }


 # size of matrix
 nr1=nrow(A); nc1=ncol(A)
 nr2=nrow(B); nc2=ncol(B)
 len1=0; len2=0
 ismatA=1; ismatB=1

 # First, make non matrix input a scalar
 if( is.null(nr1) ){
  # A is a vector
  ismatA=0
  nr1=1; nc1=1
  len1=length(A)
 }
 if( is.null(nr2) ){
  # B is a vector
  ismatB=0
  nr2=1; nc2=1
  len2=length(B)
 }

 # make A a row vec when B is a matrix
 if( ismatB   &  len1 == nr2 ){
  A=matrix(A,1)
  nc1=len1
 }
 # make B a col vec when A is a matrix
 if( ismatA  &  len2 == nc1 ){
  B=matrix(B,,1)
  nr2=len2
 }

 # Both A and B are vectors of the same size -> scalar product
 if( is.vector(A)*is.vector(B)*(length(A)>1)*(length(B)>1)*
     (length(A)==length(B))  ){
  A=matrix(A,1); B=matrix(B,,1)
  nc1=len1; nr2=len2
 }

 # Both A and B are scalar
 if( nr1*nr2*nc1*nc2 == 1 ){
  A=matrix(A,1,1); B=matrix(B,1,1)
 }


 if( debug ) Print(nr1,nc1,nr2,nc2, len1,len2, ismatA, ismatB)

 rn=rownames(A); cn=colnames(B)


 if( nr1 == 1  &  nc1 == 1  &  ismatB ){

  # scalar A times B matrix
  B=matrix(B,nr2,nc2)
  ff=c(as.character(A))
  C=matrix("",nr2,nc2)
  for( i in 1:nr2 )
   for( j in 1:nc2 ){
    C[i,j]=times2(ff, B[i,j], prefix="", symbol=symbol)
   }

 }
 else if( nr2 == 1  &  nc2 == 1  &  ismatA ){

  # A matrix times sclar B
  A=matrix(A,nr1,nc1)
  ff=c(as.character(B))
  C=matrix("",nr1,nc1)
  for( i in 1:nr1 )
   for( j in 1:nc1 ){
    C[i,j]=times2(A[i,j],ff, prefix="", symbol=symbol)
   }

 }
 else{

  if( nc1 != nr2 ){
   cat("\n error1:(matTimes) The sizes of matrices/vectors do not conform.\n")
   return()
  }

  C=matrix("",nr1,nc2)
  for( i in 1:nr1 ){
   for( j in 1:nc2 ){
    temp=times2( A[i,1], B[1,j], prefix="", method=method, symbol=symbol )
    if( nc1 > 1 ){
     for( k in 2:nc1 ){
      temp=paste( temp, times2( A[i,k], B[k,j], method=method, symbol=symbol ) )
     }
    } # k
    C[i,j]=temp
   } # j
  } # i

 }

 # names
 dimnames(C)=list(rn,cn)

 # final treatment
 if( simplify ==1  ){
  C=Simplify(C)
 }

 if( padd == 1){
  C=gsub("\\+", " \\+ ", C)
  C=gsub("-", " - ", C)
 }

 # attributes of the result
 if( !is.null(SMdim$nrows) ) attributes(C)$SMdim=SMdim
 if( resfs == 1 ) attributes(C)$fullsymb=1

 return( C )

} # end of matTimes



times2 <- function( a, b, prefix="+", method=1, symbol="*" ){
 # scalar a x scalar b
 # Shin-ichi Mayekawa
 # 20150724
 # renamed as times2: 20150823
 # nterms: 20150824
 # symbol: 20200527
 #
 # This should be called from matTimes or Times
 # This will return
 #   prefix a * b
 #

 nta=nterms(a); ntb=nterms(b)
 if( nta > 1 ) a=paste("(",a,")", collapse="")
 if( ntb > 1 ) b=paste("(",b,")", collapse="")

 if( method == 1 ){
  # dafault elaborated method

  # 0
  if( a == "0" |  b == "0" ) return( paste(prefix,"0" ) )

  # remove signs
  sa=1; sb=1
  if( a1 <- substr(a,1,1) %in% c("+","-") ){
   if( a1 == "+" ) sa=1
   else sa=-1
   a=substr(a,2,nchar(a))
  }
  if( b1 <- substr(b,1,1) %in% c("+","-") ){
   sb=b1
   if( b1 == "+" ) sb=1
   else sb=-1
   b=substr(b,2,nchar(b))
  }

  # prefix
  pp="+"
  if( sa*sb == -1 ){
   if( prefix %in% c("","+") ) pp="-"
   else pp="+"
  }
  if( prefix == ""  &  pp == "+" ) pp=NULL

  # res

  res=paste( pp, paste( a, symbol, b, sep="" ) )

  # 1 x b  or a x 1
  if( a == "1" ) res=paste( pp, b )
  if( b == "1" ) res=paste( pp, a )
  if( a == "1" & b == "1" ) res=paste( pp, 1 )

  #if( substr(res,1,1) == " " ) res=substr(res,2,nchar(res))
  res=trim(res,all=1)

  return( res )

 }
 else if( method == 2 ){
  # educational method

  sa=1; sb=1
  if( a1 <- substr(a,1,1) %in% c("+","-") ){
   a=paste( "(", a, ")", sep="")
  }
  if( b1 <- substr(b,1,1) %in% c("+","-") ){
   b=paste( "(", b, ")", sep="")
  }

  pp="+"
  if( prefix == "" ) pp=NULL

  res=paste( pp, paste( a, symbol, b, sep="" ) )

  if( substr(res,1,1) == " " ) res=substr(res,2,nchar(res))

  return( res )

 }

} # end of times2


