#' Modify Mathematical Operators for Symbolic Math
#'
#' Japanese help file: \link[lazy.symbolic]{modify_math_operators_JPH} \cr
#'
#'
#' @param where The environment name. (parent or global)
#' @param print = 0 to skip the message
#'
#' @details
#' The following native mathematical operators will be modified
#' so that they can handle symbolic objects.
#' \preformatted{
#'  +     Addition
#'  -     Subtraction
#'  *     Scalar Multiplication
#'  /     Scalar Division
#'  \%*\%   Matrix Multiplication
#'  t   transposition
#' }
#'
#' These operators will be available in the environment specified by
#' \code{where}. The default is \code{parent.frame()}, the calling environment.
#'
#'   To remove these operators, submit the following code: \cr
#'       rm( `+`, `-`, `*`, `\%*\%`, `/` , t )  .
#'
#' In addition, we have \cr
#' \preformatted{
#' \%@\%   Kronecker product
#' }
#'
#' @examples
#' # in global environment
#' # remove all
#' rm(list=ls())
#'
#' # Create symbolic matrices
#' A=demomat(3,2,root="a")
#' B=demomat(3,2,root="b")
#'
#' # Below causes error.
#' # A+B
#'
#' # Load modified operators
#' modify_math_operators()
#' ls()
#'
#' # This works!
#' A+B
#'
#' # Remove operators
#' rm(`+`, `-`, `*`, `%*%`,  `/`, t)
#'
#'
#' # In a function environment
#' func1 <- function(){
#'  printm("in function before call", ls())
#'  modify_math_operators()
#'  printm("in function after call", ls())
#'  printm(A+B)
#' }
#'
#' printm("in the calling environment before call", ls())
#' func1()
#' printm("in the calling environment after", ls())
#'
#'
#' @export
#'

modify_math_operators <- function( where="parent", print=1 ){
 # modify mathematical operators
 # Shin-ichi Mayekawa
 # 20250315
 # default="parent": 20250316
 #

 `+` <- function( A, B=NULL ){
  # + operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   if( is.null(B) ) return( Plus(0,A))
   else return( Plus(A,B) )
  }
  else{
   if( is.null(B) )  return(  .Primitive("+")(0,A) )
   else return(  .Primitive("+")(A,B) )
  }
 } # end of +



 `-` <- function( A, B=NULL ){
  # - operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   if( is.null(B) ) return( Minus(0,A))
   else return( Minus(A,B) )
  }
  else{
   if( is.null(B) )  return(  .Primitive("-")(0,A) )
   else return(  .Primitive("-")(A,B) )
  }
 } # end of -



 `*` <- function( A, B ){
  # * operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   return( Times(A,B) )
  }
  else{
   return(   .Primitive("*")(A, B) )
  }
 } # end of *



 `%*%` <- function( A, B ){
  # %*% operator
  # Shin-ichi Mayekawa
  # 20150925DNC

  if( is.character(A) | is.character(B) ){
   return( matTimes(A,B) )
  }
  else{
   return(   .Primitive("%*%")(A, B) )
  }
 } # end of %*%


 `/` <- function( A, B ){
  # / operator
  # Shin-ichi Mayekawa
  # 20250301

  if( is.character(A) | is.character(B) ){
   return( paste("(", A, ")/(", B, ")", sep="") )
  }
  else{
   return(   .Primitive("/")(A, B) )
  }
 } # end of /


 t <- function( A ){
  # transposition
  # Shin-ichi Mayekawa
  # 20230725dnc
  #

  if( is.data.frame(A) ){
   # A is a data frame
   return( base::t(A) )
  }
  if( is.list(A) || length(dim(A)) > 2 || !is.null(attributes(A)$SMdim) ){
   # A is a list or array, or super mat.
   return( lazy.mat::transp(A) )
  }
  else if( is.null(attributes(A)$fullsymb) ){
   # A is not a list, nor array, nor a super mat, nor a fully symbolic matrix
   return( base::t(A) )
  }
  else{
   # A is a fully symbolic matrix
   return( lazy.symbolic::tp(A) )
  }

 } # end of t




 # put in the specified env.
 values=named_list(`+`, `-`, `*`, `%*%`,  `/`, t)
 opnames=c("+", "-", "*", "%*%",  "/", "t")
 opnames2=paste(paste("`",opnames[-length(opnames)],"`",sep=""),collapse=", ")
 opnames2=c(opnames2,", t")
 if( where == "global" ){
  # generate variables in the global environment
  tmp=mapply( assign, opnames, values, MoreArgs=list(envir=.GlobalEnv) )
 }
 else{
  # in calling env
  # callingfunc1 = as.list(sys.call(-1))[[1]]
  # callingfunc2 = as.list(sys.call(-2))
  # if( length(callingfunc2) == 0 ) callingfunc2="global env."
  # else callingfunc2=callingfunc2[[1]]
  mapply( assign, opnames, values, MoreArgs=list(envir=parent.frame()) )
 }

 if( print > 0 ){
  cat( "\n\n**\n The following mathematical operators have been modified \n")
  if( where == "global" ){
   cat( " and made available in the global environment. \n" )
   cat("    ",  opnames, "\n" )
   cat( "  To remove these operators, submit the following code: \n"
        , "     rm(",   opnames2,  ")  .  \n**\n\n")
  }
  else
   cat( " and made available in the following environment. \n" )
   print(parent.frame())
 }

} # end of modify_math_operators



