#' Count the Number of Terms in an Expression
#'
#'
#' @param exp expression vector or matrix
#' @param debug = 1 to print the intermediate result
#'
#' @return
#' A vector or matrix of the same size as the input
#'
#' @details
#' *************************************************************************\cr
#' ********** There is a discrepancy between this and find_terms. **********\cr
#' *************************************************************************\cr
#'
#' @examples
#' nterms( "((a)) - ((a)+(b)) + (a-b) - (x-y)*(((z+1)))" )
#' nterms("-a")
#' nterms("+a")
#' nterms("a+")
#' nterms("(a)")
#' nterms("x*a/(1-y)")
#' nterms("-a+b-0")
#' nterms(Simplify("-a+b-0"))
#' nterms("a %p% b %m% c%t%d%T%e + f%@@%g")
#'
#' @export
#'

nterms <- function( exp, debug=0 ){
 # count the # of terms in an expression
 # Shin-ichi Mayekawa
 # 20150824
 # specialoperators: 20150919
 #

 # mark
 MARK=";"

 # is exp a scalar?
 if( is.vector(exp) ) lenexp=length(exp)
 else lenexp=nrow(exp)*ncol(exp)

 # if not scalar, use recursive call
 if( lenexp > 1  |  is.matrix(exp) ){
  exp[]=unlist(lapply( exp, nterms, debug=debug ))
  return( exp )
 }


 ############# main starts here. #####################################

 # trim exp
 exp=gsub( " ", "", exp )

 # mark special operators
 exp=mark_soper( exp )

 lenexp=nchar(exp)
 MARKs=paste(rep(MARK,lenexp),collapse="")
 exp0=exp

 # find the location of left parens
 temp=unlist(strsplit(exp,""))
 loc1=which( unlist( gregexpr( "\\(", temp ) ) == 1 )
 loc20=which( unlist( gregexpr( "\\)", temp ) ) == 1 )
 # if( debug ) Print(temp,loc1,loc20, printrownames=1)

 np=length(loc1)

 if( np > 0 ){

  # find the matching parens
  loc2=loc1
  for( i in 1:np ){
   loc2[i]=locbalpar( exp, loc1[i] )
  }

  # range=mapply( ":", loc1, loc2 )
  range=cbind(loc1,loc2)
  if( debug ) Print(temp,range, printrownames=1)

  # location of top-level parens
  k=1
  loc=1
  for( i in 1:np ){
   ll=which( range[,1] > range[k,2] )
   if( length(ll) == 0 ) break
   loc=c( loc, k <- ll[1] )
  }
  range1=range[loc,,drop=0]
  np1=nrow(range1)
  if( debug ) Print(np1,loc,range1,range)

  # mark the inside of parens
  for( i in 1:np1 ){
   substr( exp, range1[i,1], range1[i,2] )=
    substr(MARKs,1,range1[i,2]-range1[i,1]+1)
  }

 }

 if( debug ) Print(exp0,"/",exp)

 # count the terms
 if( substr(exp,1,1) %in% c("+","-","!","#")  ) exp=substr(exp,2,nchar(exp))
 nterms=length(strsplit(exp,"[-+!#]")[[1]])

 return( nterms )

} # end of nterms




