#' Express the i-j element of an Matrix Expression using Summation Operators
#'
#' @param exp An full symbplic matrix expression
#' @param row row subscript name
#' @param col column subscript name
#' @param root root of the intermediate subscripts
#' @param sublist a character vector containing the intermediate subscript names
#' @param simple = 0 to use full spec: {subscript, from , to}
#' @param expand = 0 not to expand (A+B) or (A*B) etc.
#' @param print = 1 to print the result
#' @param debug = 1 to print the intermediate result
#'
#' @details
#' The summation of an expression over subscript k is expressed as \cr
#' \code{ s( exp, {k} )}  \cr
#' or \cr
#' \code{ s( exp, {k, from, to} ) }
#' \cr\cr
#' The lowercase object is treated as a column vector. (See examples below.)
#' \cr\cr
#' When row="i" and col="j": \cr
#' \preformatted{
#'  mat2sum( "A \%*\% B" ) -> s(A[i,s1]*B[s1,j], {s1})
#'  mat2sum( "A \%*\% B", root="z" ) -> s(A[ij,z1]*B[z1,j], {z1})
#'  mat2sum( "A \%*\% B", simple=0 ) -> s(A[i,s1]*B[s1,j], {s1,1,ncol(A)})
#'  mat2sum( "t(A) \%*\% B" ) -> s(A[s1,i]*B[s1,j], {s1})
#'  mat2sum( "A \%*\% t(B)" ) -> s(A[i,s1]*B[j,s1], {s1})
#'  mat2sum( "A \%*\%Diag(B)\%*\%C" ) -> s(A[i,s1]*B[s1,s1]*C[s1,j],{s1})
#'  mat2sum( "A \%*\% b" ) -> s(A[i,s1]*b[s1],{s1})
#'  mat2sum( "t(a) \%*\% B" ) -> s(a[s1]*B[s1,j],{s1})
#'  }
#'  \cr
#'
#'  Identity matrix, I, and vec function are not recognized.
#'
#'  Diag or diag function will be recognized.
#'
#'  Use parentheses to mark Hadamar and Kronecker products.
#'  \cr
#'  \cr
#'  The resulting summation expression can be expanded and evaluated by
#'  sumEval function.  (lazy.symbolic2 package is required.)
#'
#'
#' @examples
#' mat2sum( "A %*% B" )
#' mat2sum( "A %*% B", root="z" )
#' mat2sum( "A %*% B", simple=0 )
#' mat2sum( "t(A) %*% B" )
#' mat2sum( "A %*% t(B)" )
#' mat2sum( "A %*%Diag(B)%*%C" )
#' mat2sum( "A %*% b" )
#' mat2sum( "t(a) %*% B" )
#' mat2sum( "a %*% t(b)" )
#' mat2sum( "A%*%(B+C)" )
#' mat2sum( "A%*%(B+C)", expand=0 )
#'
#' exp="t(A)%*%t(B)%*%C - D%*%t(E)"
#' mat2sum( exp )
#' mat2sum( exp, root="k" )
#' mat2sum( exp, sublist=letters[11:20] )
#' exp="A%*%diag(B)%*%C - diag(D)%*%E%*%diag(F)%*%G + diag(H)"
#' mat2sum( exp )
#'
#'
#' sexp0=mat2sum( "A%*%(B+C)", expand=0 )
#' sexp=mat2sum( "A%*%(B+C)" )
#' sexpe=Expand(c(sexp0,sexp))
#' Print(sexp0,sexp,sexpe)
#'
#' @export
#'

mat2sum <- function( exp, row="i", col="j", root="s", sublist=NA
                   , simple=1, expand=1, print=0, debug=0 ){
 # matrix to summation
 # Shin-ichi Mayekawa
 # 20150930,1001
 # expand: 20151002
 # lowercase as vector: 20151002
 # roxgen2: 20220412
 # bugfix for t(a)%*%diag(D): 20240707
 #


 mattype <- function( mat ){
  # check if mat is transposed or diag or vec
  # Shin-ichi Mayekawa
  # 20150930
  #

  type=""
  nfj=nchar(mat)

  if( nfj >= 8  & tolower(substr(mat,1,6)) == "t(vec("
           &  substr(mat,nfj,nfj) == ")" ){
   type="tvec"
   mat=substr(mat,7,nfj-1)
  }
  else if( nfj >= 4  & tolower(substr(mat,1,2)) == "t("
           &  substr(mat,nfj,nfj) == ")" ){
   type="t"
   mat=substr(mat,3,nfj-1)
  }
  else if( nfj >= 7  & tolower(substr(mat,1,5)) == "diag("
           &  substr(mat,nfj,nfj) == ")" ){
   type="diag"
   mat=substr(mat,6,nfj-1)
  }
  else if( nfj >= 6  & tolower(substr(mat,1,4)) == "vec("
           &  substr(mat,nfj,nfj) == ")" ){
   type="vec"
   mat=substr(mat,5,nfj-1)
  }

  return( named_list( type, mat ) )

 } # end of mattype



 # compress exp
 exp00=exp
 exp=gsub( " ", "", exp )

 temp=find_terms( exp )
 nterms=temp$nterms
 term=temp$term
 oper=temp$oper

 if( debug ) Print(nterms, oper, term )

 # for each term
 for( i in 1:nterms ){

  termi=term[i]

  # find factor
  temp2=find_factors( termi )
  nfactors=temp2$nfactors
  fact=temp2$factor
  foper=temp2$oper

  if( debug ) Print( i, foper, fact )

  # Hadamar and Kronecker product: %@%  and *
  removed=0
  jj=1
  fact1=fact[1]; foper1=foper[1]
  for( j in 2:nfactors ){
   if( debug ) Print(j,jj,foper[j],fact[j],fact1[jj],removed)
   if( foper[j] %in% c("%@%","*") ){
    if( debug ) Print("found", jj,fact1[jj],fact[j])
    removed=c(removed,j)
    fact1[jj]=paste( "(", fact1[jj], foper[j], fact[j], ")", sep="" )
#    if( removed[length(removed)] != j-1 ) jj=jj+1
   }
   else{
    jj=jj+1
    fact1[jj]=fact[j]
    foper1[jj]=foper[j]
   }
  }

  if( length(fact) > length(fact1) ){
   if( debug ) Print( removed, foper1, fact1 )
   foper=foper1; fact=fact1
   nfactors=length(fact)
   if( nfactors == 1 ){
    termi=paste(foper,fact,sep="")
   }
  }

  if( debug ) Print(  nfactors, foper, fact )

  if( nfactors == 1 ){
   temp=mattype( termi )
   # just check if it is transposed or not.
   if( temp$type == "t" ){
    term[i]=paste( temp$mat, "[", col, ",", row, "]", sep="" )
   }
   else{
    term[i]=paste( termi, "[", row, ",", col, "]", sep="" )
   }
   next # to next i
  } # end of nfactors = 1


  # last diag
  lastdiag=0
  temp=mattype( fact[nfactors] )
  if( regexpr("^[a-z]+([a-z]|[_.]|[0-9])*$",temp$mat)[[1]] > 0 ) vec=1
  else vec=0
  if( temp$type == "diag"  |  (temp$type == "t"  &  vec ) ){
   lastdiag=1
   lastmat=temp$mat
   nfactors=nfactors-1
  }

  # make sublist long enough
  if( all(!is.na(sublist)) ){
   if( length(sublist) < nfactors ){
    sublist2=character(nfactors)
    sublist2[1:length(sublist)]=sublist
    sublist2[(length(sublist)+1):nfactors]=
     paste(root,(length(sublist)+1):nfactors, sep="")
    sublist=sublist2
   }
 }

  subnum=1   # subscript number
  sub0=row   # the last subscript
  res1=NULL; res2=NULL

  # factors (matrices) in the middle
  if( nfactors >= 2 ){
   for( j in 1:(nfactors-1) ){
    fj=fact[j]; nfj=nchar(fj)

    if( all(is.na(sublist)) ) sub1=paste(root,subnum,sep="")
    else sub1=sublist[subnum]

    temp=mattype( fj )
    if( regexpr("^[a-z]+([a-z]|[_.]|[0-9])*$",temp$mat)[[1]] > 0 ) vec=1
    else vec=0
    mat=paste( fj, "[", sub0, ",", sub1, "]", sep="" )
    if( debug ) Print(j,fj,vec,temp$type,temp$mat)
    if( vec ) mat=paste( fj, "[", sub0, "]", sep="" )
    if( temp$type == "diag" ){
     if( !vec )
      res1=paste( res1, temp$mat, "[", sub0, ",",  sub0, "]*", sep="" )
     else if( vec ) res1=paste( res1, temp$mat, "[", sub0, "]*", sep="" )
     next
     # no summation for diag
    }
#    if( temp$type == "t"  &  vec ){
#     res1=paste( res1, temp$mat, "[", sub0, "]*", sep="" )
#     next
#    }
    if( temp$type == "t" ){
     # switch subscript if transposed
     fj=temp$mat
     mat=paste( fj, "[", sub1, ",", sub0, "]", sep="" )
     if( vec ) mat=paste( fj, "[", sub1, "]", sep="" )
    }
    # top part:       s( body *    <------------------------ * is assumed here.
    res1=paste( res1, "s(", mat, "*", sep="" )
    # bottom part:    , {sub, from, to} )
    spec=sub1
    if( !simple ){
     if( temp$type != "t" )
      spec=paste( spec, ",1,ncol(", fj, ")", sep="")
     else if( temp$type == "t" )
      spec=paste( spec, ",1,nrow(", fj, ")", sep="")
    }
    res2=paste( ",{", spec, "})", res2, sep="" )
    sub0=sub1
    subnum=subnum+1
   } # end of j
  }

  # last factor (matrix)
  fj=fact[nfactors]; nfj=nchar(fj)
  temp=mattype( fj )
  if( regexpr("^[a-z]+([a-z]|[_.]|[0-9])*$",temp$mat)[[1]] > 0 ) vec=1
  else vec=0

  if( debug ){ j=nfactors
   Print(j,fj,vec,temp$type,temp$mat, lastdiag)
  }

  mat=paste( fj, "[", sub0, ",", col, "]", sep="" )
  if( vec ) mat=paste( fj, "[", sub0, "]", sep="" )
  if( debug ) Print("*0", mat)
  if( temp$type == "diag" ){
   if( !vec )
    res1=paste( res1, temp$mat, "[", sub0, ",",  sub0, "]", sep="" )
   else if( vec ) res1=paste( res1, temp$mat, "[", sub0, "]", sep="" )
  }
  else if( temp$type == "t"  &  vec ){   #### 20240707
   if( lastdiag != 1 ) res1=paste( res1, temp$mat, "[", sub0, "]*", sep="" )
   else if( lastdiag == 1 ){
    res1=paste( res1, temp$mat, "[", col, "]*", sep="" )  #### 20240707
    mat=paste( lastmat, "[", col,",", col, "]", sep="" )  #### 20240707
    res1=paste( res1, mat, sep="" )  #### 20240707
   }
   if( debug ) Print("*2",res1, mat)
  }
  else{
   if( temp$type == "t" ){
    fj=temp$mat
    mat=paste( fj, "[", col, ",", sub0, "]", sep="" )
    if( vec ) mat=paste( fj, "[",  sub0, "]", sep="" )
   }
   if( lastdiag == 1 ){
    if( !vec )
     mat=paste( mat, "*", lastmat, "[", col, ",", col, "]", sep="" )
    else if( vec ) mat=paste( mat, "*", lastmat, "[", col, "]", sep="" )
   }
   res1=paste( res1, mat, sep="" )
   res1=paste( res1, res2, sep="" )
  }

  # term[i]=paste( "(", res1, ")" )
  term[i]=res1

  if( debug ) Print(res1)

 } # end of i loop for term


 # expand (A+B)[i,j]  -> (A[i,j]+B[i,j])
 if( expand ){
  for( i in 1:nterms ){
   term[i]=gsub("\\(([[:alnum:]])\\+([[:alnum:]])\\)(\\[[^]]+\\])"
             ,"(\\1\\3\\+\\2\\3)", term[i])
   term[i]=gsub("\\(([[:alnum:]])\\-([[:alnum:]])\\)(\\[[^]]+\\])"
                ,"(\\1\\3\\-\\2\\3)", term[i])
   term[i]=gsub("\\(([[:alnum:]])\\*([[:alnum:]])\\)(\\[[^]]+\\])"
             ,"(\\1\\3\\*\\2\\3)", term[i])
   ### ncol()  We need this to aoid error in ExpandSum
   term[i]=gsub("ncol\\(\\(([[:alnum:]])[-+*]([[:alnum:]])\\)\\)"
                ,"ncol(\\1)", term[i])

  }
 }

 if( debug ) Print( oper, term)

 # combine terms
 exp=paste( paste( " ", oper, " ", term, sep="" ), collapse="" )
 exp=trim(exp,leading=1)

 if( print ) Print(exp)

 return( exp )

} # end of mat2sum







#' Information of Nested Summations
#'
#' @param exp expression to be analyzed
#' @param sort = 1 to sort the result descendingly by level \cr
#' = -1 to sort the result ascendingly
#' @param print = 1 to print the result
#' @param debug = 1 to print intermediate result
#'
#' @details
#' This function analyses an expression containing summations.
#'
#' @return
#' A list of range, parent, level, sub, srange, body. \cr
#' where \cr
#' range starting and ending location of each summation:  s( body, {sub,range})
#' \cr
#' parent parent summation number: parent of top = 0 \cr
#' level  summation level number: level of top = 1 \cr
#' sub subscript of the summation\cr
#' srange range of the summation: a matrix consisting of (from,to) \cr
#' body body of the summation \cr
#' oper preceding operator or paren etc \cr
#' locspec starting location of the spec.
#'
#' @examples
#' sexp="s( s( (s(a3,{i3})+s(a4,{i4}))*s( s(a6,{i6}), {i5}), {i2}), {i1,1,n1})"
#' suminfo( sexp, print=1 )
#'
#'
#' @export
#'


suminfo <- function( exp, sort=0, print=0, debug=0 ){
 # get summation information
 # Shin-ichi Mayekawa
 # 20150930
 # character srange: 20151001
 # sort: 20151001
 # oper: 20151004
 # bugfix:20151005
 # print and sort=-1: 20151007
 #

 # compress exp
 exp00=exp
 exp=gsub( " ", "", exp )

 # find the location of s( ..., {} ).
 loc1=gregexpr( "s\\(", exp )[[1]]
 np=length(loc1)

 if( debug ) Print(exp, np,loc1)

 # Is there any paren?
 if( any(loc1 == -1) ){
  return(NULL)
 }


 # find the matching parens
 loc2=loc1
 for( i in 1:np ){
  loc2[i]=locbalpar( exp, loc1[i]+1 )
 }
 # range=mapply( ":", loc1, loc2 )
 range=cbind(loc1,loc2)
 if( debug ) Print(range, printrownames=1)

 # location of non summations
 loc3=which( substr(rep(exp,np),loc2-1,loc2-1) != "}" )

 if( length(loc3) > 0 ) range=range[-loc3,]
 ns=nrow(range)
 if( debug ) Print(  ns, range )

 # find structure
 parent=numeric(ns); level=parent; locspec=parent
 sub=character(ns); srange=character(ns); body=character(ns)

 # first summation
 parent[1]=0; level[1]=1
 lbp=locbalpar(exp,range[1,2]-1)
 locspec[1]=lbp
 temp=substr( exp, lbp+1, range[1,2]-2 )
 temp2=strsplit( temp, ",")[[1]]
 sub[1]=temp2[1]
 if( length(temp2) >= 2 ) srange[1]=temp2[2]
 if( length(temp2) >= 3 ) srange[1]=paste(srange[1],temp2[3],sep=",")
 body[1]=substr(exp,range[1,1]+2,lbp-2)
 if( debug ) Print(i,temp, lbp)

 if( ns >= 2 ){
  for( i in 2:ns ){

   # index
   lbp=locbalpar(exp,range[i,2]-1)
   locspec[i]=lbp
   temp=substr( exp, lbp+1, range[i,2]-2 )
   temp2=strsplit( temp, ",")[[1]]
   sub[i]=temp2[1]
   if( length(temp2) >= 2 ) srange[i]=temp2[2]
   if( length(temp2) >= 3 ) srange[i]=paste(srange[i],temp2[3],sep=",")
   body[i]=substr(exp,range[i,1]+2,lbp-2)
   if( debug ) Print(i,temp, lbp)

   # parent etc
   f=range[i,1]; t=range[i,2]
   loc4=which( range[1:(i-1),2] > range[i,2] )
   if( length(loc4) > 0 ){
    parent[i]=loc4[length(loc4)]
    level[i]=level[parent[i]]+1
   }
   else level[i]=1

  } # end of i loop
 }

 oper=character(ns)
 for( i in 1:ns ){
  if( range[i,1] > 1 ) oper[i]=substr(exp,range[i,1]-1,range[i,1]-1)
 }


 if( sort !=0 ){
  if( sort == 1 ) od=order(-level)
  else od=order(level)
  level=level[od]
  range=range[od,]
  parent=parent[od]
  locspec=locspec[od]
  sub=sub[od]
  srange=srange[od]
  body=body[od]
  oper=oper[od]
 }

 if( length(range) == 2 ) range=matrix(range,1) # 20151005

 if( print ) Print(range, parent,level, sub, srange, locspec, oper, body)


 return( named_list( range, parent, level, sub, srange, body, oper, locspec ) )


} # suminfo




#' Evaluate Summation:
#'
#' This function converts the summation to sum and evaluates it. \cr
#' \code{ s(exp[i],{i,1,3}) -> exp[1]+exp[2]+exp[3] }
#'
#' @param sexp An expression containing summation:  s( body, {sub,from,to} )
#' @param values replacement pattern for the ranges of subscripts
#' @param eval = 0 to leave Sigma notation
#' @param debug = 1 to print the intermediate result
#'
#' @details
#' For each summation found, replace it by its expansion as Sum.
#' The replacement proceeds from the bottom to top levels.
#'
#' @examples
#' sumEval( "s(s(a[i,j],{i,1,ni}),{j,1,nj})" , values="ni=2; nj=3")
#'
#' ( sexp=mat2sum("A%*%v",simple=0) )
#' sumEval( sexp, values="ncol(A)=3" )
#'
#' @export
#'

sumEval <- function( sexp, values=NULL, eval=1, debug=0 ){
 # Evaluate an expression containing summation
 # Shin-ichi Mayekawa
 # 20151001
 # ExpandSum renamed: 20151003
 # values: 20151006
 #


 Sigma <- function( body, sub, srange, eval=0, debug=10 ){
  # expand summation
  # Shin-ichi Mayekawa
  # 20151001
  #

  # from, to
  ss=strsplit(srange,",")[[1]]
  if( length(ss) < 2 ){
   cat("\nerror1:(Sigma) The input summation does not have full srange.\n")
   return()
  }

  pat=paste( "unlist(lapply(",ss[1], ":", ss[2]
             , ",function(x){gsub('", sub, "',x,'", body, "')}))", sep="" )
  pat=paste( "(Sum(", pat, "))", sep="" )
  if( debug ) Print(pat)

  if( eval ){
   pat=eval(parse(text=pat))
   if( debug ) Print("after eval", pat)
  }

  return( pat )

 } # end of Sigma



 ######################################################################
 # Simple case starts here.


 # check if sexp is a acalar
 if( length(sexp) > 1 ){
  cat("\nerror1:(sumEval) Input must not be a vector or a matrix.\n")
  return( sexp )
 }

 # compress sexp
 sexp=gsub( " ", "", sexp )

 # no summation
 if( regexpr(",\\{[^{}]+\\}",sexp)[[1]] == -1 ){
  return( sexp )
 }


 # evaluate with values
 #if( is.null(values) ){
 #  cat("\nerror1:(sumEval) values parameter is not given.\n")
 #  return( sexp )
 # }

 if( !is.null(values) )
  sexp=matReplace( sexp, values=values )

 if( debug ) Print("after eval", sexp)

 # sum info
 si=suminfo( sexp, sort=1 )
 if( is.null(si) ){
  return( sexp )
 }

 sub=si$sub
 body=si$body
 srange=si$srange
 ns=length(sub)

 if( debug ) Print(ns,body,sub,srange)

 # error check
 if( any( unlist(regexpr( "[[:alpha:]]", srange ) ) > 0 ) ){
  cat("\nerror1:(sumEval) Not all the ranges of the subscript are numeric.\n")
  cat("You must supply numeric values by values= parameter.\n")
  return( sexp )
 }

 if( ns >= 1 ){
  for( i in 1:ns ){

   if( debug ) Print("top of",i)
   target=paste("s(", body[i],",{",sub[i],",",srange[i],"})", sep="")
   pat=paste("(",Sigma( body[i], sub[i], srange[i], eval=eval, debug=debug )
             , ")", sep="")
   if( debug ) Print(pat)
   sexp=sub( target, pat, sexp, fixed=1 )
   body[]=sub( target, pat, body, fixed=1 )

   if( debug ) Print(target,"\n",sexp, body)

  } # end of i for summation
 }

 return( sexp )

} # end of sumEval




#' Expand (Distribute) summation
#'
#' @param sexp An expression containing summation
#' @param frombottom = 1 to expand the bottom level summation
#' @param all = 1 to expand all the summations
#' @param debug = 1 to print  intermediate result
#'
#' @details
#' \preformatted{
#' s( a + b, {scope} )  ->  s( a, {scope} ) + s( b, {scope} )
#' }
#'
#'
#' @examples
#' sumExpand("s( a + b, {scope} ) ")
#'
#' exp <- "s(a[i],{i})+s( b*( s(c[i,j],{j}) - s(d[i,j],{j}) ), {i} )"
#' sumExpand( exp)
#'
#' exp <- "s( a[i]+b[i] - s( c[ij]-d[ij], {j} ), {i})"
#' sumExpand( exp, frombottom=1)
#' sumExpand( exp )
#' sumExpand( exp, all=1 )
#'
#' @export
#'

sumExpand <- function( sexp, frombottom=0, all=0, debug=0 ){
 # Expand summation: s(body1+body2,{scope}) -> s(body,{scope})+s(body2,{scope})
 # Shin-ichi Mayekawa
 # 20151003,04
 # all: 20151005
 # no summation: 20151006
 #


 # check if sexp is a acalar
 if( length(sexp) > 1 ){
  cat("\nerror1:(sumExpand) Input must not be a vector or a matrix.\n")
  return( sexp )
 }


 if( all == 1 ){
  res0=sexp
  for( llll in 1:10 ){
   res=Recall( res0, frombottom=frombottom, all=0, debug=debug )
   if( res0 == res ) break
   res0=res
  }
  return( res )
 }



 ######################################################################
 # Simple case starts here.

 # compress and expand it first
 sexp=Expand(sexp)

 # no summation
 if( regexpr(",\\{[^{}]+\\}",sexp)[[1]] == -1 ){
  return( sexp )
 }

 temp=suminfo( sexp, sort=frombottom)
 ns=length(temp$level)

 if( debug ) Print(ns, sexp, temp)

 # find the summation whose body is expandable
 j=0
 for( i in 1:ns ){
  tempj=find_terms( temp$body[i] )
  if( tempj$nterms > 1 ){
   j=i; break
  }
 }

 if( j == 0 ){
  return( sexp )
 }

 ntj=tempj$nterms
 rangej=temp$range[j,]
 subj=temp$sub[j]
 operj=temp$oper[j]
 srangej=temp$srange[j]

 if( debug ){
  cat("\nThe ",j,"-th summation has an expandable body.\n")
  Print(tempj)
  Print(ntj,rangej,subj,srangej,operj)
 }

 top=NULL
 if( rangej[1] > 1 ) top=substr(sexp,1,rangej[1]-1)
 bottom=NULL
 if( rangej[2] < nchar(sexp) ) bottom=substr(sexp, rangej[2]+1, nchar(sexp) )
 right=paste( ",{", subj, srangej, "})", sep="" )

 if( debug ) Print( top, bottom, right)

 new=paste("s(", tempj$oper[1], tempj$term[1], right, sep="")
 if( ntj > 1 ){
  for( i in 2:ntj ){
   if( tempj$oper[i] == "+" )
    new=paste(new, "+", "s(", tempj$term[i], right, sep="")
   else if( tempj$oper[i] == "-" )
    new=paste(new, "-", "s(", tempj$term[i], right, sep="")
  }
 }

 res=paste( top, "(", new, ")", bottom, sep="")

 if( debug ) Print(res)

 res=Simplify( res )

 return( res )

} # end of sumExpand





#' Move summation to the right
#'
#' @param sexp An expression containing summation
#' @param frombottom = 1 to expand the bottom level summation
#' @param all = 1 to move all the movable summations
#' @param debug = 1 to print  intermediate result
#'
#' @details
#' \preformatted{
#' s( s( a[i,j]*b[j,k]*c[k,l], {k}) , {j}) ->
#'                          s( a[i,j]*s( b[j,k]*c[k,l], {k}) , {j})
#' }
#'
#'
#' @examples
#' sumMoveIn( "s( s( a[i,j]*b[j,k]*c[k,l], {k}) , {j})" )
#' sumMoveIn( "s( s( a[i,j]/b[j,k]*c[k,l], {k}) , {j})" )
#' sumMoveIn( "s( s( c[k,l]/a[i,j]/b[j,k], {k}) , {j})" )
#'
#' exp="s( x * s( s( a[i,j]*b[j,k]*c[k,l], {k}) , {j}),  {l})"
#' sumMoveIn( exp, all=1 )
#'
#' exp="s( x + s( s( a[i,j]*b[j,k]*c[k,l], {k}) - y, {j}),  {l})"
#' sumMoveIn( exp, all=1 )
#'
#' @export
#'

sumMoveIn <- function( sexp, frombottom=1, all=0, debug=0 ){
 # Move summation to the right
 # Shin-ichi Mayekawa
 # 20151004,05
 # all: 20151005
 # strict check for sub free: 20151005
 # no summation: 20151006
 #


 # check if sexp is a acalar
 if( length(sexp) > 1 ){
  cat("\nerror1:(sumMoveIn) Input must not be a vector or a matrix.\n")
  return( sexp )
 }


 if( all == 1 ){
  res0=sexp
  for( llll in 1:10 ){
   if( debug ) cat("\n\n####### Top of fixed point iteration ", llll,"...\n")
   res=Recall( res0, frombottom=frombottom, all=0, debug=debug )
   if( res0 == res ) break
   res0=res
  }
  return( res )
 }



 ######################################################################
 # Simple case starts here.


 # compress and expand it first
 sexp=sumExpand(sexp, all=1)

 # no summation
 if( regexpr(",\\{[^{}]+\\}",sexp)[[1]] == -1 ){
  return( sexp )
 }

 # Assuming that the body consists of one term.

 # sum info
 temp=suminfo( sexp, sort=frombottom)
 ns=length(temp$level)

 if( debug ) Print(ns, sexp, temp)

 found=0
 for( i in 1:ns){

  # find the movable summation
  range=temp$range[i,]
  sub=temp$sub[i]
  body=temp$body[i]
  srange=temp$srange[i]

  if( nterms(body) > 1 ){
   if( debug ) Print(body)
   cat("\nwarning1:(sumMoveIn) Use this function when the expression"
       ," can be expanded to the form where each summation contain"
       ," only one term\n")
  }

  # decompose body into factors
  temp1=find_factors( body )
  nf=temp1$nfactors
  foper=temp1$oper
  fact=temp1$factor

  # find those factors w/o subscript sub
  nosub=numeric(nf)
  for( j in 1:nf ){
   # pat=paste( "(\\[", sub, ")|(", sub, "\\])", sep="")
   pat=paste( "(\\[",sub,"\\])|(\\[", sub, ",)|(,", sub, "\\])", sep="")
   if( regexpr( pat,fact[j] )[[1]]  < 0 ) nosub[j]=1
  }
  if( debug ) Print(nf,foper,fact,nosub,sub)

  # see if there are sub-free factors
  if( !all( nosub == 0 ) ){
   found=1
   soper2="*"
   if( i < ns ) soper2=temp$oper[i+1]
   if( soper2 %in% c("+","-", "","(") ) soper2=""
   break
  }

 } # end of i loop

 if( !found ) return( sexp )

 if( debug ){
  Print("summation found!!.", i,temp1)
 }


 # one or more factors free of sub found.
 loc=which(nosub==1)
 loc2=which(nosub==0)
 if( foper[loc[1]] == "/" ){
  foper[loc[1]]=""; fact[loc[1]]=paste("1/",fact[loc[1]],sep="")
 }
 if( length(loc2) > 0  && foper[loc2[1]] == "/" ){
  fact[loc2[1]]=paste("1/",fact[loc2[1]],sep="")
 }
 if( length(loc2) > 0 ) foper[loc2[1]]=""
 soper="*"

 # sub-free factors
 nosubf=expr <- paste( paste( foper[loc], fact[loc], sep="" ), collapse="" )
 # factos with sub
 subf=expr <- paste( paste( foper[loc2], fact[loc2], sep="" ), collapse="" )

 if( debug ) Print(loc,sub,nosubf,subf)

 # move summation in
 if( subf != "" ){
  new=paste(nosubf,soper,"s(",subf,",{",sub,srange,"})", sep="")
 }
 else{
  new=paste(nosubf,"*range(",sub,")",soper2, sep="")
 }
 if( debug ) Print(new)

 top=NULL
 if( range[1] > 1 ) top=substr(sexp,1,range[1]-1)
 bottom=NULL
 if( range[2] < nchar(sexp) ) bottom=substr(sexp, range[2]+1, nchar(sexp) )

 res=paste(top,new,bottom,sep="")

 # final
 res=gsub("**","*",res,fixed=1)
 if( substr(res,nchar(res),nchar(res)) == "*" )
  res=strtrim(res,nchar(res)-1)

 if( debug ) Print(res)

 return( res )


} # end of sumMoveIn


#' Simplify an expression containing summation
#' @param sexp expression containing summation
#' @details This is sumMoveIn function with all=1 and frombottom=0. \cr
#' @examples
#' sexp="s( x + s( s( a[i,j]*b[j,k]*c[k,l], {k}) - y, {j}),  {l})"
#' sumSimplify( sexp )
#' @export
#'

sumSimplify <- function( sexp ){
 sumMoveIn( sexp, frombottom=0, all=1 )
} # end of sumSimplify



