#' Transposition of a Fully Symbolic Partitioned Matrix
#'
#' @param A fully symbolic partitioned matrix
#' @param func = "transp" if necessary.
#' @param func2 = transp if  necessary.
#'
#' @details
#' tp is defined as \cr
#' \code{ map( func, func2(A) ) }
#'
#' If A does not have fullsymb attribute nore SMdim attribute, that is, \cr
#' \code{ is.null(attributes(A)$fullsymb) == TRUE}, and \cr
#' \code{ is.null(attributes(A)$SEdim) == TRUE}, \cr
#' \code{tp} is equal to \code{t}. \cr
#' If A has SMdim attribute, \code{tp} is equal to \code{transp}.
#'
#' @examples
#' # Standard Matrix
#' A=demomat(2,3,root="a")
#' tA=tp(A)
#' Printb(tA)
#'
#' # Semi Symbolic Partitioned Matrix:  Super Matrix with SMdim attributes
#' A=demomat(c(2,3),c(3,4),root="a")
#' Printb(A)
#' tA=tp(A)
#' Printb(tA)
#'
#' # Fully Symbolic Partitioned Matrix:
#' # No SMdim attribute but fullsymb attribute
#' Ap=demomat(2,3,root="A", fullsymb=1)
#' Printb(Ap)
#' tAp=tp(Ap)
#' Printb(tAp)
#' 
#' tp(demomat(3,3,root="A", fullsymb=1,shape="diag"))
#' 
#' @export
#'

tp <- function( A, func="t", func2=t ){
 # transposition of a fully symbolic matrix
 # Shin-ichi Mayekawa
 # 20200526,27
 # simplify: 20200527,28
 #

 if( !is.null(attributes(A)$SMdim) ){
  return( transp(A) )
 }
 else if( is.null(attributes(A)$fullsymb) ){
  return( t(A) )
 }

 res=map(func, func2(A))

 # simplify some trivial result
 #   1        2      3     4          5    6
 #   (top | *)( t(  )(sign)( number )( )  )($ or */ ) -> top sign num $
 v3="(^|[-+*`~])(t\\()([-+]*)([0-9]*)(\\))($|[-+*/`~])"
 res=gsub( v3, "\\1\\3\\4\\6", res )

 #   1        2        3     4           5      6
 #   (top | *)( t(t(  )(sign)( variable )( ))  )($ or */ ) -> top sign var $
 v3="(^|[-+*`~])(t\\(t\\()([-+]*)([[:alpha:]][0-9_]*)(\\)\\))($|[-+*/`~])"
 res=gsub( v3, "\\1\\3\\4\\6", res )

 #   1        2        3     4         5      6
 #   (top | *)( t(t(  )(sign)( number )( ))  )($ or */ ) -> top sign num $
 v3="(^|[-+*`~])(t\\(t\\()([-+]*)([0-9]*)(\\)\\))($|[-+*/`~])"
 res=gsub( v3, "\\1\\3\\4\\6", res )

 return( res )

} # end of tp


