#' Conversion of the vec of Matrix Product using vec operator
#' and Kronecker Product
#'
#' @param A The first matrix or Identity matrix
#' @param B The second matrix or Identity matrix
#' @param C The third matrix or Identity matrix
#'
#' @details
#' This function converts \code{vec(A\%*\%B\%*\%C)} to
#' \code{(t(C)\%@\%A) \%*\% vec(B)}. \cr
#'
#' Note that the input matrices can be numeric, semi symbolic,
#' or full symbolic.
#'
#' When the number of input matrices is less than 3,
#' by choosing the matrices to be set equal to I,
#' various results can be obtained.
#'
#' @return
#' An expression as the result of conversion.
#'
#'
#' @examples
#' # full symbolic result
#' # general form
#' vec2vecK( "A","B","C")
#'
#' # with two matrices
#' vec2vecK( "A","B")
#' vec2vecK( "A",,"B")
#' vec2vecK( ,"A","B")
#'
#' \dontrun{
#' # checking with semi symbolic matrices
#' A <- demomat(2,2,root="a")
#' B <- demomat(2,2,root="b")
#' C <- demomat(2,2,root="c")
#'
#' # general form
#' res1 <- vec(A%*%B%*%C)
#' res2 <- vec2vecK(A, B, C)
#'
#' # two matrices
#' v0 <- vec(A%*%B)
#' v1 <- vec2vecK( A, B, diag(2) )
#' v2 <- vec2vecK( A, diag(2), B )
#' v3 <- vec2vecK( diag(2), A, B )
#' Print(v0,v1,v2,v3)
#' }
#' 
#' @export
#'

vec2vecK <- function( A="I", B="I", C="I" ){
 # vec to vec and Kronecker Product
 # Shin-ichi Mayekawa
 # 20200629
 #

 if( !is.null(attributes(A)$fullsymb) || !is.null(attributes(B)$fullsymb)
     || !is.null(attributes(C)$fullsymb) ){
  cat("\n\nerror1:(tr2vecK) Cannot handle fully symbolic partitioned"
      , "matrices.\n\n")
  return( NULL )
 }


 if( length(A)==1 && length(B)==1 && length(C)==1 ){
  # fully symbolic case
  # v=(t(C)%@%A) %*% vec(B)
  v=paste("(t(", C, ")%@%(", A, "))%*%vec(", B ,")", sep="" )
  v=gsub( "t(I)", "I", v, fixed=1 )
  v=gsub( "t\\(t\\(([[:alpha:]][[:alnum:]_]*)\\)\\)", "\\1", v )
 }
 else{
  # semi symbolic case
  v=matTimes( matKTimes(t(C),A, ra=1), vec(B) )
 }

 return( v )

} # end of vec2vecK



