% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/invGE.R
\name{invGE}
\alias{invGE}
\title{Matrix Inversion by Gaussian Elimination}
\usage{
invGE(A, loc = 1:nrow(A), nopermute = 1, eps = 1e-09, simplify = 1, print = 0)
}
\arguments{
\item{A}{Input Matrix (nueric or symbolic)}

\item{loc}{vector of pivot locations}

\item{nopermute}{= 0 to find the best pivot when A is numeric.}

\item{eps}{eps for zero.}

\item{simplify}{= 0 not to implify the result when A is character.}

\item{print}{= 1 to print the intermediate results.}
}
\value{
\code{ cbind(I, inv(A) } \cr
The inverse of \code{A} is given as \code{cbind(I,inv{A})[,-c(1:nrow(A))]}.
}
\description{
Matrix Inversion by Gaussian Elimination
}
\details{
This function performs the Gaussian elimination on \code{A} matrix.
 When \code{loc=1:nrow(A)},
 the elimination starts from column 1 to column \code{n=ncol(A)} as:
 \preformatted{
  B0 <- cbind( A, I )
  B1 <- T1 \%*\% B0
  B2 <- T2 \%*\% B1
   :        :
  Bk <- Tk \%*\% Bk-1
   :        :
  Bn <- Tn \%*\% Bn-1
 }
 where Tk matrix makes the k-th column of Bk matrix a unit vector
 whose k-th element is 1 and 0 elsewhere.
 \code{B[i,i]} is used as the pivot.

 At the end of the iteration, we have
 \preformatted{
   Bn = cbind( I, inv(A) )
 }
 In other words
 \preformatted{
  inv(A) = Tn \%*\% Tn-1 \%*\% ... \%*\% T2 \%*\% T1
 }

When \code{loc != 1:nrow(A)}, at each step, \code{B[loc[i],loc[i]]} is used
as the pivot.

When \code{length(loc) < nrow(A)}, the intermediate result will be returned.
}
\examples{

# Symbolic matrix A.

n <- 2
Aa <- demomat(n,n,root="a")

set.seed(1701)
An <- matrix(rnorm(n*n),n,n)

res <- invGE(Aa, print=1)
II <- res[,c(1:n)]
invAa <- res[,-c(1:n)]

# numerical checking
Eval(II, Aa=An)
# Eval(Aa\%*\%invAa, Aa=An)
Eval(matTimes(Aa,invAa), Aa=An)


# Numeric matrix A.

set.seed(1702)
n <- 4
A <- matrix( sample(0:9,n*n,replace=1), n,n )

# result
res <- invGE(A, print=1)
invA <- res[,-(1:n)]
printm(A, invA, A\%*\%invA, fmt=".2")



# Example of permutation
# needs permutation
set.seed(1701)
n=4
A=matrix( sample(0:9,n*n,replace=1), n,n )

# wrong result
invA=invGE(A, print=1)[,c(1:n)]
printm( A\%*\%invA )

# correct result by nopermute=0 option
invA=invGE(A, print=1, nopermute=0)[,-c(1:n)]
printm( A\%*\%invA, fmt="7.3"  )

# correct by specifying loc argument
invA=invGE(A, print=1, loc=c(1,4,3,2))[,-c(1:n)]
printm( A\%*\%invA, fmt="6.3" )



# non full rank matrix A.
set.seed(1701)
n <- 4
A <- matrix( sample(0:9,n*n,replace=1), n,n )
A[,4] <- A[,1]+A[,2]

res <- invGE(A, print=1, nopermute=0)
invA <- res[,-c(1:n)]
round( A\%*\%invA ,6 )
chkginv(A,invA)



}
